/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.math.BigDecimal;
import java.util.HashSet;
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.core.OutOfMemoryErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;
/**
 * Abstract class to be extended by all Grid2DSquareCellAbstract 
 * factories.
 */
public abstract class Grid2DSquareCellAbstractFactory {
    
    /**
     * A directory for swapping.
     */
    protected File directory;
    
    /**
     * The number of rows in a chunk.
     */
    protected int chunkNrows;
    
    /**
     * The number of columns in a chunk.
     */
    protected int chunkNcols;
    
    /**
     * The Dimensions
     */
    protected BigDecimal[] dimensions;
    
    /**
     * A container of other Grid2DSquareCellAbstract references.
     */
    protected HashSet grid2DSquareCells;
    
    /**
     * The GridStatisticsAbstract
     */
    protected GridStatisticsAbstract gridStatistics;
    
    /**
     * handleOutOfMemoryError
     */
    protected boolean handleOutOfMemoryError;
    
    /**
     * Returns a copy of this.directory.
     */
    public File getDirectory() {
        return new File( this.directory.getPath() );
    }
    
    /**
     * Sets this.directory to directory.
     */
    public void setDirectory( 
            File directory ) {
        this.directory = directory;
    }
    
    /**
     * Returns a copy of this.chunkNrows.
     */
    public int getChunkNrows() {
        return this.chunkNrows;
    }
    
    /**
     * Sets this.chunkNrows to chunkNrows
     */
    public void setChunkNrows( 
            int chunkNrows ) {
        this.chunkNrows = chunkNrows;
    }
    
    /**
     * Returns a copy of this.chunkNcols.
     */
    public int getChunkNcols() {
        return this.chunkNcols;
    }
    
    /**
     * Sets this.chunkNcols to chunkNcols
     */
    public void setChunkNcols( 
            int chunkNcols ) {
        this.chunkNcols = chunkNcols;
    }
    
    /**
     * Returns this.dimensions
     */
    public BigDecimal[] getDimensions() {
        return this.dimensions;
    }
        
    /**
     * Sets this.dimensions to dimensions
     */
    public void setDimensions( 
            BigDecimal[] dimensions ) {
        this.dimensions = dimensions;
    }
    
    /**
     * Returns this.grid2DSquareCells
     */
    public HashSet getGrid2DSquareCells() {
        return this.grid2DSquareCells;
    }
    
    /**
     * Sets this.grid2DSquareCells to grid2DSquareCells
     */
    public void setGrid2DSquareCells( 
            HashSet grid2DSquareCells ) {
        this.grid2DSquareCells = grid2DSquareCells;
    }
    
    /**
     * Returns this.gridStatistics
     */
    public GridStatisticsAbstract getGridStatistics() {
        return this.gridStatistics;
    }
    
    /**
     * Sets this.gridStatistics to gridStatistics
     */
    public void setGridStatistics( 
            GridStatisticsAbstract gridStatistics ) {
        this.gridStatistics = gridStatistics;
    }
    
    /**
     * Returns this.handleOutOfMemoryError
     */
    public boolean getHandleOutOfMemoryError() {
        return this.handleOutOfMemoryError;
    }
    
    /**
     * Sets this.handleOutOfMemoryError to handleOutOfMemoryError
     */
    public void setHandleOutOfMemoryError( 
            boolean handleOutOfMemoryError ) {
        this.handleOutOfMemoryError = handleOutOfMemoryError;
    }
    
    //////////////////////
    // Default Creation //
    //////////////////////
    
    /**
     * Returns an Grid2DSquareCellAbstract loaded from this.directory.
     */
    public Grid2DSquareCellAbstract create() {
        return create( this.directory );
    }
    
    /////////////////////////
    // Create from scratch //
    /////////////////////////
    
    /**
     * Returns an Grid2DSquareCellAbstract with all values as
     * noDataValues.
     * @param nrows The nrows for the construct.
     * @param ncols The ncols for the construct.
     */
    public Grid2DSquareCellAbstract create(
            long nrows,
            long ncols ) {
        return create( 
                nrows, 
                ncols, 
                this.dimensions );
    }
    
    /**
     * Returns an Grid2DSquareCellAbstract with all values as
     * noDataValues.
     * @param nrows The nrows for the construct.
     * @param ncols The ncols for the construct.
     * @param dimensions The cellsize and bounding box details for the construct.
     */
    public Grid2DSquareCellAbstract create(
            long nrows,
            long ncols,
            BigDecimal[] dimensions ) {
        return create( 
                FileCreator.createNewFile( this.directory ),
                nrows, 
                ncols,
                dimensions );
    }
    
    /**
     * Returns an Grid2DSquareCellAbstract with all values as
     * noDataValues.
     * @param directory The directory for swapping to file.
     * @param nrows The nrows for the construct.
     * @param ncols The ncols for the construct.
     * @param dimensions The cellsize and bounding box details for the construct.
     */
    public Grid2DSquareCellAbstract create(
            File directory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions ) {
        return create( 
                directory, 
                nrows, 
                ncols, 
                dimensions, 
                this.grid2DSquareCells,
                this.handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract grid with all values as
     * noDataValues.
     * @param directory The directory for swapping to file.
     * @param nrows The nrows for the construct.
     * @param ncols The ncols for the construct.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public abstract Grid2DSquareCellAbstract create(
            File directory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError );
    
    
    //////////////////////////////////////////////////////
    // Create from an existing Grid2DSquareCellAbstract //
    //////////////////////////////////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellAbstract with all values as int values from
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     *   are obtained.
     */
    public Grid2DSquareCellAbstract create(
            Grid2DSquareCellAbstract grid2DSquareCell ) {
        return create(
                this.directory, 
                grid2DSquareCell,
                0L, 
                0L, 
                grid2DSquareCell.getNrows( handleOutOfMemoryError ) - 1L, 
                grid2DSquareCell.getNcols( handleOutOfMemoryError ) - 1L );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * grid2DSquareCell.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     *   are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats
     *   values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats
     *   values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats
     *   values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats
     *   values are used.
     */
    public Grid2DSquareCellAbstract create(
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex ) {
        File file = FileCreator.createNewFile( 
                this.directory,
                "" + startRowIndex + "_" + startColIndex + "_" + endRowIndex + "_" + endColIndex );
        return create( 
                file, 
                grid2DSquareCell, 
                startRowIndex, 
                startColIndex, 
                endRowIndex,
                endColIndex );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing data in files.
     *   Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     *   are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats
     *   values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats
     *   values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats
     *   values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats
     *   values are used.
     */
    public Grid2DSquareCellAbstract create(
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex ) {
        return create( 
                directory, 
                grid2DSquareCell,
                startRowIndex,
                startColIndex, 
                endRowIndex,
                endColIndex,
                this.grid2DSquareCells, 
                this.handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing data in files.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     *   are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats
     *   values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats
     *   values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats
     *   values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats
     *   values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public abstract Grid2DSquareCellAbstract create(
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError );
    
    
    ////////////////////////
    // Create from a File //
    ////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * gridFile. If gridFile is a directory then it is assumed to contain a
     * file called cache which can be opened into an object input stream and
     * initailised as an instance of a class extending Grid2DSquareCellAbstract.
     * @param gridFile either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellAbstract to be returned.
     */
    public Grid2DSquareCellAbstract create ( 
            File gridFile ) {
        if ( gridFile.isDirectory() ) {
            // Initialise from File(gridFile,"this")
            File thisFile = new File( 
                    gridFile, 
                    "thisFile" );
            try {
                ObjectInputStream objectInputStream = new ObjectInputStream( 
                        new BufferedInputStream( 
                        new FileInputStream( thisFile ) ) );
                return create( 
                        this.directory,
                        gridFile,
                        objectInputStream );
            } catch ( Exception e0 ) {
                System.out.println( e0 );
                e0.printStackTrace();
            }
        }
        // Assume it is ESRI asciigrid
        ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter( gridFile );
        Object[] header = eSRIAsciigridImporter.readHeaderObject();
        long ncols = ( Long ) header[ 0 ];
        long nrows = ( Long ) header[ 1 ];
        double noDataValue = ( Double ) header[ 5 ];
        eSRIAsciigridImporter.close();
        String gridName = gridFile.getName().substring( 0, gridFile.getName().length() - 4 );
        String directoryName = new String (
                gridFile.getParentFile() +
                System.getProperty( "file.separator" ) + 
                gridName + 
                this.getClass().getName() +
                "_chunkNrows(" + chunkNrows + ")_chunkNcols(" + chunkNcols + ")" );
        //this.directory = FileCreator.createNewFile( new File( directoryName ) );
        this.directory = new File( directoryName );
        this.directory.mkdirs();
        return create( 
                this.directory, 
                gridFile, 
                0L, 
                0L, 
                nrows - 1L, 
                ncols - 1L );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellAbstract to be returned.
     * @param startRowIndex The topmost row index of the grid represented in
     *   gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in
     *   gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in
     *   gridFile thats values are used.
     * Default:
     * grid2DSquareCells to null;
     * handleOutOfMemoryError to true.
     */
    public Grid2DSquareCellAbstract create(
            File directory,
            File gridFile,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex ) {
        return create( 
                directory, 
                gridFile, 
                startRowIndex, 
                startColIndex, 
                endRowIndex, 
                endColIndex, 
                this.grid2DSquareCells, 
                this.handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellAbstract to be returned.
     * @param startRowIndex The topmost row index of the grid represented in
     *   gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in
     *   gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public abstract Grid2DSquareCellAbstract create (
            File directory,
            File gridFile,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError );
            
    
    /////////////////////////
    // Create from a cache //
    /////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from gridFile.
     * @param directory The directory for swapping to file.
     * @param gridFile A file containing the data to be used in construction.
     * @param ois The ObjectInputStream to construct from.
     */
    public Grid2DSquareCellAbstract create(
            File directory,
            File gridFile,
            ObjectInputStream ois ) {
        return create( 
                directory, 
                gridFile,
                ois, 
                this.grid2DSquareCells, 
                this.handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from gridFile.
     * @param directory The directory for swapping to file.
     * @param gridFile A file containing the data to be used in construction.
     * @param ois The ObjectInputStream to construct from.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public abstract Grid2DSquareCellAbstract create(
            File directory,
            File gridFile,
            ObjectInputStream ois,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError );
}
