/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.lang.OutOfMemoryError;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciiGridImporter;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;

/**
 * A class for representing grids of double precision values.
 * @see Grid2DSquareCellAbstract
 */
public class Grid2DSquareCellDouble extends Grid2DSquareCellAbstract implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * For storing the NODATA value of the grid, which by default is
     * Double.NEGATIVE_INFINITY.
     * N.B. Double.NaN should not be used.
     * N.B. Care should be taken so that noDataValue is not a data value.
     */
    protected double noDataValue = Double.NEGATIVE_INFINITY;
    
    /**
     * Creates a new Grid2DSquareCellDouble
     */
    public Grid2DSquareCellDouble() {
        initGrid2DSquareCellDouble();
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble.
     * Warning!! Concurrent modification may occur if directory is in use.
     * If a completely new instance is wanted then use:
     * Grid2DSquareCellDouble(
     *   File,
     *   Grid2DSquareCellDoubleAbstract,
     *   Grid2DSquareCellDoubleChunkAbstractFactory,
     *   int,
     *   int,
     *   long,
     *   long,
     *   long,
     *   long,
     *   double,
     *   HashSet)
     * which can be accessed via a Grid2DSquareCellDoubleFactory.
     * @param directory The File directory to be used for swapping.
     * @param gridFile The File directory containing the File names thisFile that
     *   the ois was constructed from.
     * @param ois The ObjectInputStream used in first attempt to construct this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellDouble(
            File directory,
            File gridFile,
            ObjectInputStream ois,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        initGrid2DSquareCellDouble(
                directory,
                gridFile,
                ois,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble with each cell value equal to
     * noDataValue and all chunks of the same type.
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered for creating chunks.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param nrows The number of rows of cells.
     * @param ncols The number of columns of cells.
     * @param dimensions The cellsize, xmin, ymin, xmax and ymax.
     * @param noDataValue The noDataValue.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        initGrid2DSquareCellDouble(
                gridStatistics,
                directory,
                grid2DSquareCellDoubleChunkFactory,
                chunkNrows,
                chunkNcols,
                nrows,
                ncols,
                dimensions,
                noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble based on values in grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which this is
     *   to be constructed.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param startRowIndex The grid2DSquareCell row index which is the bottom
     *   most row of this.
     * @param startColIndex The grid2DSquareCell column index which is the left
     *   most column of this.
     * @param endRowIndex The grid2DSquareCell row index which is the top most
     *   row of this.
     * @param endColIndex The grid2DSquareCell column index which is the right
     *   most column of this.
     * @param noDataValue The noDataValue for this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        initGrid2DSquareCellDouble(
                gridStatistics,
                directory,
                grid2DSquareCell,
                grid2DSquareCellDoubleChunkFactory,
                chunkNrows,
                chunkNcols,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Creates a new Grid2DSquareCellDouble with values obtained from gridFile.
     * Currently gridFile must be a directory of a Grid2DSquareCellDouble or
     * Grid2DSquareCellInt or a ESRI Asciigrid format file with a name ending
     * ".asc".
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param startRowIndex The grid2DSquareCell row index which is the bottom
     *   most row of this.
     * @param startColIndex The grid2DSquareCell column index which is the left
     *   most column of this.
     * @param endRowIndex The grid2DSquareCell row index which is the top most
     *   row of this.
     * @param endColIndex The grid2DSquareCell column index which is the right
     *   most column of this.
     * @param noDataValue The noDataValue for this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are 
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            File gridFile,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        initGrid2DSquareCellDouble(
                gridStatistics,
                directory,
                gridFile,
                grid2DSquareCellDoubleChunkFactory,
                chunkNrows,
                chunkNcols,
                startRowIndex,
                startColIndex,
                endRowIndex,
                endColIndex,
                noDataValue,
                grid2DSquareCells,
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a string description of the instance. Basically the values of
     * each field.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are 
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String toString( 
            boolean handleOutOfMemoryError ) {
        try {
            return "Grid2DSquareCellDouble( " +
                    //"noDataValue( " + getNoDataValue( handleOutOfMemoryError ) + " ), " +
                    "noDataValue( " + this.noDataValue + " ), " +
                    super.toString( 0, handleOutOfMemoryError ) + " )";
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return toString( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.
     * @see Grid2DSquareCellDouble()
     */
    private void initGrid2DSquareCellDouble() {
        try {
            initGrid2DSquareCellAbstract();
            this.noDataValue = Double.NEGATIVE_INFINITY;
            initMemoryReserve();
        } catch ( OutOfMemoryError oome0 ) {
            System.err.println(
                    "OutOfMemoryError \n" + 
                    this.getClass().toString() +
                    ".initGrid2DSquareCellDouble(); " );
            oome0.printStackTrace();
            throw oome0;
        }
    }
    
    /**
     * Initialises this.
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble from which
     *   the fields of this are set.
     * @param initTransientFields Iff true then transient fields of this are set
     *   with those of grid2DSquareCellDouble.
     */
    private void initGrid2DSquareCellDouble(
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            boolean initTransientFields ) {
        try {
            this.noDataValue = grid2DSquareCellDouble.noDataValue;
            super.initGrid2DSquareCellAbstract( grid2DSquareCellDouble );
            if ( initTransientFields ) {
                this.grid2DSquareCellChunks = 
                        grid2DSquareCellDouble.grid2DSquareCellChunks;
                this.grid2DSquareCells = 
                        grid2DSquareCellDouble.grid2DSquareCells;
                this.memoryReserve = 
                        grid2DSquareCellDouble.memoryReserve;
            }
        } catch ( OutOfMemoryError oome0 ) {
            System.err.println(
                    "OutOfMemoryError \n" + 
                    this.getClass().toString() +
                    ".initGrid2DSquareCellDouble( \n" +
                    "    Grid2DSquareCellDouble, \n" +
                    "    boolean ); " );
            oome0.printStackTrace();
            throw oome0;
        }
    }
    
    /**
     * Initialises this.
     * @see Grid2DSquareCellDouble(
     *      File,
     *      File,
     *      ObjectInputStream,
     *      HashSet,
     *      boolean)
     * @param directory The File directory to be used for swapping.
     * @param gridFile The File directory containing the File named thisFile that
     *   the ois was constructed from.
     * @param ois The ObjectInputStream used in first attempt to construct this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are 
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            File directory,
            File gridFile,
            ObjectInputStream ois,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells = grid2DSquareCells;
            initGrid2DSquareCellDouble(
                    directory,
                    gridFile,
                    ois );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCells = grid2DSquareCells;
                //swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initGrid2DSquareCellDouble(
                            directory,
                            gridFile,
                            ois,
                            grid2DSquareCells,
                            handleOutOfMemoryError );
                } else {
                    throw oome0;
                }
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.
     * @see Grid2DSquareCellDouble(
     *      File,
     *      File,
     *      ObjectInputStream,
     *      HashSet,
     *      boolean)
     * @param directory The File directory to be used for swapping.
     * @param gridFile The File directory containind the File thisFile that the
     *   ois was constructed from.
     * @param ois The ObjectInputStream used in first attempt to construct this.
     */
    private void initGrid2DSquareCellDouble(
            File directory,
            File gridFile,
            ObjectInputStream ois ) {
        this.directory = directory;
        if ( this.grid2DSquareCells.size() > 0 ) {
            Iterator ite = this.grid2DSquareCells.iterator();
            this.memoryReserve = ( ( Grid2DSquareCellAbstract ) ite.next() ).memoryReserve;
        } else {
            initMemoryReserve();
        }
        File thisFile = new File(
                gridFile,
                "thisFile" );
        try {
            initGrid2DSquareCellDouble(
                    ( Grid2DSquareCellDouble ) ois.readObject(),
                    false );
                    //true );
            ois.close();
            // Set the refernce to this in the Grid Chunks
            initGrid2DSquareCellChunks( gridFile );
            Iterator chunkIterator = this.grid2DSquareCellChunks.values().iterator();
            while ( chunkIterator.hasNext() ) {
                Grid2DSquareCellChunkAbstract grid2DSquareCellChunkAbstract = ( Grid2DSquareCellChunkAbstract ) chunkIterator.next();
                grid2DSquareCellChunkAbstract.grid2DSquareCell = this;
            }
            // Set the reference to this in the Grid Statistics  
            this.gridStatistics.grid2DSquareCell = this;
        } catch ( Exception e0 ) {
            //if ( e0 instanceof InvalidClassException || e0 instanceof ClassCastException ) {
            if ( e0.getClass() == ClassCastException.class ) {
                try {
                    ois.close();
                    ois = new ObjectInputStream(
                            new BufferedInputStream(
                            new FileInputStream( thisFile ) ) );
                    // If the object is a Grid2DSquareCellInt
                    Grid2DSquareCellIntFactory grid2DSquareCellIntFactory = new Grid2DSquareCellIntFactory( directory );
                    Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCellIntFactory.create(
                            directory,
                            gridFile,
                            ois );
                    Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory( directory );
                    Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCellDoubleFactory.create( grid2DSquareCellInt );
                    initGrid2DSquareCellDouble(
                            grid2DSquareCellDouble,
                            false );
                    initGrid2DSquareCellChunks( gridFile );
                } catch ( IOException ioe0 ) {
                    ioe0.printStackTrace();
                }
            } else {
                System.err.println( e0.getMessage() );
                //throw e0;
            }
        }
        //initGrid2DSquareCellChunks( gridFile );
        this.grid2DSquareCells.add( this );
    }
    
    /**
     * Initialises this.
     * @see Grid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long ,
     *      BigDecimal[],
     *      double,
     *      HashSet,
     *      boolean );
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered for creating chunks.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param nrows The number of rows of cells.
     * @param ncols The number of columns of cells.
     * @param dimensions The cellsize, xmin, ymin, xmax and ymax.
     * @param noDataValue The noDataValue.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells = grid2DSquareCells;
            initGrid2DSquareCellDouble(
                    gridStatistics,
                    directory,
                    grid2DSquareCellDoubleChunkFactory,
                    chunkNrows,
                    chunkNcols,
                    nrows,
                    ncols,
                    dimensions,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCells = grid2DSquareCells;
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initGrid2DSquareCellDouble(
                            gridStatistics,
                            directory,
                            grid2DSquareCellDoubleChunkFactory,
                            chunkNrows,
                            chunkNcols,
                            nrows,
                            ncols,
                            dimensions,
                            noDataValue,
                            grid2DSquareCells,
                            handleOutOfMemoryError );
                } else {
                    throw oome0;
                }
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.
     * @see Grid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long ,
     *      BigDecimal[],
     *      double,
     *      HashSet,
     *      boolean );
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered for creating chunks.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param nrows The number of rows of cells.
     * @param ncols The number of columns of cells.
     * @param dimensions The cellsize, xmin, ymin, xmax and ymax.
     * @param noDataValue The noDataValue.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            double noDataValue ) {
        this.directory = directory;
        if ( this.grid2DSquareCells.size() > 0 ) {
            Iterator ite = this.grid2DSquareCells.iterator();
            this.memoryReserve = ( ( Grid2DSquareCellAbstract ) ite.next() ).memoryReserve;
        } else {
            initMemoryReserve();
        }
        this.gridStatistics = gridStatistics;
        this.gridStatistics.init( this );
        this.directory = directory;
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.nrows = nrows;
        this.ncols = ncols;
        this.dimensions = dimensions;
        initNoDataValue( noDataValue );
        this.grid2DSquareCells = grid2DSquareCells;
        this.name = directory.getName();
        initNChunkRows();
        initNChunkCols();
        long nChunks = getNChunks();
        this.grid2DSquareCellChunks = new HashMap( ( int ) nChunks );
        int chunkRowIndex = Integer.MIN_VALUE;
        int chunkColIndex = Integer.MIN_VALUE;
        int loadedChunkCount = 0;
        boolean isLoadedChunk = false;
        boolean grid2DSquareCellsSwapped = false;
        int cri0 = 0;
        int cci0 = 0;
        int cci1 = 0;
        ChunkID chunkID = new ChunkID();
        Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                this,
                chunkID );
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                do {
                    try {
                        // Try to load chunk.
                        chunkID = new ChunkID(
                                this.nChunkCols,
                                chunkRowIndex,
                                chunkColIndex );
                        grid2DSquareCellDoubleChunk = grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                                this,
                                chunkID );
                        chunkID = new ChunkID(
                                this.nChunkCols,
                                chunkRowIndex,
                                chunkColIndex );
                        this.grid2DSquareCellChunks.put(
                                chunkID,
                                grid2DSquareCellDoubleChunk );
                        isLoadedChunk = true;
                    } catch ( OutOfMemoryError oome0 ) {
                        // Encountering an OutOfMemoryError while executing this block is fatal.
                        clearMemoryReserve();
                        System.out.println(
                                "Encountered an OutOfMemoryError initialising " +
                                "Grid2DSquareCellDoubleChunkAbstract "
                                + ( loadedChunkCount + 1 ) + " out of " + nChunks );
                        if ( ! grid2DSquareCellsSwapped ) {
                            swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                            grid2DSquareCellsSwapped = true;
                        }
                        System.out.println(
                                "swapToFileGrid2DSquareCellChunks( \n" +
                                "cri0 ( " + cri0+ " ), \n" +
                                "cci0 ( " + cci0 + " ), \n" +
                                "cri1 ( " + chunkRowIndex + " ), \n" +
                                "cci1 ( " + cci1 + " ) );");
                        swapToFileGrid2DSquareCellChunks(
                                cri0,
                                cci0,
                                chunkRowIndex,
                                cci1 );
                        cri0 = chunkRowIndex;
                        cci0 = chunkColIndex;
                        initMemoryReserve();
                    }
                } while ( ! isLoadedChunk );
                isLoadedChunk = false;
                loadedChunkCount ++;
                cci1 = chunkColIndex;
            }
            System.out.println("Done chunkRow " + chunkRowIndex );
        }
        this.grid2DSquareCells.add( this );
    }
    
    /**
     * Initialises this.
     * @see initGrid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      Grid2DSquareCellAbstract,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long,
     *      long,
     *      long,
     *      double,
     *      HashSet,
     *      boolean );
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which this is
     *   to be constructed.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param startRowIndex The grid2DSquareCell row index which is the bottom
     *   most row of this.
     * @param startColIndex The grid2DSquareCell column index which is the left
     *   most column of this.
     * @param endRowIndex The grid2DSquareCell row index which is the top most
     *   row of this.
     * @param endColIndex The grid2DSquareCell column index which is the right
     *   most column of this.
     * @param noDataValue The noDataValue for this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        try {
            initGrid2DSquareCellDouble(
                    gridStatistics,
                    directory,
                    grid2DSquareCell,
                    grid2DSquareCellDoubleChunkFactory,
                    chunkNrows,
                    chunkNcols,
                    startRowIndex,
                    startColIndex,
                    endRowIndex,
                    endColIndex,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCells = grid2DSquareCells;
                //swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0L ) {
                    initGrid2DSquareCellDouble(
                            gridStatistics,
                            directory,
                            grid2DSquareCell,
                            grid2DSquareCellDoubleChunkFactory,
                            chunkNrows,
                            chunkNcols,
                            startRowIndex,
                            startColIndex,
                            endRowIndex,
                            endColIndex,
                            noDataValue,
                            grid2DSquareCells,
                            handleOutOfMemoryError );
                } else {
                    throw oome0;
                }
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.
     * @see initGrid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      Grid2DSquareCellAbstract,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long,
     *      long,
     *      long,
     *      double,
     *      HashSet,
     *      boolean );
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which this is
     *   to be constructed.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The number of rows of cells in any chunk.
     * @param chunkNcols The number of columns of cells in any chunk.
     * @param startRowIndex The grid2DSquareCell row index which is the bottom
     *   most row of this.
     * @param startColIndex The grid2DSquareCell column index which is the left
     *   most column of this.
     * @param endRowIndex The grid2DSquareCell row index which is the top most
     *   row of this.
     * @param endColIndex The grid2DSquareCell column index which is the right
     *   most column of this.
     * @param noDataValue The noDataValue for this.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue ) {
        if ( this.grid2DSquareCells == null ) {
            this.grid2DSquareCells = new HashSet();
            initMemoryReserve();
        }
        if ( this.grid2DSquareCells.size() > 0 ) {
            Iterator ite = this.grid2DSquareCells.iterator();
            this.memoryReserve = ( ( Grid2DSquareCellAbstract ) ite.next() ).memoryReserve;
        } else {
            initMemoryReserve();
        }
        this.directory = directory;
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.nrows = endRowIndex - startRowIndex + 1L;
        this.ncols = endColIndex - startColIndex + 1L;
        initNoDataValue( noDataValue );
        //this.grid2DSquareCells = grid2DSquareCells;
        this.name = directory.getName();
        BigDecimal[] grid2DSquareCellDimensions = grid2DSquareCell.dimensions;
        this.dimensions = new BigDecimal[ grid2DSquareCellDimensions.length ];
        initNChunkRows();
        initNChunkCols();
        long nChunks = getNChunks();
        this.grid2DSquareCellChunks = new HashMap( ( int ) nChunks );
        this.dimensions[ 0 ] = new BigDecimal( grid2DSquareCellDimensions[ 0 ].toString() );
        BigDecimal startColIndexBigDecimal = new BigDecimal( ( long ) startColIndex );
        BigDecimal startRowIndexBigDecimal = new BigDecimal( ( long ) startRowIndex );
        BigDecimal nrowsBigDecimal = new BigDecimal( Long.toString( nrows ) );
        BigDecimal ncolsBigDecimal = new BigDecimal( Long.toString( ncols ) );
        this.dimensions[ 1 ] = grid2DSquareCellDimensions[ 1 ].add( startColIndexBigDecimal.multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 2 ] = grid2DSquareCellDimensions[ 2 ].add( startRowIndexBigDecimal.multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 3 ] = this.dimensions[ 1 ].add( ncolsBigDecimal.multiply( this.dimensions[ 0 ] ) );
        this.dimensions[ 4 ] = this.dimensions[ 2 ].add( nrowsBigDecimal.multiply( this.dimensions[ 0 ] ) );
        this.gridStatistics = gridStatistics;
        this.gridStatistics.init( this );
        int chunkRowIndex = Integer.MIN_VALUE;
        int chunkColIndex = Integer.MIN_VALUE;
        int loadedChunkCount = 0;
        boolean isLoadedChunk = false;
        int chunkRow = Integer.MIN_VALUE;
        int chunkCol = Integer.MIN_VALUE;
        long row = Long.MIN_VALUE;
        long col = Long.MIN_VALUE;
        boolean grid2DSquareCellsSwapped = false;
        int cri0 = 0;
        int cci0 = 0;
        int cci1 = 0;
        double cellDouble = Double.NEGATIVE_INFINITY;
        ChunkID chunkID = new ChunkID();
        Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                this,
                chunkID );
        if ( gridStatistics.getClass() == GridStatistics0.class ) {
            for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                    do {
                        try {
                            // Try to load chunk.
                            chunkID = new ChunkID(
                                    this.nChunkCols,
                                    chunkRowIndex,
                                    chunkColIndex );
                            grid2DSquareCellDoubleChunk = grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                                    this,
                                    chunkID );
                            chunkID = new ChunkID(
                                    this.nChunkCols,
                                    chunkRowIndex,
                                    chunkColIndex );
                            this.grid2DSquareCellChunks.put(
                                    chunkID,
                                    grid2DSquareCellDoubleChunk );
                            row = ( ( long ) chunkRowIndex * ( long ) this.chunkNrows );
                            for ( chunkRow = 0; chunkRow < this.chunkNrows; chunkRow ++ ) {
                                col = ( ( long ) chunkColIndex * ( long ) this.chunkNcols );
                                for ( chunkCol = 0; chunkCol < this.chunkNcols; chunkCol ++ ) {
                                    cellDouble = grid2DSquareCell.getCellDouble(
                                            row,
                                            col );
                                    initCell(
                                            row,
                                            col,
                                            cellDouble );
                                    col ++;
                                }
                                row ++;
                            }
                            isLoadedChunk = true;
                        } catch ( OutOfMemoryError oome0 ) {
                            // Encountering an OutOfMemoryError while executing this block is fatal.
                            clearMemoryReserve();
                            System.out.println(
                                    "Encountered an OutOfMemoryError initialising " +
                                    "Grid2DSquareCellDoubleChunkAbstract "
                                    + ( loadedChunkCount + 1 ) + " out of " + nChunks );
                            if ( ! grid2DSquareCellsSwapped ) {
                                swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                grid2DSquareCellsSwapped = true;
                            }
                            System.out.println(
                                    "swapToFileChunks( \n" +
                                    "cri0 ( " + cri0+ " ), \n" +
                                    "cci0 ( " + cci0 + " ), \n" +
                                    "cri1 ( " + chunkRowIndex + " ), \n" +
                                    "cci1 ( " + cci1 + " ) );" );
                            swapToFileGrid2DSquareCellChunks(
                                    cri0,
                                    cci0,
                                    chunkRowIndex,
                                    cci1 );
                            cri0 = chunkRowIndex;
                            cci0 = chunkColIndex;
                            initMemoryReserve();
                        }
                    } while ( ! isLoadedChunk );
                    isLoadedChunk = false;
                    loadedChunkCount ++;
                    cci1 = chunkColIndex;
                }
                System.out.println("Done chunkRow " + chunkRowIndex );
            }
        } else {
            // gridStatistics.getClass() == GridStatistics1.class
            // initCellFast(long,long,double) is to be used inplace of
            // initCell(long,long,double).
            for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
                for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                    do {
                        try {
                            // Try to load chunk.
                            chunkID = new ChunkID(
                                    this.nChunkCols,
                                    chunkRowIndex,
                                    chunkColIndex );
                            grid2DSquareCellDoubleChunk = grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                                    this,
                                    chunkID );
                            chunkID = new ChunkID(
                                    this.nChunkCols,
                                    chunkRowIndex,
                                    chunkColIndex );
                            this.grid2DSquareCellChunks.put(
                                    chunkID,
                                    grid2DSquareCellDoubleChunk );
                            row = ( ( long ) chunkRowIndex * ( long ) this.chunkNrows );
                            for ( chunkRow = 0; chunkRow < this.chunkNrows; chunkRow ++ ) {
                                col = ( ( long ) chunkColIndex * ( long ) this.chunkNcols );
                                for ( chunkCol = 0; chunkCol < this.chunkNcols; chunkCol ++ ) {
                                    cellDouble = grid2DSquareCell.getCellDouble(
                                            row,
                                            col );
                                    initCellFast(
                                            row,
                                            col,
                                            cellDouble );
                                    col ++;
                                }
                                row ++;
                            }
                            isLoadedChunk = true;
                        } catch ( OutOfMemoryError oome0 ) {
                            // Encountering an OutOfMemoryError while executing this block is fatal.
                            clearMemoryReserve();
                            System.out.println(
                                    "Encountered an OutOfMemoryError initialising " +
                                    "Grid2DSquareCellDoubleChunkAbstract "
                                    + ( loadedChunkCount + 1 ) + " out of " + nChunks );
                            if ( ! grid2DSquareCellsSwapped ) {
                                swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                grid2DSquareCellsSwapped = true;
                            }
                            System.out.println(
                                    "swapToFileChunks( \n" +
                                    "cri0 ( " + cri0+ " ), \n" +
                                    "cci0 ( " + cci0 + " ), \n" +
                                    "cri1 ( " + chunkRowIndex + " ), \n" +
                                    "cci1 ( " + cci1 + " ) );" );
                            swapToFileGrid2DSquareCellChunks(
                                    cri0,
                                    cci0,
                                    chunkRowIndex,
                                    cci1 );
                            cri0 = chunkRowIndex;
                            cci0 = chunkColIndex;
                            initMemoryReserve();
                        }
                    } while ( ! isLoadedChunk );
                    isLoadedChunk = false;
                    loadedChunkCount ++;
                    cci1 = chunkColIndex;
                }
                System.out.println("Done chunkRow " + chunkRowIndex );
            }
        }
        this.grid2DSquareCells.add( this );
    }
    
    /**
     * Initialises this.
     * @see this.Grid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      File,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long,
     *      long,
     *      long,
     *      double,
     *      HashSet,
     *      boolean )
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue The noDataValue for this.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            File gridFile,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells = grid2DSquareCells;
            initGrid2DSquareCellDouble(
                    gridStatistics,
                    directory,
                    gridFile,
                    grid2DSquareCellDoubleChunkFactory,
                    chunkNrows,
                    chunkNcols,
                    startRowIndex,
                    startColIndex,
                    endRowIndex,
                    endColIndex,
                    noDataValue,
                    handleOutOfMemoryError );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                this.grid2DSquareCells = grid2DSquareCells;
                //swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0L ) {
                    initGrid2DSquareCellDouble(
                            gridStatistics,
                            directory,
                            gridFile,
                            grid2DSquareCellDoubleChunkFactory,
                            chunkNrows,
                            chunkNcols,
                            startRowIndex,
                            startColIndex,
                            endRowIndex,
                            endColIndex,
                            noDataValue,
                            grid2DSquareCells,
                            handleOutOfMemoryError );
                } else {
                    throw oome0;
                }
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.
     * @see thisGrid2DSquareCellDouble(
     *      GridStatisticsAbstract,
     *      File,
     *      File,
     *      Grid2DSquareCellDoubleChunkAbstractFactory,
     *      int,
     *      int,
     *      long,
     *      long,
     *      long,
     *      long,
     *      double,
     *      HashSet,
     *      boolean )
     * @param gridStatistics The GridStatisticsAbstract to accompany this.
     * @param directory The File directory to be used for swapping.
     * @param gridFile Either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellDouble to be returned.
     * @param grid2DSquareCellDoubleChunkFactory The
     *   Grid2DSquareCellDoubleChunkAbstractFactory prefered to construct chunks
     *   of this.
     * @param chunkNrows The Grid2DSquareCellDouble chunkNrows.
     * @param chunkNcols The Grid2DSquareCellDouble chunkNcols.
     * @param startRowIndex The topmost row index of the grid stored as gridFile.
     * @param startColIndex The leftmost column index of the grid stored as gridFile.
     * @param endRowIndex The bottom row index of the grid stored as gridFile.
     * @param endColIndex The rightmost column index of the grid stored as gridFile.
     * @param noDataValue The noDataValue for this.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void initGrid2DSquareCellDouble(
            GridStatisticsAbstract gridStatistics,
            File directory,
            File gridFile,
            Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory,
            int chunkNrows,
            int chunkNcols,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            double noDataValue,
            boolean handleOutOfMemoryError ) {
        if ( this.grid2DSquareCells.size() > 0 ) {
            Iterator ite = this.grid2DSquareCells.iterator();
            this.memoryReserve = ( ( Grid2DSquareCellAbstract ) ite.next() ).memoryReserve;
        } else {
            initMemoryReserve();
        }
        // Setting directory allows for it having being moved.
        this.directory = directory;
        if ( gridFile.isDirectory() ) {
            /* Initialise from
             * new File(
             *     gridFile,
             *     "thisFile" );
             */
            if ( true ) {
                Grid2DSquareCellDoubleFactory grid2DSquareCellDoubleFactory = new Grid2DSquareCellDoubleFactory(
                        directory,
                        chunkNrows,
                        chunkNcols,
                        grid2DSquareCellDoubleChunkFactory,
                        noDataValue );
                File thisFile = new File(
                        gridFile,
                        "thisFile"  );
                ObjectInputStream ois = null;
                try {
                    ois = new ObjectInputStream(
                            new BufferedInputStream(
                            new FileInputStream( thisFile ) ) );
                    Grid2DSquareCellDouble gridFileGrid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCellDoubleFactory.create(
                            directory,
                            thisFile,
                            ois );
                    Grid2DSquareCellDouble gridFileGrid2DSquareCell = grid2DSquareCellDoubleFactory.create(
                            directory,
                            gridFileGrid2DSquareCellDouble,
                            startRowIndex,
                            startColIndex,
                            endRowIndex,
                            endColIndex,
                            grid2DSquareCells,
                            handleOutOfMemoryError );
                    initGrid2DSquareCellDouble(
                            gridFileGrid2DSquareCell,
                            false );
                } catch ( IOException ioe0 ) {
                    System.err.println( ioe0.getMessage() );
                    ioe0.printStackTrace();
                    //throw ioe0;
                }
            }
            initGrid2DSquareCellChunks( gridFile );
        } else {
            // Assume ESRI AsciiFile
            this.chunkNrows = chunkNrows;
            this.chunkNcols = chunkNcols;
            this.nrows = endRowIndex - startRowIndex + 1L;
            this.ncols = endColIndex - startColIndex + 1L;
            initNoDataValue( noDataValue );
            this.grid2DSquareCells = grid2DSquareCells;
            this.name = directory.getName();
            initNChunkRows();
            initNChunkCols();
            long nChunks = getNChunks();
            this.grid2DSquareCellChunks = new HashMap( ( int ) nChunks );
            this.dimensions = new BigDecimal[ 5 ];
            this.gridStatistics = gridStatistics;
            this.gridStatistics.init( this );
            String filename = gridFile.getName();
            int loadedChunkCount = 0;
            boolean isLoadedChunk = false;
            double value = this.noDataValue;
            boolean grid2DSquareCellsSwapped = false;
            if ( filename.endsWith( "asc" ) ) {
                ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter( gridFile );
                Object[] header = eSRIAsciigridImporter.readHeaderObject();
                //long inputNcols = ( Long ) header[ 0 ];
                //long inputNrows = ( Long ) header[ 1 ];
                this.dimensions[ 0 ] = ( BigDecimal ) header[ 4 ];
                this.dimensions[ 1 ] = ( ( BigDecimal ) header[ 2 ] ).add( dimensions[ 0 ].multiply( new BigDecimal( startColIndex ) ) );
                this.dimensions[ 2 ] = ( ( BigDecimal ) header[ 3 ] ).add( dimensions[ 0 ].multiply( new BigDecimal( startRowIndex ) ) );
                this.dimensions[ 3 ] = this.dimensions[ 1 ].add( new BigDecimal( Long.toString( this.ncols ) ).multiply( this.dimensions[ 0 ] ) );
                this.dimensions[ 4 ] = this.dimensions[ 2 ].add( new BigDecimal( Long.toString( this.nrows ) ).multiply( this.dimensions[ 0 ] ) );
                double gridFileNoDataValue = ( Double ) header[ 5 ];
                int chunkRowIndex = Integer.MIN_VALUE;
                int chunkColIndex = Integer.MIN_VALUE;
                int cachedAndClearedChunkCount = 0;
                int cri0 = 0;
                int cci0 = 0;
                int cri1 = 0;
                int cci1 = 0;
                ChunkID chunkID = new ChunkID();
                boolean initialisedChunksToClear = true;
                long row = Long.MIN_VALUE;
                long col = Long.MIN_VALUE;
                boolean isInitCellDone = false;
                //boolean cacheRowMinor = true;
                //ChunkID nextChunkToSwap = new ChunkID(
                //        this.nChunkCols,
                //        0,
                //        0 );
                Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = 
                        grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                        this,
                        chunkID );
                // Initialise Chunks
                for ( chunkRowIndex = 0; chunkRowIndex < nChunkRows; chunkRowIndex ++ ) {
                    //for ( chunkRowIndex = nChunkRows - 1; chunkRowIndex >= 0; chunkRowIndex -- ) {
                    for ( chunkColIndex = 0; chunkColIndex < nChunkCols; chunkColIndex ++ ) {
                        do {
                            try {
                                chunkID = new ChunkID(
                                        this.nChunkCols,
                                        chunkRowIndex,
                                        chunkColIndex );
                                grid2DSquareCellDoubleChunk = 
                                        grid2DSquareCellDoubleChunkFactory.createGrid2DSquareCellDoubleChunk(
                                        this,
                                        chunkID );
                                chunkID = new ChunkID(
                                        this.nChunkCols,
                                        chunkRowIndex,
                                        chunkColIndex );
                                this.grid2DSquareCellChunks.put(
                                        chunkID,
                                        grid2DSquareCellDoubleChunk );
                                isLoadedChunk = true;
                            } catch ( OutOfMemoryError oome0 ) {
                                // Encountering an OutOfMemoryError while executing this block is fatal.
                                clearMemoryReserve();
                                System.out.println(
                                        "Encountered an OutOfMemoryError initialising " +
                                    "Grid2DSquareCellDoubleChunkAbstract "
                                    + ( loadedChunkCount + 1 ) + " out of " + nChunks );
                            if ( ! grid2DSquareCellsSwapped ) {
                                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                    grid2DSquareCellsSwapped = true;
                                }
                                System.out.println(
                                        "swapToFileChunks( \n" +
                                        "cri0 ( " + cri0 + " ), \n" +
                                        "cci0 ( " + cci0 + " ), \n" +
                                        "cri1 ( " + chunkRowIndex + " ), \n" +
                                        "cci1 ( " + cci1 + " ), \n" +
                                        "handleOutOfMemoryError( " + handleOutOfMemoryError + " ) );" );
                                swapToFileGrid2DSquareCellChunks(
                                        cri0,
                                        cci0,
                                        chunkRowIndex,
                                        cci1,
                                        handleOutOfMemoryError );
                                cri0 = chunkRowIndex;
                                cci0 = chunkColIndex;
                                initMemoryReserve();
                            }
                        } while ( ! isLoadedChunk );
                        isLoadedChunk = false;
                        loadedChunkCount ++;
                        cci1 = chunkColIndex;
                    }
                    System.out.println("Done chunkRow " + chunkRowIndex );
                }
                System.out.println("First stage of initialisation complete. Reading data into initialised Chunks");
                
                // Read Data into Chunks
                if ( ( int ) gridFileNoDataValue == Integer.MIN_VALUE ) {
                    if ( gridStatistics.getClass() == GridStatistics0.class ) {
                        for ( row = ( this.nrows - 1 ); row > -1; row -- ) {
                            for ( col = 0; col < this.ncols; col ++ ) {
                                value = eSRIAsciigridImporter.readDouble();
                                do {
                                    try {
                                        initCell(
                                                row,
                                                col,
                                                value,
                                                handleOutOfMemoryError );
                                        isInitCellDone = true;
                                    } catch ( OutOfMemoryError oome0 ) {
                                        clearMemoryReserve();
                                        if ( ! grid2DSquareCellsSwapped ) {
                                            swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                            grid2DSquareCellsSwapped = true;
                                        }
                                        if ( initialisedChunksToClear ) {
                                            chunkRowIndex = getChunkRowIndex( row );
                                            if ( ( chunkRowIndex + 1 ) > cri0 ) {
                                                System.out.println(
                                                        "swapToFileChunks( \n" +
                                                        "cri0 ( " + cri0 + " ), \n" +
                                                        "cci0 ( " + cci0 + " ), \n" +
                                                        "cri1 ( " + ( this.nChunkRows - 1 ) + " ), \n" +
                                                        "cci1 ( " + cci1 + " ), \n" +
                                                        "handleOutOfMemoryError( " + handleOutOfMemoryError + " );");
                                                swapToFileGrid2DSquareCellChunks(
                                                        cri0,
                                                        cci0,
                                                        this.nChunkRows - 1,
                                                        cci1,
                                                        handleOutOfMemoryError );
                                            }
                                            initialisedChunksToClear = false;
                                            cri0 = this.nChunkRows - 1;
                                        }
                                        cri1 = getChunkRowIndex( row ) + 1;
                                        if ( cri0 == cri1 ) {
                                            System.out.println(
                                                    "Warning: This is likely to be very slow. \n" +
                                                    "Cannot completley be store chunk row " + cri0 +
                                                    "in memory. This will result in a large swapping \n" +
                                                    "overhead! It might be so slow that it is best to \n" +
                                                    "reprogram and try using a different type of \n" +
                                                    "Grid2DSquareCellDoubleChunkAbstractFactory than \n" +
                                                    grid2DSquareCellDoubleChunkFactory.toString() );
                                            boolean sufficientMemoryToContinue = false;
                                            do {
                                                try {
                                                    ChunkID swappedChunk = writeToFileGrid2DSquareCellChunkExcept(
                                                            getChunkRowIndex( row ),
                                                            getChunkColIndex( col ) );
                                                    //System.out.println( "Swapped chunk " + swappedChunk.toString() );
                                                    clearFromCacheGrid2DSquareCellChunk(
                                                            swappedChunk,
                                                            handleOutOfMemoryError );
                                                    initMemoryReserve();
                                                    sufficientMemoryToContinue = true;
                                                } catch ( OutOfMemoryError oome1 ) {
                                                    clearMemoryReserve();
                                                }
                                            } while ( ! sufficientMemoryToContinue );
                                        } else {
                                            System.out.println(
                                                    "swapToFileChunks( \n" +
                                                    "cri0 ( " + cri1 + " ), \n" +
                                                    "cci0 ( 0 ), \n" +
                                                    "cri1 ( " + cri0 + " ), \n" +
                                                    "cci1 ( 0 ), \n" +
                                                    "handleOutOfMemoryError( " + handleOutOfMemoryError + " );");
                                            swapToFileGrid2DSquareCellChunks(
                                                    cri1,
                                                    0,
                                                    cri0,
                                                    0,
                                                    handleOutOfMemoryError );
                                            cri0 = cri1 + 1;
                                        }
                                    }
                                } while ( ! isInitCellDone );
                                isInitCellDone = false;
                            }
                            if ( row % 64 == 0 ) {
                                System.out.println( "Done row " + row );
                            }
                        }
                    } else {
                        // gridStatistics.getClass() == GridStatistics1.class
                        for ( row = ( this.nrows - 1 ); row > -1; row -- ) {
                            for ( col = 0; col < this.ncols; col ++ ) {
                                value = eSRIAsciigridImporter.readDouble();
                                do {
                                    try {
                                        initCellFast(
                                                row,
                                                col,
                                                value );
                                        isInitCellDone = true;
                                    } catch ( OutOfMemoryError oome0 ) {
                                        clearMemoryReserve();
                                        if ( ! grid2DSquareCellsSwapped ) {
                                            swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                            grid2DSquareCellsSwapped = true;
                                        }
                                        if ( initialisedChunksToClear ) {
                                            chunkRowIndex = getChunkRowIndex( row );
                                            if ( ( chunkRowIndex + 1 ) > cri0 ) {
                                                System.out.println(
                                                        "swapToFileChunks( \n" +
                                                        "cri0 ( " + cri0 + " ), \n" +
                                                        "cci0 ( " + cci0 + " ), \n" +
                                                        "cri1 ( " + ( this.nChunkRows - 1 ) + " ), \n" +
                                                        "cci1 ( " + cci1 + " ), \n" +
                                                        "handleOutOfMemoryError( " + handleOutOfMemoryError + " );");
                                                swapToFileGrid2DSquareCellChunks(
                                                        cri0,
                                                        cci0,
                                                        this.nChunkRows - 1,
                                                        cci1,
                                                        handleOutOfMemoryError );
                                            }
                                            initialisedChunksToClear = false;
                                            cri0 = this.nChunkRows - 1;
                                        }
                                        cri1 = getChunkRowIndex( row ) + 1;
                                        if ( cri0 == cri1 ) {
                                            System.out.println(
                                                    "Warning: This is likely to be very slow. \n" +
                                                    "Cannot completley be store chunk row " + cri0 +
                                                    "in memory. This will result in a large swapping \n" +
                                                    "overhead! It might be so slow that it is best to \n" +
                                                    "reprogram and try using a different type of \n" +
                                                    "Grid2DSquareCellDoubleChunkAbstractFactory than \n" +
                                                    grid2DSquareCellDoubleChunkFactory.toString() );
                                            boolean sufficientMemoryToContinue = false;
                                            do {
                                                try {
                                                    ChunkID swappedChunk = writeToFileGrid2DSquareCellChunkExcept(
                                                            getChunkRowIndex( row ),
                                                            getChunkColIndex( col ) );
                                                    //System.out.println( "Swapped chunk " + swappedChunk.toString() );
                                                    clearFromCacheGrid2DSquareCellChunk(
                                                            swappedChunk,
                                                            handleOutOfMemoryError );
                                                    initMemoryReserve();
                                                    sufficientMemoryToContinue = true;
                                                } catch ( OutOfMemoryError oome1 ) {
                                                    clearMemoryReserve();
                                                }
                                            } while ( ! sufficientMemoryToContinue );
                                        } else {
                                            System.out.println(
                                                    "swapToFileChunks( \n" +
                                                    "cri0 ( " + cri1 + " ), \n" +
                                                    "cci0 ( 0 ), \n" +
                                                    "cri1 ( " + cri0 + " ), \n" +
                                                    "cci1 ( 0 ), \n" +
                                                    "handleOutOfMemoryError( " + handleOutOfMemoryError + " ) );");
                                            swapToFileGrid2DSquareCellChunks(
                                                    cri1,
                                                    0,
                                                    cri0,
                                                    0,
                                                    handleOutOfMemoryError);
                                            initMemoryReserve();
                                            cri0 = cri1 + 1;
                                        }
                                    }
                                } while ( ! isInitCellDone );
                                isInitCellDone = false;
                            }
                            if ( row % 64 == 0 ) {
                                System.out.println( "Done row " + row );
                            }
                        }
                    }
                } else {
                    if ( gridStatistics.getClass() == GridStatistics0.class ) {
                        for ( row = ( this.nrows - 1 ); row > -1; row -- ) {
                            for ( col = 0; col < this.ncols; col ++ ) {
                                value = eSRIAsciigridImporter.readDouble();
                                do {
                                    try {
                                        if ( value != gridFileNoDataValue ) {
                                            initCell(
                                                    row,
                                                    col,
                                                    value );
                                        } else {
                                            initCell(
                                                    row,
                                                    col,
                                                    this.noDataValue );
                                        }
                                        isInitCellDone = true;
                                    } catch ( OutOfMemoryError oome0 ) {
                                        clearMemoryReserve();
                                        if ( ! grid2DSquareCellsSwapped ) {
                                            swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                            grid2DSquareCellsSwapped = true;
                                        }
                                        if ( initialisedChunksToClear ) {
                                            chunkRowIndex = getChunkRowIndex( row );
                                            if ( ( chunkRowIndex + 1 ) > cri0 ) {
                                                System.out.println(
                                                        "swapToFileChunks( \n" +
                                                        "cri0 ( " + cri0 + " ), \n" +
                                                        "cci0 ( " + cci0 + " ), \n" +
                                                        "cri1 ( " + ( this.nChunkRows - 1 ) + " ), \n" +
                                                        "cci1 ( " + cci1 + " ), \n" +
                                                        "handleOutOfMemoryError( " + handleOutOfMemoryError + " );");
                                                swapToFileGrid2DSquareCellChunks(
                                                        cri0,
                                                        cci0,
                                                        this.nChunkRows - 1,
                                                        cci1,
                                                        handleOutOfMemoryError );
                                            }
                                            initialisedChunksToClear = false;
                                            cri0 = this.nChunkRows - 1;
                                        }
                                        cri1 = getChunkRowIndex( row ) + 1;
                                        if ( cri0 == cri1 ) {
                                            System.out.println(
                                                    "Warning: This is likely to be very slow. \n" +
                                                    "Cannot completley be store chunk row " + cri0 +
                                                    "in memory. This will result in a large swapping \n" +
                                                    "overhead! It might be so slow that it is best to \n" +
                                                    "reprogram and try using a different type of \n" +
                                                    "Grid2DSquareCellDoubleChunkAbstractFactory than \n" +
                                                    grid2DSquareCellDoubleChunkFactory.toString() );
                                            boolean sufficientMemoryToContinue = false;
                                            do {
                                                try {
                                                    ChunkID swappedChunk = writeToFileGrid2DSquareCellChunkExcept(
                                                            getChunkRowIndex( row ),
                                                            getChunkColIndex( col ) );
                                                    //System.out.println( "Swapped chunk " + swappedChunk.toString() );
                                                    clearFromCacheGrid2DSquareCellChunk(
                                                            swappedChunk,
                                                            handleOutOfMemoryError );
                                                    initMemoryReserve();
                                                    sufficientMemoryToContinue = true;
                                                } catch ( OutOfMemoryError oome1 ) {
                                                    clearMemoryReserve();
                                                }
                                            } while ( ! sufficientMemoryToContinue );
                                        } else {
                                            System.out.println(
                                                    "swapToFileChunks( \n" +
                                                    "cri0 ( " + cri1 + " ), \n" +
                                                    "cci0 ( 0 ), \n" +
                                                    "cri1 ( " + cri0 + " ), \n" +
                                                    "cci1 ( 0 ), \n" +
                                                    "handleOutOfMemoryError ( " + handleOutOfMemoryError + " ) );");
                                            swapToFileGrid2DSquareCellChunks(
                                                    cri1,
                                                    0,
                                                    cri0,
                                                    0,
                                                    handleOutOfMemoryError );
                                            initMemoryReserve();
                                            cri0 = cri1 + 1;
                                        }
                                    }
                                } while ( ! isInitCellDone );
                                isInitCellDone = false;
                            }
                            if ( row % 64 == 0 ) {
                                System.out.println( "Done row " + row );
                            }
                        }
                    } else {
                        // gridStatistics.getClass() == GridStatistics1.class
                        for ( row = ( this.nrows - 1 ); row > -1; row -- ) {
                            for ( col = 0; col < this.ncols; col ++ ) {
                                value = eSRIAsciigridImporter.readDouble();
                                do {
                                    try {
                                        if ( value != gridFileNoDataValue ) {
                                            initCellFast(
                                                    row,
                                                    col,
                                                    value );
                                        } else {
                                            initCellFast(
                                                    row,
                                                    col,
                                                    this.noDataValue );
                                        }
                                        isInitCellDone = true;
                                    } catch ( OutOfMemoryError oome0 ) {
                                        clearMemoryReserve();
                                        if ( ! grid2DSquareCellsSwapped ) {
                                            swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
                                            grid2DSquareCellsSwapped = true;
                                        }
                                        if ( initialisedChunksToClear ) {
                                            chunkRowIndex = getChunkRowIndex( row );
                                            if ( ( chunkRowIndex + 1 ) > cri0 ) {
                                                System.out.println(
                                                        "swapToFileChunks( \n" +
                                                        "cri0 ( " + cri0 + " ), \n" +
                                                        "cci0 ( " + cci0 + " ), \n" +
                                                        "cri1 ( " + ( this.nChunkRows - 1 ) + " ), \n" +
                                                        "cci1 ( " + cci1 + " ), \n" +
                                                        "handleOutOfMemoryError( " + handleOutOfMemoryError + " );");
                                                swapToFileGrid2DSquareCellChunks(
                                                        cri0,
                                                        cci0,
                                                        this.nChunkRows - 1,
                                                        cci1,
                                                        handleOutOfMemoryError );
                                            }
                                            initialisedChunksToClear = false;
                                            cri0 = this.nChunkRows - 1;
                                        }
                                        cri1 = getChunkRowIndex( row ) + 1;
                                        if ( cri0 == cri1 ) {
                                            System.out.println(
                                                    "Warning: This is likely to be very slow. \n" +
                                                    "Cannot completley be store chunk row " + cri0 +
                                                    "in memory. This will result in a large swapping \n" +
                                                    "overhead! It might be so slow that it is best to \n" +
                                                    "reprogram and try using a different type of \n" +
                                                    "Grid2DSquareCellDoubleChunkAbstractFactory than \n" +
                                                    grid2DSquareCellDoubleChunkFactory.toString() );
                                            boolean sufficientMemoryToContinue = false;
                                            do {
                                                try {
                                                    ChunkID swappedChunk = writeToFileGrid2DSquareCellChunkExcept(
                                                            getChunkRowIndex( row ),
                                                            getChunkColIndex( col ) );
                                                    //System.out.println( "Swapped chunk " + swappedChunk.toString() );
                                                    clearFromCacheGrid2DSquareCellChunk( swappedChunk, true );
                                                    initMemoryReserve();
                                                    sufficientMemoryToContinue = true;
                                                } catch ( OutOfMemoryError oome1 ) {
                                                    clearMemoryReserve();
                                                }
                                            } while ( ! sufficientMemoryToContinue );
                                        } else {
                                            System.out.println(
                                                    "swapToFileGrid2DSquareCellChunks( \n" +
                                                    "cri0 ( " + cri1 + " ), \n" +
                                                    "cci0 ( 0 ), \n" +
                                                    "cri1 ( " + cri0 + " ), \n" +
                                                    "cci1 ( 0 ), \n" +
                                                    "handleOutOfMemoryError ( " + handleOutOfMemoryError + " ) );");
                                            swapToFileGrid2DSquareCellChunks(
                                                    cri1,
                                                    0,
                                                    cri0,
                                                    0,
                                                    handleOutOfMemoryError );
                                            initMemoryReserve();
                                            cri0 = cri1 + 1;
                                        }
                                    }
                                } while ( ! isInitCellDone );
                                isInitCellDone = false;
                            }
                            if ( row % 64 == 0 ) {
                                System.out.println( "Done row " + row );
                            }
                        }
                    }
                }
            }
        }
        this.grid2DSquareCells.add( this );
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunks.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellDoubleChunk(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getGrid2DSquareCellDoubleChunk(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunks.
     */
    protected Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk(
            int chunkRowIndex,
            int chunkColIndex ) {
        ChunkID chunkID = new ChunkID(
                this.nChunkCols,
                chunkRowIndex,
                chunkColIndex );
        return getGrid2DSquareCellDoubleChunk( chunkID );
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunksAbstract for the given ChunkID
     * @param chunkID
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellDoubleChunk( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return getGrid2DSquareCellDoubleChunk(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns grid2DSquareCellDoubleChunksAbstract for the given ChunkID
     * @param chunkID
     */
    protected Grid2DSquareCellDoubleChunkAbstract getGrid2DSquareCellDoubleChunk(
            ChunkID chunkID ) {
        //boolean handleOutOfMemoryError = true;
        boolean containsKey = false;
        //boolean isInGrid = isInGrid( 
        //        chunkID, 
        //        handleOutOfMemoryError );
        boolean isInGrid = isInGrid( chunkID );
        if ( isInGrid ) {
            containsKey = this.grid2DSquareCellChunks.containsKey( chunkID );
            if ( ! containsKey ) {
                //loadIntoCacheChunk( 
                //        chunkID, 
                //        handleOutOfMemoryError );
                loadIntoCacheChunk( chunkID );
            }
            return ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellChunks.get( chunkID );
        }
        return null;
    }
    
    /**
     * If newValue and oldValue are the same then statistics won't change. A test might be
     * appropriate in set cell so that this method is not called. Also want to keep track if
     * underlying data has changed for getting statistics of GridStatistics1 type.
     * @param newValue The value replacing oldValue.
     * @param oldValue The value being replaced.
     * @param noDataValue this.noDataValue
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    private void upDateGridStatistics(
            double newValue,
            double oldValue,
            double noDataValue,
            boolean handleOutOfMemoryError ) {
        try {
            upDateGridStatistics(
                    newValue,
                    oldValue,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                upDateGridStatistics(
                        newValue,
                        oldValue,
                        noDataValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * If newValue and oldValue are the same then statistics won't change. A test might be
     * appropriate in set cell so that this method is not called. Also want to keep track if
     * underlying data has changed for getting statistics of GridStatistics1 type.
     * @param newValue The value replacing oldValue.
     * @param oldValue The value being replaced.
     * @param noDataValue this.noDataValue
     */
    private void upDateGridStatistics(
            double newValue,
            double oldValue,
            double noDataValue ) {
        if ( this.gridStatistics instanceof GridStatistics0 ) {
            if ( oldValue != noDataValue ) {
                BigDecimal oldValueBigDecimal = new BigDecimal( oldValue );
                this.gridStatistics.nonNoDataValueCountBigInteger = 
                        this.gridStatistics.nonNoDataValueCountBigInteger.subtract( BigInteger.ONE );
                this.gridStatistics.sumBigDecimal = 
                        this.gridStatistics.sumBigDecimal.subtract( oldValueBigDecimal );
                if ( oldValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                    this.gridStatistics.minCountBigInteger = 
                            this.gridStatistics.minCountBigInteger.subtract( BigInteger.ONE );
                }
                if ( oldValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                    this.gridStatistics.maxCountBigInteger = 
                            this.gridStatistics.maxCountBigInteger.subtract( BigInteger.ONE );
                }
            }
            if ( newValue != noDataValue ) {
                BigDecimal newValueBigDecimal = new BigDecimal( newValue );
                this.gridStatistics.nonNoDataValueCountBigInteger = 
                        this.gridStatistics.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                this.gridStatistics.sumBigDecimal = 
                        this.gridStatistics.sumBigDecimal.add( newValueBigDecimal );
                if ( newValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == -1 ) {
                    this.gridStatistics.minBigDecimal = newValueBigDecimal;
                    this.gridStatistics.minCountBigInteger = BigInteger.ONE;
                } else {
                    if ( newValueBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                        this.gridStatistics.minCountBigInteger = 
                                this.gridStatistics.minCountBigInteger.add( BigInteger.ONE );
                    } else {
                        if ( this.gridStatistics.minCountBigInteger.compareTo( BigInteger.ONE ) == -1 ) {
                            // The GridStatistics need recalculating
                            this.gridStatistics.update();
                        }
                    }
                }
                if ( newValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == -1 ) {
                    this.gridStatistics.maxBigDecimal = newValueBigDecimal;
                    this.gridStatistics.maxCountBigInteger = BigInteger.ONE;
                } else {
                    if ( newValueBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                        this.gridStatistics.maxCountBigInteger = 
                                this.gridStatistics.maxCountBigInteger.add( BigInteger.ONE );
                    } else {
                        if ( this.gridStatistics.maxCountBigInteger.compareTo( BigInteger.ONE ) == -1 ) {
                            // The GridStatistics need recalculating
                            this.gridStatistics.update();
                        }
                    }
                }
            }
        } else {
            if ( newValue != oldValue ) {
                ( ( GridStatistics1 ) this.gridStatistics ).setIsUpToDate( false );
            }
        }
    }
    
    /**
     * Returns noDataValue.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getNoDataValue(
            boolean handleOutOfMemoryError ) {
        try {
            return this.noDataValue;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNoDataValue( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the noDataValue of this as a BigDecimal.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public BigDecimal getNoDataValueBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            if ( Double.isInfinite( noDataValue ) ) {
                // Cannot convert to a BigDecimal!
                return null;
            }
            return new BigDecimal( noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNoDataValueBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises noDataValue.
     * @param noDataValue The value this.noDataValue is initialised to.
     */
    protected final void initNoDataValue(
            double noDataValue ) {
        // This method is called before this.grid2DSquareCellChunks are
        // constructed so no OutOfMemoryError handling is needed.
        if ( Double.isNaN( noDataValue ) ) {
            this.noDataValue = Double.NEGATIVE_INFINITY;
            System.out.println( "noDataValue cannot be set to NaN! Initialised noDataValue as " + this.noDataValue );
            //throw new Exception( "noDataValue cannot be set to NaN" );
        } else {
            this.noDataValue = noDataValue;
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex else returns noDataValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCell(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCell(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCell(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex else returns noDataValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     */
    protected double getCell(
            long cellRowIndex,
            long cellColIndex ) {
        boolean isInGrid = isInGrid(
                cellRowIndex,
                cellColIndex );
        if ( isInGrid ) {
            int chunkRowIndex = getChunkRowIndex( cellRowIndex );
            int chunkColIndex = getChunkColIndex( cellColIndex );
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk = getGrid2DSquareCellChunk(
                    chunkRowIndex,
                    chunkColIndex );
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                return ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell(
                        ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * this.chunkNrows ) ) ),
                        ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * this.chunkNcols ) ) ),
                        noDataValue );
            } else {
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                    return ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell(
                            ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * this.chunkNrows ) ) ),
                            ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * this.chunkNcols ) ) ),
                            noDataValue );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                        return ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell(
                                ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * this.chunkNrows ) ) ),
                                ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * this.chunkNcols ) ) ),
                                noDataValue );
                    } else {
                        return ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell(
                                ( int ) ( cellRowIndex - ( ( long ) ( chunkRowIndex * this.chunkNrows ) ) ),
                                ( int ) ( cellColIndex - ( ( long ) ( chunkColIndex * this.chunkNcols ) ) ),
                                noDataValue );
                    }
                }
            }
        }
        return noDataValue;
    }
    
    /**
     * Returns the value at position given by chunk row index chunkRowIndex,
     * chunk column index chunkColIndex, chunk cell row index chunkCellRowIndex,
     * chunk cell column index chunkCellColIndex.
     * @param chunkRowIndex The chunk row index of the cell thats value is
     *   returned.
     * @param chunkColIndex The chunk column index of the cell thats value is
     *   returned.
     * @param chunkCellRowIndex The chunk cell row index of the cell thats
     *   value is returned.
     * @param chunkCellColIndex The chunk cell column index of the cell thats
     *   value is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCell(
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCell(
                    grid2DSquareCellChunk,
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCell(
                        grid2DSquareCellChunk,
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at position given by chunk row index chunkRowIndex,
     * chunk column index chunkColIndex, chunk cell row index chunkCellRowIndex,
     * chunk cell column index chunkCellColIndex.
     * @param chunkRowIndex The chunk row index of the cell thats value is
     *   returned.
     * @param chunkColIndex The chunk column index of the cell thats value is
     *   returned.
     * @param chunkCellRowIndex The chunk cell row index of the cell thats
     *   value is returned.
     * @param chunkCellColIndex The chunk cell column index of the cell thats
     *   value is returned.
     */
    public double getCell(
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
            return ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).getCell(
                    chunkCellRowIndex,
                    chunkCellColIndex,
                    noDataValue );
        } else {
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                return ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            } else {
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                    return ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).getCell(
                            chunkCellRowIndex,
                            chunkCellColIndex,
                            noDataValue );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class  ) {
                        return ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).getCell(
                                chunkCellRowIndex,
                                chunkCellColIndex,
                                noDataValue );
                    } else {
                        return noDataValue;
                    }
                }
            }
        }
    }
    
    /**
     * For returning the value of the cell containing point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCell(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getCell(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCell(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell containing point given by
     * x-coordinate x, y-coordinate y as a int.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     */
    public final double getCell(
            double x,
            double y ) {
        return getCell(
                getCellRowIndex( y ),
                getCellColIndex( x ) );
    }
    
    /**
     * For returning the value of the cell with cell ID cellID as a int.
     * @param cellID the CellID of the cell.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCell(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return getCell(
                    cellID.cellRowIndex,
                    cellID.cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellID.cellRowIndex ),
                            getChunkColIndex( cellID.cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCell(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value at x-coordinate x, y-coordinate y and setting it
     * to newValue.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param newValue .
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double setCell(
            double x,
            double y,
            double newValue,
            boolean handleOutOfMemoryError ) {
        try {
            return setCell(
                    x,
                    y,
                    newValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return setCell(
                        x,
                        y,
                        newValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value at x-coordinate x, y-coordinate y and setting it
     * to newValue.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param newValue .
     */
    public final double setCell(
            double x,
            double y,
            double newValue ) {
        return setCell(
                getCellRowIndex( x ),
                getCellColIndex( y ),
                newValue );
    }
    
    /**
     * For returning the value of the cell with cell ID cellID and setting it
     * to newValue.
     * @param cellID the CellID of the cell.
     * @param newValue .
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double setCell(
            CellID cellID,
            double newValue,
            boolean handleOutOfMemoryError ) {
        try {
            return setCell(
                    cellID.cellRowIndex,
                    cellID.cellColIndex,
                    newValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellID.cellRowIndex ),
                            getChunkColIndex( cellID.cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return setCell(
                        cellID,
                        newValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value at cellRowIndex, cellColIndex and setting it to
     * newValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     * @param newValue .
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell(
            long cellRowIndex,
            long cellColIndex,
            double newValue,
            boolean handleOutOfMemoryError ) {
        try {
            return setCell(
                    cellRowIndex,
                    cellColIndex,
                    newValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( handleOutOfMemoryError ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                initMemoryReserve();
                return setCell(
                        cellRowIndex,
                        cellColIndex,
                        newValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value at cellRowIndex, cellColIndex and setting it to
     * newValue.
     * @param cellRowIndex .
     * @param cellColIndex .
     * @param newValue .
     */
    public double setCell(
            long cellRowIndex,
            long cellColIndex,
            double newValue ) {
        int chunkRowIndex = getChunkRowIndex( cellRowIndex );
        int chunkColIndex = getChunkColIndex( cellColIndex );
        int chunkCellRowIndex = getChunkCellRowIndex( cellRowIndex );
        int chunkCellColIndex = getChunkCellColIndex( cellColIndex );
        //if ( inGrid( cellRowIndex, cellColIndex ) ) {
        Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk 
                = getGrid2DSquareCellDoubleChunk(
                chunkRowIndex,
                chunkColIndex );
        return setCell(
                grid2DSquareCellDoubleChunk,
                chunkRowIndex,
                chunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex,
                newValue );
        //} else {
        //    return getNoDataValue();
        //}
    }
    
    /**
     * For returning the value of the cell in chunk given by chunkRowIndex and
     * chunkColIndex and cell in the chunk given by chunkCellColIndex and
     * chunkCellRowIndex and setting it to newValue.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param newValue
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double newValue,
            boolean handleOutOfMemoryError ) {
        try {
            return setCell(
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex,
                    newValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( 
                            chunkID );
                }
                initMemoryReserve();
                return setCell(
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        newValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For returning the value of the cell in chunk given by chunkRowIndex and
     * chunkColIndex and cell in the chunk given by chunkCellColIndex and
     * chunkCellRowIndex and setting it to newValue.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param newValue
     */
    public double setCell(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double newValue ) {
        //if ( inGrid( chunkRowIndex, chunkColIndex, chunkCellRowIndex, chunkCellColIndex ) ) {
        Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk =  getGrid2DSquareCellDoubleChunk(
                chunkRowIndex,
                chunkColIndex );
        return setCell(
                grid2DSquareCellDoubleChunk,
                chunkRowIndex,
                chunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex,
                newValue );
        //} else {
        //    return getNoDataValue();
        //}
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex and sets it to newValue.
     * @param chunkRowIndex .
     * @param chunkColIndex .
     * @param newValue .
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double setCell(
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double newValue,
            boolean handleOutOfMemoryError ) {
        try {
            return setCell(
                    grid2DSquareCellChunk,
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex,
                    newValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return setCell(
                        grid2DSquareCellChunk,
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        newValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex and sets it to newValue.
     * @param chunkRowIndex .
     * @param chunkColIndex .
     * @param newValue .
     */
    public double setCell(
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double newValue ) {
        boolean handleOutOfMemoryError = false; //?
        double result = getNoDataValue( handleOutOfMemoryError );
        if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
            result = ( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellChunk ).setCell(
                    chunkCellRowIndex,
                    chunkCellColIndex,
                    newValue,
                    result );
        } else {
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                result = ( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellChunk ).setCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        newValue,
                        result );
            } else {
                if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                    result = ( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellChunk ).setCell(
                            chunkCellRowIndex,
                            chunkCellColIndex,
                            newValue,
                            result );
                } else {
                    if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class  ) {
                        result = ( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellChunk ).setCell(
                                chunkCellRowIndex,
                                chunkCellColIndex,
                                newValue,
                                result );
                    }  else {
                        System.err.println(
                                "Error in "
                                + getClass().getName() + ".setCell(Grid2DSquareCellDoubleChunkAbstract,int,int,int,int,double) \n" +
                                "unable to handle Grid2DSquareCellChunkAbstract " + grid2DSquareCellChunk.toString() );
                        return result;
                    }
                }
            }
        }
        // Update Statistics
        upDateGridStatistics(
                newValue,
                result,
                noDataValue );
        return result;
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected void initCell(
            long cellRowIndex,
            long cellColIndex,
            double valueToInitialise,
            boolean handleOutOfMemoryError ) {
        try {
            initCell(
                    cellRowIndex,
                    cellColIndex,
                    valueToInitialise );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                initCell(
                        cellRowIndex,
                        cellColIndex,
                        valueToInitialise,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     */
    protected void initCell(
            long cellRowIndex,
            long cellColIndex,
            double valueToInitialise ) {
        boolean isInGrid = isInGrid(
                cellRowIndex,
                cellColIndex );
        if ( isInGrid ) {
            int chunkRowIndex = getChunkRowIndex( cellRowIndex );
            int chunkColIndex = getChunkColIndex( cellColIndex );
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols,
                    chunkRowIndex,
                    chunkColIndex );
            //boolean containsKey = this.grid2DSquareCellChunks.containsKey( chunkID ); // For debugging
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = 
                    getGrid2DSquareCellDoubleChunk( chunkID );
            grid2DSquareCellDoubleChunk.initCell(
                    ( int ) ( cellRowIndex - ( ( long ) chunkRowIndex * ( long ) chunkNrows ) ),
                    ( int ) ( cellColIndex - ( ( long ) chunkColIndex * ( long ) chunkNcols ) ),
                    valueToInitialise );
            // Update Statistics
            if ( valueToInitialise != noDataValue ) {
                BigDecimal cellBigDecimal = new BigDecimal( valueToInitialise );
                this.gridStatistics.nonNoDataValueCountBigInteger = 
                        this.gridStatistics.nonNoDataValueCountBigInteger.add( BigInteger.ONE );
                this.gridStatistics.sumBigDecimal = this.gridStatistics.sumBigDecimal.add( cellBigDecimal );
                if ( cellBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == -1 ) {
                    this.gridStatistics.minCountBigInteger = BigInteger.ONE;
                    this.gridStatistics.minBigDecimal = cellBigDecimal;
                } else {
                    if ( cellBigDecimal.compareTo( this.gridStatistics.minBigDecimal ) == 0 ) {
                        this.gridStatistics.minCountBigInteger = 
                                this.gridStatistics.minCountBigInteger.add( BigInteger.ONE );
                    }
                }
                if ( cellBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 1 ) {
                    this.gridStatistics.maxCountBigInteger = BigInteger.ONE;
                    this.gridStatistics.maxBigDecimal = cellBigDecimal;
                } else {
                    if ( cellBigDecimal.compareTo( this.gridStatistics.maxBigDecimal ) == 0 ) {
                        this.gridStatistics.maxCountBigInteger = 
                                this.gridStatistics.maxCountBigInteger.add( BigInteger.ONE );
                    }
                }
            }
        }
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex and does nothing about
     * this.gridStatistics
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected void initCellFast(
            long cellRowIndex,
            long cellColIndex,
            double valueToInitialise,
            boolean handleOutOfMemoryError ) {
        try {
            initCellFast(
                    cellRowIndex,
                    cellColIndex,
                    valueToInitialise );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                initCellFast(
                        cellRowIndex,
                        cellColIndex,
                        valueToInitialise,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initilises the value at cellRowIndex, cellColIndex and does nothing about
     * this.gridStatistics
     * @param cellRowIndex
     * @param cellColIndex
     * @param valueToInitialise
     */
    protected void initCellFast(
            long cellRowIndex,
            long cellColIndex,
            double valueToInitialise ) {
        boolean isInGrid = isInGrid(
                cellRowIndex,
                cellColIndex );
        if ( isInGrid ) {
            int chunkRowIndex = getChunkRowIndex( cellRowIndex );
            int chunkColIndex = getChunkColIndex( cellColIndex );
            int chunkNrows = this.chunkNrows;
            int chunkNcols = this.chunkNcols;
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols,
                    chunkRowIndex,
                    chunkColIndex );
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = 
                    getGrid2DSquareCellDoubleChunk( chunkID );
            grid2DSquareCellDoubleChunk.initCell(
                    ( int ) ( cellRowIndex - ( ( long ) chunkRowIndex * ( long ) chunkNrows ) ),
                    ( int ) ( cellColIndex - ( ( long ) chunkColIndex * ( long ) chunkNcols ) ),
                    valueToInitialise );
        }
    }
    
    /**
     * Returns a double[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double[] getCells(
            double x,
            double y,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCells(
                    x,
                    y,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCells(
                        x,
                        y,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a double[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     */
    public double[] getCells(
            double x,
            double y,
            double distance ) {
        return getCells(
                x,
                y,
                getCellRowIndex( y ),
                getCellColIndex( x ),
                distance );
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     *   values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double[] getCells(
            long cellRowIndex,
            long cellColIndex,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCells(
                    cellRowIndex,
                    cellColIndex,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCells(
                        cellRowIndex,
                        cellColIndex,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     *   circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     *   values are returned.
     */
    protected double[] getCells(
            long cellRowIndex,
            long cellColIndex,
            double distance ) {
        return getCells(
                getCellXDouble( cellColIndex ),
                getCellYDouble( cellRowIndex ),
                cellRowIndex,
                cellColIndex,
                distance );
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x The x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y The y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param cellRowIndex The row index at y.
     * @param cellColIndex The col index at x.
     * @param distance The radius of the circle for which intersected cell
     *   values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double[] getCells(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCells(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCells(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a int[] of all cell values for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x The x-coordinate of the circle centre from which cell values are
     *   returned.
     * @param y The y-coordinate of the circle centre from which cell values are
     *   returned.
     * @param cellRowIndex The row index at y.
     * @param cellColIndex The col index at x.
     * @param distance The radius of the circle for which intersected cell
     *   values are returned.
     */
    protected double[] getCells(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double distance ) {
        double[] cells;
        int cellDistance = ( int ) Math.ceil( distance / dimensions[ 0 ].doubleValue() );
        cells = new double[ ( ( 2 * cellDistance ) + 1 ) * ( ( 2 * cellDistance ) + 1 ) ];
        long row = Long.MIN_VALUE;
        long col = Long.MIN_VALUE;
        double thisX = Double.MIN_VALUE;
        double thisY = Double.MIN_VALUE;
        int count = 0;
        for ( row = cellRowIndex - cellDistance; row <= cellRowIndex + cellDistance; row ++ ) {
            thisY = getCellYDouble( cellRowIndex );
            for ( col = cellColIndex - cellDistance; col <= cellColIndex + cellDistance; col ++ ) {
                thisX = getCellXDouble( cellColIndex );
                if ( Utilities.distance( x, y, thisX, thisY ) <= distance ) {
                    cells[ count ] = getCell(
                            row,
                            col );
                    count ++;
                }
            }
        }
        // Trim cells
        System.arraycopy(
                cells,
                0,
                cells,
                0,
                count );
        return cells;
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDouble(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDouble(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDouble(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     */
    protected double getNearestValueDouble(
            double x,
            double y ) {
        double result = getCell(
                x,
                y );
        if ( result == noDataValue ) {
            result = getNearestValueDouble(
                    x,
                    y,
                    getCellRowIndex( y ),
                    getCellColIndex( x ),
                    noDataValue );
        }
        return result;
    }
    
    /**
     * Returns the average of the nearest data values to position given by row
     * index rowIndex, column index colIndex
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDouble(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDouble(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDouble(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to position given by row
     * index rowIndex, column index colIndex
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     */
    protected double getNearestValueDouble(
            long cellRowIndex,
            long cellColIndex ) {
        double result = getCell(
                cellRowIndex,
                cellColIndex );
        if ( result == noDataValue ) {
            result = getNearestValueDouble(
                    getCellXDouble( cellColIndex ),
                    getCellYDouble( cellRowIndex ),
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        }
        return result;
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected double getNearestValueDouble(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDouble(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDouble(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        noDataValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the average of the nearest data values to point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex the row index from which average of the nearest data
     *   values is returned
     * @param cellColIndex the column index from which average of the nearest data
     *   values is returned
     */
    protected double getNearestValueDouble(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue ) {
        CellID nearestCellID = getNearestCellID(
                x,
                y,
                cellRowIndex,
                cellColIndex );
        double nearestValue = getCell(
                cellRowIndex,
                cellColIndex );
        if ( nearestValue == noDataValue ) {
            // Find a value Seeking outwards from nearestCellID
            // Initialise visitedSet1
            HashSet visitedSet = new HashSet();
            HashSet visitedSet1 = new HashSet();
            visitedSet.add( nearestCellID );
            visitedSet1.add( nearestCellID );
            // Initialise toVisitSet1
            HashSet toVisitSet1 = new HashSet();
            long row = Long.MIN_VALUE;
            long col = Long.MIN_VALUE;
            CellID cellID0 = new CellID();
            boolean isInGrid = false;
            for ( row = -1; row < 2; row ++ ) {
                for ( col = -1; col < 2; col ++ ) {
                    if ( ! ( row == 0 && col == 0 ) ) {
                        isInGrid = isInGrid(
                                cellRowIndex + row,
                                cellColIndex + col );
                        if ( isInGrid ) {
                            cellID0 = new CellID(
                                    ncols,
                                    cellRowIndex + row,
                                    cellColIndex + col );
                            toVisitSet1.add( cellID0 );
                        }
                    }
                }
            }
            // Seek
            boolean foundValue = false;
            double value;
            HashSet values = new HashSet();
            HashSet visitedSet2;
            HashSet toVisitSet2;
            Iterator iterator;
            CellID cellID1;
            while ( ! foundValue ) {
                visitedSet2 = new HashSet();
                toVisitSet2 = new HashSet();
                iterator = toVisitSet1.iterator();
                while ( iterator.hasNext() ) {
                    cellID0 = ( CellID ) iterator.next();
                    visitedSet2.add( cellID0 );
                    value = getCell( cellID0, handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        foundValue = true;
                        values.add( cellID0 );
                    } else {
                        // Add neighbours to toVisitSet2
                        for ( row = -1; row < 2; row ++ ) {
                            for ( col = -1; col < 2; col ++ ) {
                                if ( ! ( row == 0 && col == 0 ) ) {
                                    isInGrid = isInGrid(
                                            cellID0.cellRowIndex + row,
                                            cellID0.cellColIndex + col );
                                    if ( isInGrid ) {
                                        cellID1 = new CellID(
                                                ncols,
                                                cellID0.cellRowIndex + row,
                                                cellID0.cellColIndex + col );
                                        toVisitSet2.add( cellID1 );
                                    }
                                }
                            }
                        }
                    }
                }
                toVisitSet2.removeAll( visitedSet1 );
                toVisitSet2.removeAll( visitedSet2 );
                visitedSet.addAll( visitedSet2 );
                visitedSet1 = visitedSet2;
                toVisitSet1 = toVisitSet2;
            }
            double distance;
            double minDistance = Integer.MAX_VALUE;
            // Go through values and find the closest
            HashSet closest = new HashSet();
            iterator = values.iterator();
            while ( iterator.hasNext() ) {
                cellID0 = ( CellID ) iterator.next();
                distance = Utilities.distance(
                        x,
                        y,
                        getCellXDouble( cellID0 ),
                        getCellYDouble( cellID0 ) );
                if ( distance < minDistance ) {
                    closest.clear();
                    closest.add( cellID0 );
                } else {
                    if ( distance == minDistance ) {
                        closest.add( cellID0 );
                    }
                }
                minDistance = Math.min(
                        minDistance,
                        distance );
            }
            // Get cellIDs that are within distance of discovered value
            CellID[] cellIDs = getCellIDs(
                    x,
                    y,
                    minDistance );
            // Go through values and find the closest
            for ( int i = 0; i < cellIDs.length; i ++ ) {
                if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                    if ( getCell( cellIDs[ i ], handleOutOfMemoryErrorTrue ) != noDataValue ) {
                        distance = Utilities.distance(
                                x,
                                y,
                                getCellXDouble( cellIDs[ i ] ),
                                getCellYDouble( cellIDs[ i ] ) );
                        if ( distance < minDistance ) {
                            closest.clear();
                            closest.add( cellIDs[ i ] );
                        } else {
                            if ( distance == minDistance ) {
                                closest.add( cellIDs[ i ] );
                            }
                        }
                        minDistance = Math.min(
                                minDistance,
                                distance );
                    }
                }
            }
            // Go through the closest and calculate the average.
            value = 0;
            iterator = closest.iterator();
            while ( iterator.hasNext() ) {
                cellID0 = ( CellID ) iterator.next();
                value += getCell( cellID0, handleOutOfMemoryErrorTrue );
            }
            nearestValue = value / ( double ) closest.size();
        }
        return nearestValue;
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to point given by x-coordinate x, y-coordinate y.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getNearestValuesCellIDs(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValuesCellIDs(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValuesCellIDs(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to point given by x-coordinate x, y-coordinate y.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     */
    protected CellID[] getNearestValuesCellIDs(
            double x,
            double y ) {
        double value = getCell(
                x,
                y );
        if ( value == noDataValue ) {
            return getNearestValuesCellIDs(
                    x,
                    y,
                    getCellRowIndex( y ),
                    getCellColIndex( x ),
                    noDataValue );
        }
        CellID[] cellIDs = new CellID[ 1 ];
        cellIDs[ 0 ] = getCellID(
                x,
                y );
        return cellIDs;
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to position given by row index rowIndex, column index colIndex.
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getNearestValuesCellIDs(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValuesCellIDs(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValuesCellIDs(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * to position given by row index rowIndex, column index colIndex.
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     */
    protected CellID[] getNearestValuesCellIDs(
            long cellRowIndex,
            long cellColIndex ) {
        double value = getCell(
                cellRowIndex,
                cellColIndex );
        if ( value == noDataValue ) {
            return getNearestValuesCellIDs(
                    getCellXDouble( cellColIndex ),
                    getCellYDouble( cellRowIndex ),
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        }
        CellID[] cellIDs = new CellID[ 1 ];
        cellIDs[ 0 ] = getCellID(
                cellRowIndex,
                cellColIndex );
        return cellIDs;
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * nearest to point with position given by: x-coordinate x, y-coordinate y;
     * and, cell row index cellRowIndex, cell column index cellColIndex.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param noDataValue the no data value of the grid
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getNearestValuesCellIDs(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValuesCellIDs(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValuesCellIDs(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        noDataValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - The CellIDs of the nearest cells with data values
     * nearest to point with position given by: x-coordinate x, y-coordinate y;
     * and, cell row index cellRowIndex, cell column index cellColIndex.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param cellRowIndex The row index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param cellColIndex The column index from which the cell IDs of the nearest
     *   cells with data values are returned.
     * @param noDataValue the no data value of the grid
     */
    protected CellID[] getNearestValuesCellIDs(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue ) {
        CellID[] nearestCellIDs = new CellID[1];
        nearestCellIDs[0] = getNearestCellID(
                x,
                y,
                cellRowIndex,
                cellColIndex );
        double nearestCellValue = getCell(
                cellRowIndex,
                cellColIndex );
        if ( nearestCellValue == noDataValue ) {
            // Find a value Seeking outwards from nearestCellID
            // Initialise visitedSet1
            HashSet visitedSet = new HashSet();
            HashSet visitedSet1 = new HashSet();
            visitedSet.add( nearestCellIDs[0] );
            visitedSet1.add( nearestCellIDs[0] );
            // Initialise toVisitSet1
            HashSet toVisitSet1 = new HashSet();
            long row;
            long col;
            boolean isInGrid;
            CellID cellID;
            for ( row = -1; row < 2; row ++ ) {
                for ( col = -1; col < 2; col ++ ) {
                    if ( ! ( row == 0 && col == 0 ) ) {
                        isInGrid = isInGrid(
                                cellRowIndex + row,
                                cellColIndex + col );
                        if ( isInGrid ) {
                            cellID = getCellID(
                                    cellRowIndex + row,
                                    cellColIndex + col );
                            toVisitSet1.add( cellID );
                        }
                    }
                }
            }
            // Seek
            boolean foundValue = false;
            double value;
            HashSet values = new HashSet();
            HashSet visitedSet2;
            HashSet toVisitSet2;
            Iterator iterator;
            while ( ! foundValue ) {
                visitedSet2 = new HashSet();
                toVisitSet2 = new HashSet();
                iterator = toVisitSet1.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    visitedSet2.add( cellID );
                    value = getCell(
                            cellID,
                            handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        foundValue = true;
                        values.add( cellID );
                    } else {
                        // Add neighbours to toVisitSet2
                        for ( row = -1; row < 2; row ++ ) {
                            for ( col = -1; col < 2; col ++ ) {
                                if ( ! ( row == 0 && col == 0 ) ) {
                                    isInGrid = isInGrid(
                                            cellID.cellRowIndex + row,
                                            cellID.cellColIndex + col );
                                    if ( isInGrid ) {
                                        cellID = getCellID(
                                                cellID.cellRowIndex + row,
                                                cellID.cellColIndex + col );
                                        toVisitSet2.add( cellID );
                                    }
                                }
                            }
                        }
                    }
                }
                toVisitSet2.removeAll( visitedSet1 );
                toVisitSet2.removeAll( visitedSet2 );
                visitedSet.addAll( visitedSet2 );
                visitedSet1 = visitedSet2;
                toVisitSet1 = toVisitSet2;
            }
            double distance;
            double minDistance = Double.MAX_VALUE;
            // Go through values and find the closest
            HashSet closest = new HashSet();
            iterator = values.iterator();
            while ( iterator.hasNext() ) {
                cellID = ( CellID ) iterator.next();
                distance = Utilities.distance(
                        x,
                        y,
                        getCellXDouble( cellID ),
                        getCellYDouble( cellID ) );
                if ( distance < minDistance ) {
                    closest.clear();
                    closest.add( cellID );
                } else {
                    if ( distance == minDistance ) {
                        closest.add( cellID );
                    }
                }
                minDistance = Math.min(
                        minDistance,
                        distance );
            }
            // Get cellIDs that are within distance of discovered value
            CellID[] cellIDs = getCellIDs(
                    x,
                    y,
                    minDistance );
            // Go through values and find the closest
            for ( int i = 0; i < cellIDs.length; i ++ ) {
                if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                    if ( getCell( cellIDs[ i ], handleOutOfMemoryErrorTrue ) != noDataValue ) {
                        distance = Utilities.distance(
                                x,
                                y,
                                getCellXDouble( cellIDs[ i ] ),
                                getCellYDouble( cellIDs[ i ] ) );
                        if ( distance < minDistance ) {
                            closest.clear();
                            closest.add( cellIDs[ i ] );
                        } else {
                            if ( distance == minDistance ) {
                                closest.add( cellIDs[ i ] );
                            }
                        }
                        minDistance = Math.min(
                                minDistance,
                                distance );
                    }
                }
            }
            // Go through the closest and put into an array
            nearestCellIDs = new CellID[ closest.size() ];
            iterator = closest.iterator();
            int counter = 0;
            while ( iterator.hasNext() ) {
                nearestCellIDs[ counter ] = ( CellID ) iterator.next();
                counter ++;
            }
        }
        return nearestCellIDs;
    }
    
    /**
     * Returns the distance to the nearest data value from point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDoubleDistance(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDoubleDistance(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDoubleDistance(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from point given by
     * x-coordinate x, y-coordinate y as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     */
    protected double getNearestValueDoubleDistance(
            double x,
            double y ) {
        double result = getCell(
                x,
                y );
        if ( result == noDataValue ) {
            result = getNearestValueDoubleDistance(
                    x,
                    y,
                    getCellRowIndex( y ),
                    getCellColIndex( x ),
                    noDataValue );
        }
        return result;
    }
    
    /**
     * Returns the distance to the nearest data value from position given by
     * row index rowIndex, column index colIndex as a double.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDoubleDistance(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDoubleDistance(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDoubleDistance(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from position given by
     * row index rowIndex, column index colIndex as a double.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     */
    protected double getNearestValueDoubleDistance(
            long cellRowIndex,
            long cellColIndex ) {
        double result = getCell(
                cellRowIndex,
                cellColIndex );
        if ( result == noDataValue ) {
            result = getNearestValueDoubleDistance(
                    getCellXDouble( cellColIndex ),
                    getCellYDouble( cellRowIndex ),
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        }
        return result;
    }
    
    /**
     * Returns the distance to the nearest data value from: point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getNearestValueDoubleDistance(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestValueDoubleDistance(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex,
                    noDataValue );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getNearestValueDoubleDistance(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        noDataValue,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the distance to the nearest data value from: point given by
     * x-coordinate x, y-coordinate y in position given by row index rowIndex,
     * column index colIndex as a double.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     * @param cellColIndex The cell column index of the cell from which the
     *   distance nearest to the nearest cell value is returned.
     */
    protected double getNearestValueDoubleDistance(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double noDataValue ) {
        double result = getCell(
                cellRowIndex,
                cellColIndex );
        if ( result == noDataValue ) {
            // Initialisation
            long long0 = Long.MIN_VALUE;
            long long1 = Long.MIN_VALUE;
            long longMinus1 = -1;
            long longTwo = 2;
            long longZero = 0;
            boolean boolean0 = false;
            boolean boolean1 = false;
            boolean boolean2 = false;
            double double0 = Double.NEGATIVE_INFINITY;
            double double1 = Double.NEGATIVE_INFINITY;
            CellID nearestCellID = getNearestCellID(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex );
            HashSet visitedSet = new HashSet();
            HashSet visitedSet1 = new HashSet();
            visitedSet.add( nearestCellID );
            visitedSet1.add( nearestCellID );
            HashSet toVisitSet1 = new HashSet();
            long row = Long.MIN_VALUE;
            long col = Long.MIN_VALUE;
            boolean isInGrid = false;
            CellID cellID = new CellID();
            boolean foundValue = false;
            double value = Double.NEGATIVE_INFINITY;
            HashSet values = new HashSet();
            HashSet visitedSet2 = new HashSet();
            HashSet toVisitSet2 = new HashSet();
            Iterator iterator = toVisitSet1.iterator();
            double distance = Double.NEGATIVE_INFINITY;
            double minDistance = Double.MAX_VALUE;
            HashSet closest = new HashSet();
            // Find a value Seeking outwards from nearestCellID
            // Initialise toVisitSet1
            for ( row = longMinus1; row < longTwo; row ++ ) {
                for ( col = longMinus1; col < longTwo; col ++ ) {
                    boolean0 = ( row == longZero );
                    boolean1 = ( col == longZero );
                    boolean2 = ! ( boolean0 && boolean1 );
                    if ( boolean2 ) {
                        long0 = cellRowIndex + row;
                        long1 = cellColIndex + col;
                        isInGrid = isInGrid(
                                long0,
                                long1,
                                this.handleOutOfMemoryErrorTrue );
                        if ( isInGrid ) {
                            cellID = getCellID(
                                    long0,
                                    long1,
                                    this.handleOutOfMemoryErrorTrue );
                            toVisitSet1.add( cellID );
                        }
                    }
                }
            }
            // Seek
            while ( ! foundValue ) {
                visitedSet2 = new HashSet();
                toVisitSet2 = new HashSet();
                iterator = toVisitSet1.iterator();
                while ( iterator.hasNext() ) {
                    cellID = ( CellID ) iterator.next();
                    visitedSet2.add( cellID );
                    value = getCell(
                            cellID,
                            this.handleOutOfMemoryErrorTrue );
                    if ( value != noDataValue ) {
                        foundValue = true;
                        values.add( cellID );
                    } else {
                        // Add neighbours to toVisitSet2
                        for ( row = longMinus1; row < longTwo; row ++ ) {
                            for ( col = longMinus1; col < longTwo; col ++ ) {
                                boolean0 = ( row == longZero );
                                boolean1 = ( col == longZero );
                                boolean2 = ! ( boolean0 && boolean1 );
                                if ( boolean2 ) {
                                    long0 = cellID.cellRowIndex + row;
                                    long1 = cellID.cellColIndex + col;
                                    isInGrid = isInGrid(
                                            long0,
                                            long1,
                                            this.handleOutOfMemoryErrorTrue );
                                    if ( isInGrid ) {
                                        cellID = getCellID(
                                                long0,
                                                long1,
                                                this.handleOutOfMemoryErrorTrue );
                                        toVisitSet2.add( cellID );
                                    }
                                }
                            }
                        }
                    }
                }
                toVisitSet2.removeAll( visitedSet1 );
                toVisitSet2.removeAll( visitedSet2 );
                visitedSet.addAll( visitedSet2 );
                visitedSet1 = visitedSet2;
                toVisitSet1 = toVisitSet2;
            }
            // Go through values and find the closest
            iterator = values.iterator();
            while ( iterator.hasNext() ) {
                cellID = ( CellID ) iterator.next();
                double0 = getCellXDouble(
                        cellID,
                        this.handleOutOfMemoryErrorTrue );
                double1 = getCellYDouble(
                        cellID,
                        this.handleOutOfMemoryErrorTrue );
                distance = Utilities.distance(
                        x,
                        y,
                        double0,
                        double1 );
                if ( distance < minDistance ) {
                    closest.clear();
                    closest.add( cellID );
                } else {
                    if ( distance == minDistance ) {
                        closest.add( cellID );
                    }
                }
                minDistance = Math.min(
                        minDistance,
                        distance );
            }
            // Get cellIDs that are within distance of discovered value
            CellID[] cellIDs = getCellIDs(
                    x,
                    y,
                    minDistance );
            // Go through values and find the closest
            for ( int i = 0; i < cellIDs.length; i ++ ) {
                if ( ! visitedSet.contains( cellIDs[ i ] ) ) {
                    if ( getCell( cellIDs[ i ], handleOutOfMemoryErrorTrue ) != noDataValue ) {
                        distance = Utilities.distance(
                                x,
                                y,
                                getCellXDouble( cellIDs[ i ] ),
                                getCellYDouble( cellIDs[ i ] ) );
                        minDistance = Math.min(
                                minDistance,
                                distance );
                    }
                }
            }
            result = minDistance;
        } else {
            result = 0.0d;
        }
        return result;
    }
    
    /**
     * Returns current value of cell containing the point given by x-coordinate
     * x, y-coordinate y, and adds valueToAdd to that cell.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param valueToAdd the value to be added to the cell containing the point
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double addToCell(
            double x,
            double y,
            double valueToAdd,
            boolean handleOutOfMemoryError ) {
        try {
            return addToCell(
                    x,
                    y,
                    valueToAdd );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return addToCell(
                        x,
                        y,
                        valueToAdd,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns current value of cell containing the point given by x-coordinate
     * x, y-coordinate y, and adds valueToAdd to that cell.
     * @param x the x-coordinate of the point
     * @param y the y-coordinate of the point
     * @param valueToAdd the value to be added to the cell containing the point
     */
    protected double addToCell(
            double x,
            double y,
            double valueToAdd ) {
        return addToCell(
                getCellRowIndex( y ),
                getCellColIndex( x ),
                valueToAdd );
    }
    
    /**
     * Returns the value of the cell with cell ID cellID and adds valueToAdd to
     * that cell.
     * @param cellID the CellID of the cell.
     * @param valueToAdd the value to be added to the cell containing the point
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double addToCell(
            CellID cellID,
            double valueToAdd,
            boolean handleOutOfMemoryError ) {
        try {
            return addToCell(
                    cellID,
                    valueToAdd );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellID.cellRowIndex ),
                            getChunkColIndex( cellID.cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return addToCell(
                        cellID,
                        valueToAdd,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the value of the cell with cell ID cellID and adds valueToAdd to
     * that cell.
     * @param cellID the CellID of the cell.
     * @param valueToAdd the value to be added to the cell containing the point
     */
    protected double addToCell(
            CellID cellID,
            double valueToAdd ) {
        return addToCell(
                cellID.cellRowIndex,
                cellID.cellColIndex,
                valueToAdd );
    }
    
    /**
     * Returns current value of cell with row index rowIndex and column index
     * colIndex and adds valueToAdd to that cell.
     * @param cellRowIndex the row index of the cell.
     * @param cellColIndex the col index of the cell.
     * @param valueToAdd the value to be added to the cell.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * NB1. If cell is not contained in this then then returns noDataValue.
     * NB2. Adding to noDataValue is done as if adding to a cell with value of
     *   0.
     * TODO: Check Arithmetic
     */
    public double addToCell(
            long cellRowIndex,
            long cellColIndex,
            double valueToAdd,
            boolean handleOutOfMemoryError ) {
        try {
            return addToCell(
                    cellRowIndex,
                    cellColIndex,
                    valueToAdd );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return addToCell(
                        cellRowIndex,
                        cellColIndex,
                        valueToAdd,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns current value of cell with row index rowIndex and column index
     * colIndex and adds valueToAdd to that cell.
     * @param cellRowIndex the row index of the cell.
     * @param cellColIndex the col index of the cell.
     * @param valueToAdd the value to be added to the cell.
     * NB1. If cell is not contained in this then then returns noDataValue.
     * NB2. Adding to noDataValue is done as if adding to a cell with value of
     *   0.
     * TODO: Check Arithmetic
     */
    protected double addToCell(
            long cellRowIndex,
            long cellColIndex,
            double valueToAdd ) {
        boolean isInGrid = isInGrid(
                cellRowIndex,
                cellColIndex );
        if ( isInGrid ) {
            double currentValue = getCell(
                    cellRowIndex,
                    cellColIndex );
            if ( currentValue != noDataValue ) {
                if ( valueToAdd != noDataValue ) {
                    return setCell(
                            cellRowIndex,
                            cellColIndex,
                            currentValue + valueToAdd );
                }
            } else {
                if ( valueToAdd != noDataValue ) {
                    return setCell(
                            cellRowIndex,
                            cellColIndex,
                            valueToAdd );
                }
            }
        }
        return noDataValue;
    }
    
    /**
     * Returns an iterator over the cell value in this.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Iterator iterator(
            boolean handleOutOfMemoryError ) {
        try {
            return new Grid2DSquareCellDoubleIterator( this );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return iterator( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
}
