/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.OutOfMemoryError;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import uk.ac.leeds.ccg.andyt.grids.core.OutOfMemoryErrorHandler;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLongPowersOf2;
import uk.ac.leeds.ccg.andyt.grids.utilities.Utilities;
/**
 * Contains CellID and ChunkID classes, referencing and general geometry
 * methods. It also controls what methods extended classes must implement
 * acting like an interface.
 *
 * The basic geometries are ordered in set numbers of rows and columns and are
 * arranged sequentially as their base two-dimensional orthogonal coordinate
 * axes. The sequential arrangement goes along the x-axis row by row from the
 * y-axis, then up the y-axis taking each row in turn.
 *
 * TODO:
 * Handling for NumberFormatExceptions and ArithmeticExceptions in calculations
 */
public abstract class Grid2DSquareCellAbstract
        extends OutOfMemoryErrorHandler
        implements Serializable {
    
    /**
     * A version number for confidence in reloading serialised instances.
     */
    private static final long serialVersionUID = 1L;
    
    /**
     * A HashMap of swappable objects that may be juggled to release memory for
     * processing
     */
    protected transient HashSet grid2DSquareCells;
    
    /**
     * Local directory used for caching.
     * TODO
     * This should not be transient so as it is possible to ascertain where
     * the data has come from.
     */
    protected transient File directory;
    
    /**
     * The Grid2DSquareCellChunkAbstract data cache. A collection is used
     * rather than an array because an element of a collection set to null is
     * available for garbage collection whereas an element of an array set to
     * null seems not to be.
     */
    protected transient HashMap grid2DSquareCellChunks;
    
    /**
     * A reference to the grid Statistics Object.
     */
    protected GridStatisticsAbstract gridStatistics;
    
    /**
     * For storing the number of chunk rows.
     */
    protected int nChunkRows;
    
    /**
     * For storing the number of chunk columns.
     */
    protected int nChunkCols;
    
    /**
     * For storing the (usual) number of rows of cells in a chunk. The
     * number of rows in the final chunk row may be less.
     */
    protected int chunkNrows;
    
    /**
     * For storing the (usual) number of columns of cells in a chunk. The
     * number of columns in the final chunk column may be less.
     */
    protected int chunkNcols;
    
    /**
     * For storing the number of rows in the grid.
     */
    protected long nrows;
    
    /**
     * For storing the number of columns in the grid.
     */
    protected long ncols;
    
    /**
     * For storing the name of the grid.
     */
    protected String name;
    
    /**
     * For storing cellsize, minx, miny, maxx, maxy. Although maxx and maxy
     * could be easily calculated if required, for convenience they are
     * calculated and stored by default during construction.
     */
    protected BigDecimal[] dimensions;
    
    /**
     * For storing the minimum number of decimal places used to store
     * dimensions.
     */
    protected int dimensionsScale;
    
    /**
     * For storing individual locations mapped to a binary encoded long. This
     * is only used in Grid2DSquareCellDoubleChunk64CellMap and
     * Grid2DSquareCellIntChunk64CellMap. It is stored in this to save it
     * being stored in every chunk or calculated on the fly.
     */
    protected static UnsignedLongPowersOf2 unsignedLongPowersOf2;
    
    /**
     * Initialises memoryReserve.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void initMemoryReserve(
            boolean handleOutOfMemoryError ) {
        try {
            initMemoryReserve();
        } catch ( OutOfMemoryError oome0 ) {
            swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
            initMemoryReserve( handleOutOfMemoryError );
        }
    }
    
    /* Attempts to load cache from
     * new File(
     *     gridFile,
     *     "cache" );
     * @param gridFile The directory that should may contain a File named cache.
     */
    protected void initGrid2DSquareCellChunks(
            File gridFile ) {
        try {
            File cache = new File(
                    gridFile,
                    "cache"  );
            if ( cache.exists() ) {
                try {
                    ObjectInputStream ois = new ObjectInputStream(
                            new BufferedInputStream(
                            new FileInputStream( cache ) ) );
                    this.grid2DSquareCellChunks = ( HashMap ) ois.readObject();
                    ois.close();
                } catch ( Exception e0 ) {
                    System.err.println(
                            e0.getMessage() + " in \n" +
                            "initGrid2DSquareCellChunks( \n" +
                            "    File ( " + gridFile.toString() + " ) )" );
                }
            } else {
                this.grid2DSquareCellChunks = new HashMap();
            }
        } catch ( OutOfMemoryError oome0 ) {
            System.out.println(
                    "Failed to load cache due to lack of memory" );
        }
    }
    
    /**
     * Initialises Grid2DSquareCellAbstract fields of this.
     */
    protected void initGrid2DSquareCellAbstract() {
        try {
            this.chunkNcols = 1;
            this.chunkNrows = 1;
            this.dimensions = new BigDecimal[ 5 ];
            this.dimensions[ 0 ] = new BigDecimal( 1L );
            this.dimensions[ 1 ] = new BigDecimal( 0L );
            this.dimensions[ 2 ] = new BigDecimal( 0L );
            this.dimensions[ 3 ] = new BigDecimal( chunkNcols );
            this.dimensions[ 4 ] = new BigDecimal( chunkNrows );
            this.dimensionsScale = 1;
            //this.directory = FileCreator.createNewFile();
            this.directory = new File(
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " +
                    "This could do with being a very long string! " );
            this.grid2DSquareCellChunks = new HashMap();
            this.grid2DSquareCells = new HashSet();
            this.gridStatistics = new GridStatistics0();
            this.nChunkCols = 1;
            this.ncols = 1L;
            this.nrows = 1L;
            this.name = new String(
                    "DefaultName_" + getBasicDescription() );
        } catch ( OutOfMemoryError oome0 ) {
            System.err.println(
                    "OutOfMemoryError \n" +
                    this.getClass().toString() +
                    ".initGrid2DSquareCellAbstract()" );
            throw oome0;
        }
    }
    
    /**
     * Initialises non transient Grid2DSquareCellAbstract fields from
     * grid2DSquareCell
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which the
     *   non transient Grid2DSquareCellAbstract fields of this are set.
     */
    protected void initGrid2DSquareCellAbstract(
            Grid2DSquareCellAbstract grid2DSquareCell ) {
        this.chunkNcols = grid2DSquareCell.chunkNcols;
        this.chunkNrows = grid2DSquareCell.chunkNrows;
        this.dimensions = grid2DSquareCell.dimensions;
        this.dimensionsScale = grid2DSquareCell.dimensionsScale;
        //this.directory = grid2DSquareCell.directory;
        this.gridStatistics = grid2DSquareCell.gridStatistics;
        this.name = grid2DSquareCell.name;
        this.nChunkCols = grid2DSquareCell.nChunkCols;
        this.nChunkRows = grid2DSquareCell.nChunkRows;
        this.ncols = grid2DSquareCell.ncols;
        this.nrows = grid2DSquareCell.nrows;
        this.unsignedLongPowersOf2 = grid2DSquareCell.unsignedLongPowersOf2;
    }
    
    /**
     * @return a reference to this.grid2DSquareCells
     */
    protected HashSet getGrid2DSquareCells() {
        return this.grid2DSquareCells;
    }
    
    /**
     * Sets this.grid2DSquareCells to grid2DSquareCells
     * @param grid2DSquareCells The HashSet this.grid2DSquareCells is set to.
     */
    public void setGrid2DSquareCells(
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCells = grid2DSquareCells;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                setGrid2DSquareCells(
                        grid2DSquareCells,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns this.gridStatistics
     * TODO:
     * For safety, this method should either be removed and this class be made
     * implement GridStatisticsInterface. This done the methods introduced
     * would be made to call the relevant ones in this.gridStatistics. Or
     * the gridStatistics need to be made safe in that only copies of fields are
     * passed.
     */
    public GridStatisticsAbstract getGridStatistics( ) {
        return this.gridStatistics;
    }
    
    /**
     * Returns the Grid2DSquareCellChunkAbstract with ChunkID.chunkRowIndex equal
     * to chunkRowIndex and ChunkID.chunkColIndex equal to chunkColIndex.
     * @param chunkRowIndex The ChunkID.chunkRowIndex of the returned
     *   Grid2DSquareCellChunkAbstract.
     * @param chunkColIndex The ChunkID.chunkColIndex of the returned
     *   Grid2DSquareCellChunkAbstract.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellChunk(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkID = new ChunkID(
                        this.nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return getGrid2DSquareCellChunk(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the Grid2DSquareCellChunkAbstract with ChunkID.chunkRowIndex equal
     * to chunkRowIndex and ChunkID.chunkColIndex equal to chunkColIndex.
     * @param chunkRowIndex The ChunkID.chunkRowIndex of the returned
     *   Grid2DSquareCellChunkAbstract.
     * @param chunkColIndex The ChunkID.chunkColIndex of the returned
     *   Grid2DSquareCellChunkAbstract.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk(
            int chunkRowIndex,
            int chunkColIndex ) {
        ChunkID chunkID = new ChunkID(
                this.nChunkCols,
                chunkRowIndex,
                chunkColIndex );
        return getGrid2DSquareCellChunk( chunkID );
    }
    
    /**
     * Returns the Grid2DSquareCellChunkAbstract with ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the returned Grid2DSquareCellChunkAbstract.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return getGrid2DSquareCellChunk( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return getGrid2DSquareCellChunk(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the Grid2DSquareCellChunkAbstract with ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the returned Grid2DSquareCellChunkAbstract.
     */
    protected Grid2DSquareCellChunkAbstract getGrid2DSquareCellChunk(
            ChunkID chunkID ) {
        boolean isInGrid = isInGrid( chunkID );
        if ( isInGrid ) {
            boolean handleOutOfMemoryError = true;
            boolean grid2DSquareCellChunksContainsKey =
                    this.grid2DSquareCellChunks.containsKey( chunkID );
            if ( ! grid2DSquareCellChunksContainsKey ) {
                loadIntoCacheChunk(
                        chunkID,
                        handleOutOfMemoryError );
            }
            return ( Grid2DSquareCellChunkAbstract ) this.grid2DSquareCellChunks.get(
                    chunkID );
        }
        return null;
    }
    
    /**
     * Returns an iterator over the cell values in this.
     */
    public abstract Iterator iterator(
            boolean handleOutOfMemoryError );
    
    /**
     * Returns a string description of the instance.
     */
    public abstract String toString(
            boolean handleOutOfMemoryError );
    
    /**
     * Returns a string description of the Abstract fields of this instance.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @param flag This is ignored. It is simply to distinguish this method from
     *   the abstract method toString(boolean).
     */
    public String toString(
            int flag,
            boolean handleOutOfMemoryError ) {
        try {
            return toString( flag );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return toString(
                        flag,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a string description of the Abstract fields of this instance.
     * @param flag This is ignored. It is simply to distinguish this method from
     *   the abstract method toString(boolean).
     */
    protected String toString(
            int flag ) {
        boolean handleOutOfMemoryError = true;
        String result = "chunkNcols( " + this.chunkNcols + " ), " +
                "chunkNrows( " + this.chunkNrows + " ), " +
                "dimensionsScale( " + this.dimensionsScale + " ), " +
                "nChunkCols( " + this.nChunkCols + " ), " +
                "nChunkRows( " + this.nChunkRows + " ), " +
                "ncols( " + this.ncols + " ), " +
                "nrows( " + this.nrows + " ), " +
                "directory( " + this.directory + " ), " +
                "name( " + this.name + " ), " +
                "dimensions( " + this.dimensions[ 0 ].toString();
        for ( int dimensionsID = 1; dimensionsID < this.dimensions.length; dimensionsID ++ ) {
            result += ", " + dimensions[ dimensionsID ].toString();
        }
        if ( this.memoryReserve == null ) {
            result += " ), memoryReserve( null";
        } else {
            result += " ), memoryReserve.length( " + this.memoryReserve.length;
        }
        if ( this.grid2DSquareCellChunks == null ) {
            result += " ), grid2DSquareCellChunks null";
        } else {
            result += " ), grid2DSquareCellChunks.size( " + this.grid2DSquareCellChunks.size();
        }
        //result += " ), grid2DSquareCellIntProcessor( " + this.grid2DSquareCellIntProcessor.toString();
        if ( this.unsignedLongPowersOf2 == null ) {
            result += " ), unsignedLongPowersOf2( null ), ";
        } else {
            result += " ), unsignedLongPowersOf2( " + this.unsignedLongPowersOf2.toString();
        }
        result += " ), Statistics( " + this.gridStatistics.toString( handleOutOfMemoryError );
        if ( this.grid2DSquareCells == null ) {
            result += " ), grid2DSquareCells( null )";
        } else {
            result += " ), grid2DSquareCells.size( " + this.grid2DSquareCells.size() + " )";
        }
        return result;
    }
    
    /**
     * Returns a copy of directory.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public File getDirectory(
            boolean handleOutOfMemoryError ) {
        try {
            return getDirectory();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getDirectory( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of directory.
     */
    protected File getDirectory() {
        return new File( this.directory.getPath() );
    }
    
    /**
     * Returns a copy of this.name.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String getName(
            boolean handleOutOfMemoryError ) {
        try {
            return getName();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getName( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of this.name.
     */
    protected String getName() {
        return new String( this.name );
    }
    
    /**
     * Sets this.name to name.
     * @param name The String this.name is set to.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setName(
            String name,
            boolean handleOutOfMemoryError ) {
        try {
            setName( name );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                setName(
                        name,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Sets this.name to name.
     * @param name The String this.name is set to.
     */
    protected void setName( String name ) {
        this.name = name;
    }
    
    /**
     * Returns a basic description of this instance.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public String getBasicDescription(
            boolean handleOutOfMemoryError ) {
        try {
            return getBasicDescription();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getBasicDescription( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a basic description of this instance.
     */
    protected String getBasicDescription() {
        return new String( "className(" + this.getClass().getName() +
                ")_directory(" + this.directory +
                ")_nrows(" + this.nrows +
                ")_ncols(" + this.ncols +
                ")_chunkNrows(" + this.chunkNrows +
                ")_chunkNcols(" + this.chunkNcols + ")" );
    }
    
    /**
     * Sets the Grid2DSquareCellChunkAbstract with ChunkID equal to chunkID to
     * grid2DSquareCellChunk.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract that the
     * Grid2DSquareCellChunkAbstract with ChunkID equal to chunkID is set as.
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract that is set.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public void setChunk(
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk,
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            this.grid2DSquareCellChunks.put(
                    chunkID,
                    grid2DSquareCellChunk );
            System.gc();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunksExcept( chunkID );
                initMemoryReserve();
                setChunk(
                        grid2DSquareCellChunk,
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of this.ncols.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getNcols(
            boolean handleOutOfMemoryError ) {
        try {
            return this.ncols;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNcols( handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of this.nrows.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getNrows(
            boolean handleOutOfMemoryError ) {
        try {
            return this.nrows;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNrows( handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of this.nChunkRows.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getNChunkRows(
            boolean handleOutOfMemoryError ) {
        try {
            return this.nChunkRows;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNChunkRows( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.nChunkRows.
     */
    protected final void initNChunkRows() {
        //        if ( ( this.nrows % this.chunkNrows ) > 0 ) {
        //            this.nChunkRows = ( int ) (  this.nrows /  this.chunkNrows ) + 1;
        //        } else {
        //            this.nChunkRows = ( int )  this.nrows /  this.chunkNrows;
        //        }
        long longChunkNrows = ( long ) this.chunkNrows;
        if ( ( this.nrows % longChunkNrows ) != 0 ) {
            this.nChunkRows = ( int ) ( this.nrows / longChunkNrows ) + 1;
        } else {
            this.nChunkRows = ( int ) ( this.nrows / longChunkNrows );
        }
    }
    
    /**
     * Returns a copy of this.nChunkCols.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getNChunkCols(
            boolean handleOutOfMemoryError ) {
        try {
            return this.nChunkCols;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNChunkCols( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellDoubleChunkAbstracts in this as a long.
     */
    public final long getNChunks(
            boolean handleOutOfMemoryError ) {
        try {
            return getNChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * @return the number of Grid2DSquareCellDoubleChunkAbstracts in this as a long.
     */
    protected final long getNChunks() {
        long nChunks = ( long ) this.nChunkRows * ( long ) this.nChunkCols;
        if ( nChunks > Integer.MAX_VALUE ) {
            System.err.println(
                    "Error totalChunkCount > Integer.MAX_VALUE in \n" +
                    "initGrid2DSquareCellDouble( \n" +
                    "  GridStatisticsAbstract, \n" +
                    "  File, \n" +
                    "  Grid2DSquareCellDoubleChunkAbstractFactory, \n" +
                    "  int, \n" +
                    "  int, \n" +
                    "  long, \n" +
                    "  long, \n" +
                    "  BigDecimal[], \n" +
                    "  double )" );
            throw new Error();
        }
        return nChunks;
    }
    
    /**
     * Initialises this.nChunkCols.
     */
    protected final void initNChunkCols() {
        //        if ( ( this.ncols % this.chunkNcols ) > 0 ) {
        //            this.nChunkCols = ( int ) (  this.ncols /  this.chunkNcols ) + 1;
        //        } else {
        //            this.nChunkCols = ( int )  this.ncols /  this.chunkNcols;
        //        }
        long longChunkNcols = ( long ) this.chunkNcols;
        if ( ( ncols % longChunkNcols ) != 0 ) {
            this.nChunkCols = ( int ) ( this.ncols / longChunkNcols ) + 1;
        } else {
            this.nChunkCols = ( int ) ( this.ncols / longChunkNcols );
        }
    }
    
    /**
     * Returns this.chunkNrows.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNrows(
            boolean handleOutOfMemoryError ) {
        try {
            return this.chunkNrows;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNrows(
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNrows, the number of rows in Grid2DSquareCellChunkAbstract
     * with ChunkID.chunkRowIndex equal to chunkRowIndex.
     * @param chunkRowIndex The ChunkID.chunkRowIndex of the
     *   Grid2DSquareCellChunkAbstract thats chunkNrows is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNrows(
            int chunkRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNrows( chunkRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNrows(
                        chunkRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNrows, the number of rows in Grid2DSquareCellChunkAbstract
     * with ChunkID.chunkRowIndex equal to chunkRowIndex.
     * @param chunkRowIndex The ChunkID.chunkRowIndex of the
     *   Grid2DSquareCellChunkAbstract thats chunkNrows is returned.
     */
    protected final int getChunkNrows(
            int chunkRowIndex ) {
        if ( chunkRowIndex > -1 && chunkRowIndex < this.nChunkRows ) {
            if ( chunkRowIndex == ( this.nChunkRows - 1 ) ) {
                return getChunkNrowsFinalRowChunks();
            } else {
                return this.chunkNrows;
            }
        } else {
            return 0;
        }
    }
    
    /**
     * Returns this.chunkNcols.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNcols(
            boolean handleOutOfMemoryError ) {
        try {
            return this.chunkNcols;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNcols(
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNcols, the number of columns in Grid2DSquareCellChunkAbstract
     * with ChunkID.chunkColIndex equal to chunkColIndex.
     * @param chunkColIndex The ChunkID.chunkColIndex of the
     *   Grid2DSquareCellChunkAbstract thats chunkNrows is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNcols(
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNcols( chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNcols(
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNcols, the number of columns in Grid2DSquareCellChunkAbstract
     * with ChunkID.chunkColIndex equal to chunkColIndex.
     * @param chunkColIndex The ChunkID.chunkColIndex of the
     *   Grid2DSquareCellChunkAbstract thats chunkNrows is returned.
     */
    protected final int getChunkNcols(
            int chunkColIndex ) {
        if ( chunkColIndex > -1 && chunkColIndex < this.nChunkCols ) {
            if ( chunkColIndex == ( this.nChunkCols - 1 ) ) {
                return getChunkNcolsFinalColChunks();
            } else {
                return this.chunkNcols;
            }
        } else {
            return 0;
        }
    }
    
    /**
     * Returns the number of rows in the final row Chunk.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int getChunkNrowsFinalRowChunks(
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNrowsFinalRowChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNrowsFinalRowChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the number of rows in the final row Chunk.
     */
    protected final int getChunkNrowsFinalRowChunks() {
        long longNChunkRowsMinusOne = ( long ) ( this.nChunkRows - 1 );
        long longChunkNrows = ( long ) this.chunkNrows;
        return ( int ) ( this.nrows - ( longNChunkRowsMinusOne * longChunkNrows ) );
    }
    
    /**
     * Returns the number of cols in the final col Chunk
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int getChunkNcolsFinalColChunks(
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNcolsFinalColChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNcolsFinalColChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the number of cols in the final col Chunk
     */
    protected final int getChunkNcolsFinalColChunks() {
        long longNChunkColsMinusOne = ( long ) ( this.nChunkCols - 1 );
        long longChunkNcols = ( long ) this.chunkNcols;
        return ( int ) ( this.ncols - ( longNChunkColsMinusOne * longChunkNcols ) );
    }
    
    /**
     * Returns chunkNrows, the number of rows in Grid2DSquareCellChunkAbstract with
     * ChunkID equal to chunkID
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract thats number
     *   of rows is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNrows(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNrows( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNrows(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNrows, the number of rows in Grid2DSquareCellChunkAbstract with
     * ChunkID equal to chunkID
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract thats number
     *   of rows is returned.
     */
    protected final int getChunkNrows(
            ChunkID chunkID ) {
        if ( chunkID.chunkRowIndex < ( this.nChunkRows - 1 ) ) {
            return this.chunkNrows;
        } else {
            return getChunkNrowsFinalRowChunks();
        }
    }
    
    /**
     * Returns chunkNcols, the number of columns in Grid2DSquareCellChunkAbstract with
     * ChunkID equal to chunkID
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract thats number
     *   of columns is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkNcols(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkNcols( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkNcols(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns chunkNcols, the number of columns in Grid2DSquareCellChunkAbstract with
     * ChunkID equal to chunkID
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract thats number
     *   of columns is returned.
     */
    protected final int getChunkNcols(
            ChunkID chunkID ) {
        if ( chunkID.chunkColIndex < ( this.nChunkCols - 1 ) ) {
            return this.chunkNcols;
        } else {
            return getChunkNcolsFinalColChunks();
        }
    }
    
    /**
     * Returns a copy of this.dimensions.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal[] getDimensions(
            boolean handleOutOfMemoryError ) {
        try {
            //double dimensions0 = this.dimensions[0].doubleValue(); // For debugging
            //double dimensions1 = this.dimensions[1].doubleValue(); // For debugging
            //double dimensions2 = this.dimensions[2].doubleValue(); // For debugging
            //double dimensions3 = this.dimensions[3].doubleValue(); // For debugging
            //double dimensions4 = this.dimensions[4].doubleValue(); // For debugging
            int length = dimensions.length;
            BigDecimal[] result = new BigDecimal[ length ];
            System.arraycopy( dimensions, 0, result, 0, length );
            return result;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getDimensions( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a double equal to this.dimensions[0].doubleValue.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellsizeDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return this.dimensions[ 0 ].doubleValue();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellsizeDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a copy of this.dimensionsScale.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final int getDimensionsScale(
            boolean handleOutOfMemoryError ) {
        try {
            return this.dimensionsScale;
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getDimensionsScale( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Initialises this.dimensions and this.dimensionScale.
     */
    protected final void initDimensions(
            BigDecimal[] dimensions ) {
        this.dimensions = dimensions;
        this.dimensionsScale = Integer.MIN_VALUE;
        for ( int i = 0; i < dimensions.length; i ++ ) {
            this.dimensionsScale = Math.max(
                    this.dimensionsScale,
                    dimensions[ i ].scale() );
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at at Point2D.Double point
     * as a double.
     * @param point The Point2D.Double for which the cell value is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            Point2D.Double point,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble( point );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( point.getY() ),
                            getChunkColIndex( point.getX() ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                initMemoryReserve();
                return getCellDouble(
                        point,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at at Point2D.Double point
     * as a double.
     * @param point The Point2D.Double for which the cell value is returned.
     */
    protected double getCellDouble(
            Point2D.Double point ) {
        return getCellDouble(
                getChunkRowIndex( point.getY() ),
                getChunkColIndex( point.getX() ),
                getChunkCellRowIndex( point.getY() ),
                getChunkCellColIndex( point.getX() ) );
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at at point given by
     * x-coordinate x and y-coordinate y as a double.
     * @param x The x coordinate of the point at which the cell value is returned.
     * @param y The y coordinate of the point at which the cell value is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( y ),
                            getChunkColIndex( x ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                return getCellDouble(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at at point given by
     * x-coordinate x and y-coordinate y as a double.
     * @param x The x coordinate of the point at which the cell value is returned.
     * @param y The y coordinate of the point at which the cell value is returned.
     */
    protected double getCellDouble(
            double x,
            double y ) {
        return getCellDouble(
                getChunkRowIndex( y ),
                getChunkColIndex( x ),
                getChunkCellRowIndex( y ),
                getChunkCellColIndex( x ) );
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a double.
     * @param cellRowIndex The cell row index of the .
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                return getCellDouble(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a double.
     * @param cellRowIndex The cell row index of the .
     * @param cellColIndex The cell column index.
     */
    protected double getCellDouble(
            long cellRowIndex,
            long cellColIndex ) {
        return getCellDouble(
                getChunkRowIndex( cellRowIndex ),
                getChunkColIndex( cellColIndex ),
                getChunkCellRowIndex( cellRowIndex ),
                getChunkCellColIndex( cellColIndex ) );
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a double.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble(
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                return getCellDouble(
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a double.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected double getCellDouble(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        Grid2DSquareCellChunkAbstract grid2DSquareCellChunk =
                getGrid2DSquareCellChunk(
                chunkRowIndex,
                chunkColIndex );
        return getCellDouble(
                grid2DSquareCellChunk,
                chunkRowIndex,
                chunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex );
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a double.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public double getCellDouble(
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellDouble(
                    grid2DSquareCellChunk,
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                return getCellDouble(
                        grid2DSquareCellChunk,
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a double.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected double getCellDouble(
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( grid2DSquareCellChunk instanceof Grid2DSquareCellDoubleChunkAbstract ) {
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk =
                    ( Grid2DSquareCellDoubleChunkAbstract ) grid2DSquareCellChunk;
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            if ( grid2DSquareCellChunk.getClass() ==
                    Grid2DSquareCellDoubleChunk64CellMap.class ) {
                Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                        ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunk.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                        ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                        ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                        ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF =
                        ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellDoubleChunk;
                return grid2DSquareCellDoubleChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return noDataValue;
        } else {
            //( grid2DSquareCellChunk instanceof Grid2DSquareCellIntChunkAbstract )
            Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk =
                    ( Grid2DSquareCellIntChunkAbstract ) grid2DSquareCellChunk;
            Grid2DSquareCellInt grid2DSquareCellInt =
                    grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
            int noDataValue = grid2DSquareCellInt.getNoDataValue(
                    this.handleOutOfMemoryErrorFalse );
            if ( grid2DSquareCellChunk.getClass() ==
                    Grid2DSquareCellIntChunk64CellMap.class ) {
                Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                        ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunk.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                        ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                        ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                        ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                Grid2DSquareCellIntChunkRAF grid2DSquareCellIntChunkRAF =
                        ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellIntChunk;
                return ( double ) grid2DSquareCellIntChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return ( double ) noDataValue;
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a int.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellInt(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            getChunkRowIndex( cellRowIndex ),
                            getChunkColIndex( cellColIndex ) );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCellInt(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns Grid2DSquareCellChunkAbstract cell value at cell row index equal to
     * cellRowIndex, cell col index equal to cellColIndex as a int.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     */
    protected int getCellInt(
            long cellRowIndex,
            long cellColIndex ) {
        return getCellInt(
                getChunkRowIndex( cellRowIndex ),
                getChunkColIndex( cellColIndex ),
                getChunkCellRowIndex( cellRowIndex ),
                getChunkCellColIndex( cellColIndex ) );
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a int.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellInt(
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCellInt(
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a int.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected int getCellInt(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        Grid2DSquareCellChunkAbstract grid2DSquareCellChunk = getGrid2DSquareCellChunk(
                chunkRowIndex,
                chunkColIndex );
        return getCellInt(
                grid2DSquareCellChunk,
                chunkRowIndex,
                chunkColIndex,
                chunkCellRowIndex,
                chunkCellColIndex );
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a int.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public int getCellInt(
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellInt(
                    grid2DSquareCellChunk,
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                }
                initMemoryReserve();
                return getCellInt(
                        grid2DSquareCellChunk,
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex, handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell value at chunk cell row index chunkCellRowIndex, chunk cell
     * col index chunkCellColIndex of Grid2DSquareCellChunkAbstract given by chunk
     * row index chunkRowIndex, chunk col index chunkColIndex as a int.
     * @param grid2DSquareCellChunk The Grid2DSquareCellChunkAbstract containing
     *   the cell.
     * @param chunkRowIndex The cell chunk row index.
     * @param chunkColIndex The cell chunk column index.
     * @param chunkCellRowIndex The cell row index of the chunk.
     * @param chunkCellColIndex The cell column index of the chunk.
     */
    protected int getCellInt(
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk,
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        if ( grid2DSquareCellChunk instanceof Grid2DSquareCellDoubleChunkAbstract ) {
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk =
                    ( Grid2DSquareCellDoubleChunkAbstract ) grid2DSquareCellChunk;
            Grid2DSquareCellDouble grid2DSquareCellDouble =
                    grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                        ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunk64CellMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                        ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                        ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                        ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellDoubleChunkRAF.class ) {
                Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF =
                        ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellDoubleChunk;
                return ( int ) grid2DSquareCellDoubleChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return ( int ) noDataValue;
        } else {
            //if ( grid2DSquareCellChunk instanceof Grid2DSquareCellIntChunkAbstract ) {
            Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk =
                    ( Grid2DSquareCellIntChunkAbstract ) grid2DSquareCellChunk;
            Grid2DSquareCellInt grid2DSquareCellInt =
                    grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
            int noDataValue = grid2DSquareCellInt.getNoDataValue();
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                        ( Grid2DSquareCellIntChunk64CellMap ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunk64CellMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                        ( Grid2DSquareCellIntChunkArray ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkArray.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                        ( Grid2DSquareCellIntChunkJAI ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkJAI.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                        ( Grid2DSquareCellIntChunkMap ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkMap.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            if ( grid2DSquareCellChunk.getClass() == Grid2DSquareCellIntChunkRAF.class ) {
                Grid2DSquareCellIntChunkRAF grid2DSquareCellIntChunkRAF =
                        ( Grid2DSquareCellIntChunkRAF ) grid2DSquareCellIntChunk;
                return grid2DSquareCellIntChunkRAF.getCell(
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        noDataValue );
            }
            return noDataValue;
        }
    }
    
    /**
     * Returns the value at cellRowIndex, cellColIndex as a double and sets it to valueToSet.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public abstract double setCell(
            long cellRowIndex,
            long cellColIndex,
            double valueToSet,
            boolean handleOutOfMemoryError );
    
    /**
     * Returns the chunk column index for the Grid2DSquareCellChunkAbstract
     * intersecting the x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk column
     *   index returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex(
            double x,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkColIndex( x );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkColIndex(
                        x,
                        handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk column index for the Grid2DSquareCellChunkAbstract
     * intersecting the x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk column
     *   index returned.
     */
    protected final int getChunkColIndex(
            double x ) {
        return getChunkColIndex( getCellColIndex( x ) );
    }
    
    /**
     * Returns the chunk column index for the Grid2DSquareCellChunkAbstract
     * intersecting the cell column index cellColIndex.
     * @param cellColIndex
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkColIndex(
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkColIndex( cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                return getChunkColIndex(
                        cellColIndex,
                        handleOutOfMemoryError);
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk column index for the Grid2DSquareCellChunkAbstract
     * intersecting the cell column index cellColIndex.
     * @param cellColIndex
     */
    protected final int getChunkColIndex(
            long cellColIndex ) {
        return ( int ) ( cellColIndex / ( long ) this.chunkNcols );
    }
    
    /**
     * Returns the cell column Index for the cell column that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the cell column
     *   index returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex(
            double x,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellColIndex( x );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellColIndex(
                        x,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column Index for the cell column that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the cell column
     *   index returned.
     */
    protected final long getCellColIndex(
            double x ) {
        return getCellColIndex( BigDecimal.valueOf( x ) );
    }
    
    /**
     * Returns the cell column Index for the cell column that intersect the
     * x-coordinate xBigDecimal.
     * @param xBigDecimal The x-coordinate of the line intersecting the cell column
     *   index returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex(
            BigDecimal xBigDecimal,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellColIndex( xBigDecimal );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellColIndex(
                        xBigDecimal,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column Index for the cell column that intersect the
     * x-coordinate xBigDecimal.
     * @param xBigDecimal The x-coordinate of the line intersecting the cell column
     *   index returned.
     */
    protected final long getCellColIndex(
            BigDecimal xBigDecimal ) {
        int scale = Math.max(
                xBigDecimal.scale(),
                this.dimensionsScale );
        BigDecimal bigDecimal0 = xBigDecimal.subtract(
                this.dimensions[ 1 ] );
        BigDecimal bigDecimal1 = bigDecimal0.divide(
                this.dimensions[ 0 ],
                scale,
                BigDecimal.ROUND_DOWN );
        return bigDecimal1.longValue();
    }
    
    /**
     * Returns the cell column index for the cells in chunk column index 
     * chunkColIndex chunk cell column index chunkCellColIndex.
     * @param chunkColIndex
     * @param chunkCellColIndex
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellColIndex( 
            int chunkColIndex, 
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellColIndex( chunkColIndex, chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellColIndex(
                        chunkColIndex, 
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column index for the cells in chunk column index 
     * chunkColIndex chunk cell column index chunkCellColIndex.
     * @param chunkColIndex
     * @param chunkCellColIndex
     */
    protected final long getCellColIndex( 
            int chunkColIndex,
            int chunkCellColIndex ) {
        boolean handleOutOfMemoryError = true;
        return ( ( long ) chunkColIndex * ( long ) getChunkNcols( handleOutOfMemoryError ) ) + ( long ) chunkCellColIndex;
    }

    /**
     * Returns the chunk cell column Index of the cells that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk cell column
     *   index returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex(
            double x,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkCellColIndex( x );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkCellColIndex(
                        x,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell column Index of the cells that intersect the
     * x-coordinate x.
     * @param x The x-coordinate of the line intersecting the chunk cell column
     *   index returned.
     */
    protected final int getChunkCellColIndex(
            double x ) {
        return getChunkCellColIndex( getCellColIndex( x ) );
    }
    
    /**
     * Returns the chunk cell column index of the cells in the cell column index
     * cellColIndex.
     * @param cellColIndex The cell column index of the cell thats chunk cell
     *   column index is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellColIndex(
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkCellColIndex( cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkCellColIndex(
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell column index of the cells in the cell column index
     * cellColIndex.
     * @param cellColIndex The cell column index of the cell thats chunk cell
     *   column index is returned.
     */
    protected final int getChunkCellColIndex(
            long cellColIndex ) {
        return ( int ) ( cellColIndex - ( ( cellColIndex / this.chunkNcols ) * this.chunkNcols ) );
    }
    
    /**
     * Returns the chunk row index for the chunks intersecting the line given by
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk row index is
     *   returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex(
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkRowIndex( y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkRowIndex(
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk row index for the chunks intersecting the line given by
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk row index is
     *   returned.
     */
    protected final int getChunkRowIndex(
            double y ) {
        return getChunkRowIndex( getCellRowIndex( y ) );
    }
    
    /**
     * Returns the chunk row index for the chunk intersecting the cells with
     * cell row index cellRowIndex.
     * @param cellRowIndex The cell row index of the cells thats chunk row index
     *   is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkRowIndex(
            long cellRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkRowIndex( cellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkRowIndex(
                        cellRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk row index for the chunk intersecting the cells with
     * cell row index cellRowIndex.
     * @param cellRowIndex The cell row index of the cells thats chunk row index
     *   is returned.
     */
    protected final int getChunkRowIndex(
            long cellRowIndex ) {
        return ( int ) ( cellRowIndex / ( long ) this.chunkNrows );
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line thats cell row index is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex(
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellRowIndex( y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellRowIndex(
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line thats cell row index is returned.
     */
    protected final long getCellRowIndex(
            double y ) {
        return getCellRowIndex( BigDecimal.valueOf( y ) );
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the line with
     * y-coordinate yBigDecimal.
     * @param yBigDecimal The y-coordinate of the line for which the cell row index
     *   is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex(
            BigDecimal yBigDecimal,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellRowIndex( yBigDecimal );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellRowIndex(
                        yBigDecimal,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell row Index for the cells that intersect the line with
     * y-coordinate yBigDecimal.
     * @param yBigDecimal The y-coordinate of the line for which the cell row index
     *   is returned.
     */
    protected final long getCellRowIndex(
            BigDecimal yBigDecimal ) {
        int scale = Math.max(
                yBigDecimal.scale(),
                this.dimensionsScale );
        BigDecimal bigDecimal0 = yBigDecimal.subtract(
                this.dimensions[ 2 ] );
        BigDecimal bigDecimal1 = bigDecimal0.divide(
                this.dimensions[ 0 ],
                scale,
                BigDecimal.ROUND_DOWN );
        return bigDecimal1.longValue();
    }
    
    /**
     * Returns the cell row index for the cells in chunk row index 
     * chunkRowIndex chunk cell row index chunkCellRowIndex.
     * @param chunkRowIndex
     * @param chunkCellRowIndex
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final long getCellRowIndex( 
            int chunkRowIndex, 
            int chunkCellRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellRowIndex( chunkRowIndex, chunkCellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellRowIndex(
                        chunkRowIndex, 
                        chunkCellRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the cell column index for the cells in chunk column index 
     * chunkColIndex chunk cell column index chunkCellColIndex.
     * @param chunkColIndex
     * @param chunkCellColIndex
     */
    protected final long getCellRowIndex( 
            int chunkRowIndex,
            int chunkCellRowIndex ) {
        boolean handleOutOfMemoryError = true;
        return ( ( long ) chunkRowIndex * ( long ) getChunkNrows( handleOutOfMemoryError ) ) + ( long ) chunkCellRowIndex;
    }
    
    /**
     * Returns the chunk cell row Index of the cells that intersects the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk cell row index is
     *   returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex(
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkCellRowIndex( y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkCellRowIndex(
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell row Index of the cells that intersects the line with
     * y-coordinate y.
     * @param y The y-coordinate of the line for which the chunk cell row index is
     *   returned.
     */
    protected final int getChunkCellRowIndex(
            double y ) {
        return getChunkCellRowIndex( getCellRowIndex( y ) );
    }
    
    /**
     * Returns the chunk cell row index of the cells with cell row index equal to
     * cellRowIndex.
     * @param cellRowIndex The cell row index of the cells that chunk cell row
     *   index is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final int getChunkCellRowIndex(
            long cellRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getChunkCellRowIndex( cellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getChunkCellRowIndex(
                        cellRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the chunk cell row index of the cells with cell row index equal to
     * cellRowIndex.
     * @param cellRowIndex The cell row index of the cells that chunk cell row
     *   index is returned.
     */
    protected final int getChunkCellRowIndex(
            long cellRowIndex ) {
        return ( int ) ( cellRowIndex - ( ( cellRowIndex / this.chunkNrows ) * this.chunkNrows ) );
    }
    
    /**
     * Returns a new CellID of the cell given by cell row index cellRowIndex, cell
     * column index cellColIndex. A CellID is returned even if that cell would not
     * be in the grid.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ){
        try {
            return getCellID(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellID(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a new CellID of the cell given by cell row index cellRowIndex, cell
     * column index cellColIndex. A CellID is returned even if that cell would not
     * be in the grid.
     * @param cellRowIndex The cell row index.
     * @param cellColIndex The cell column index.
     */
    protected final CellID getCellID(
            long cellRowIndex,
            long cellColIndex ) {
        return new CellID( this.ncols, cellRowIndex, cellColIndex );
    }
    
    /**
     * Returns a CellID of the cell given by x-coordinate x, y-coordinate y even
     * if that cell would not be in the grid.
     * @param x The x-coordinate.
     * @param y The y-coordinate.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID getCellID(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellID(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellID(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID of the cell given by x-coordinate x, y-coordinate y even
     * if that cell would not be in the grid.
     * @param x The x-coordinate.
     * @param y The y-coordinate.
     */
    protected final CellID getCellID(
            double x,
            double y ) {
        return new CellID(
                this.ncols,
                getCellRowIndex( y ),
                getCellColIndex( x ) );
    }
    
    /**
     * A method to write this instance to Files located in the directory returned by
     * getDirectory().
     * @param swapToFileCache Iff true then this.grid2DSquareCellChunks
     *   is written to new File( getDirectory(), "cache" ).
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFile(
            boolean swapToFileCache,
            boolean handleOutOfMemoryError )
            throws IOException {
        try {
            writeToFile( swapToFileCache );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                writeToFile(
                        swapToFileCache,
                        handleOutOfMemoryError );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method to write this instance to Files located in the directory returned by
     * getDirectory().
     * @param swapToFileCache Iff true then this.grid2DSquareCellChunks
     *   is written to new File(getDirectory(),"cache").
     */
    protected void writeToFile(
            boolean swapToFileCache )
            throws IOException {
        try {
            writeToFileGrid2DSquareCellChunks();
            if ( swapToFileCache ) {
                // Write out thisCache
                File file = new File(
                        this.directory,
                        "cache" );
                ObjectOutputStream oos = new ObjectOutputStream(
                        new BufferedOutputStream(
                        new FileOutputStream( file ) ) );
                oos.writeObject( this.grid2DSquareCellChunks );
                oos.flush();
                oos.close();
            }
            // Write out this.
            File file = new File(
                    this.directory,
                    "thisFile" );
            ObjectOutputStream oos = new ObjectOutputStream(
                    new BufferedOutputStream(
                    new FileOutputStream( file ) ) );
            oos.writeObject( this );
            oos.flush();
            oos.close();
        } catch ( IOException ioe0 ) {
            System.err.println( ioe0.getMessage() );
            throw ioe0;
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks. The first
     * Grid2DSquareCellChunkAbstract attempted to be written is that with a
     * chunk row index of 0, and a chunk column index of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract which was written to file.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunk(
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunk();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks. The first
     * Grid2DSquareCellChunkAbstract attempted to be written is that with a
     * chunk row index of 0, and a chunk column index of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract which was swapped or null.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunk() {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkID;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                if ( writeToFileGrid2DSquareCellChunk( chunkID ) ) {
                    return chunkID;
                }
            }
        }
        return null;
    }
    
    /**
     * A method that attempts to write to File a seriailized version of the
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks given by;
     * chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * @return True if Grid2DSquareCellChunkAbstract on file is up to date.
     * @param chunkRowIndex The chunk row index of the
     *   Grid2DSquareCellChunkAbstract to be written.
     * @param chunkColIndex The chunk column index of the
     *   Grid2DSquareCellChunkAbstract to be written.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean writeToFileGrid2DSquareCellChunk(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunk(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunk(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of the
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks given by;
     * chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * @return True if Grid2DSquareCellChunkAbstract on file is up to date.
     * @param chunkRowIndex The chunk row index of the
     *   Grid2DSquareCellChunkAbstract to be written.
     * @param chunkColIndex The chunk column index of the
     *   Grid2DSquareCellChunkAbstract to be written.
     */
    protected final boolean writeToFileGrid2DSquareCellChunk(
            int chunkRowIndex,
            int chunkColIndex ) {
        ChunkID chunkID = new ChunkID(
                this.nChunkCols,
                chunkRowIndex,
                chunkColIndex );
        return writeToFileGrid2DSquareCellChunk( chunkID );
    }
    
    /**
     * A method that attempts to write to File a seriailized version of the
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks given by chunkID.
     * @return True if Grid2DSquareCellChunkAbstract on file is up to date.
     * @param chunkID The chunkID of the Grid2DSquareCellChunkAbstract to be written.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean writeToFileGrid2DSquareCellChunk(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunk( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunk(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of the
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks given by chunkID.
     * @return True if Grid2DSquareCellChunkAbstract on file is up to date.
     * @param chunkID The chunkID of the Grid2DSquareCellChunkAbstract to be written.
     */
    protected final boolean writeToFileGrid2DSquareCellChunk(
            ChunkID chunkID ) {
        boolean result = true;
        Object object = this.grid2DSquareCellChunks.get( chunkID );
        if ( object != null ) {
            Grid2DSquareCellChunkAbstract grid2DSquareCellChunk =
                    ( Grid2DSquareCellChunkAbstract ) object;
            if ( ! grid2DSquareCellChunk.getIsSwapUpToDate() ) {
                File file = new File(
                        this.directory,
                        chunkID.getChunkRowIndex() + "_" + chunkID.getChunkColIndex() );
                file.getParentFile().mkdirs();
                try {
                    file.createNewFile();
                    ObjectOutputStream oos = new ObjectOutputStream(
                            new BufferedOutputStream(
                            new FileOutputStream( file ) ) );
                    oos.writeObject( grid2DSquareCellChunk );
                    oos.flush();
                    oos.close();
                } catch ( IOException ioe0 ) {
                    ioe0.printStackTrace();
                }
                System.gc();
                grid2DSquareCellChunk.setIsSwapUpToDate( true );
            }
        } else {
            result = false;
        }
        return result;
    }
    
    /**
     * A method that attempts to write to File seriailized versions of all
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunks(
            boolean handleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                writeToFileGrid2DSquareCellChunks( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of all
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks.
     */
    protected final void writeToFileGrid2DSquareCellChunks() {
        Iterator iterator = this.grid2DSquareCellChunks.keySet().iterator();
        ChunkID chunkID;
        while ( iterator.hasNext() ) {
            chunkID = ( ChunkID ) iterator.next();
            if ( isInCache( chunkID ) ) {
                writeToFileGrid2DSquareCellChunk( chunkID );
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks that
     * have ChunkID equal to those in chunkIDs.
     * @param chunkIDs A HashSet containing the ChunkID of the
     *   Grid2DSquareCellChunkAbstract to be written to file.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunks(
            HashSet chunkIDs,
            boolean handleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunks( chunkIDs );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                writeToFileGrid2DSquareCellChunks(
                        chunkIDs,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks that
     * have ChunkID equal to those in chunkIDs.
     * @param chunkIDs A HashSet containing the ChunkID of the
     *   Grid2DSquareCellChunkAbstract to be written to file.
     */
    protected final void writeToFileGrid2DSquareCellChunks(
            HashSet chunkIDs ) {
        Iterator iterator = chunkIDs.iterator();
        ChunkID chunkID;
        while( iterator.hasNext() ) {
            chunkID = ( ChunkID ) iterator.next();
            if ( isInCache( chunkID ) ) {
                writeToFileGrid2DSquareCellChunk( chunkID );
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the chunk not to be written.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunksExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunksExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                writeToFileGrid2DSquareCellChunksExcept(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the chunk not to be written.
     */
    protected final void writeToFileGrid2DSquareCellChunksExcept(
            ChunkID chunkID ) {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkIDToswapToFile;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkIDToswapToFile = new ChunkID(
                        this.nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                if (  isInCache( chunkIDToswapToFile ) &&
                        ! chunkIDToswapToFile.equals( chunkID ) ) {
                    writeToFileGrid2DSquareCellChunk(
                            chunkRowIndex,
                            chunkColIndex );
                }
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with ChunkID equal to chunkID. The first chunk that is attempted
     * to be written is that with a chunk row index of 0, and a chunk column index
     * of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkID The ChunkID of the chunk not to be swapToFiled.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunkExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                        chunkID );
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunkExcept(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with ChunkID equal to chunkID. The first chunk that is attempted
     * to be written is that with a chunk row index of 0, and a chunk column index
     * of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkID The ChunkID of the chunk not to be swapToFiled.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            ChunkID chunkID ) {
        return writeToFileGrid2DSquareCellChunkExcept(
                chunkID.getChunkRowIndex(),
                chunkID.getChunkColIndex() );
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk
     * row index of 0, and a chunk column index of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunkExcept(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunkExcept(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk
     * row index of 0, and a chunk column index of 0.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExcept(
            int chunkRowIndex,
            int chunkColIndex ) {
        int cri;
        int cci;
        ChunkID chunkID = null;
        for ( cri = 0; cri < this.nChunkRows; cri ++ ) {
            for ( cci = 0; cci< this.nChunkCols; cci ++ ) {
                if ( ! ( ( cri == chunkRowIndex ) && ( cci == chunkColIndex ) ) )  {
                    chunkID = new ChunkID(
                            this.nChunkCols,
                            cri,
                            cci );
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk(
                                cri,
                                cci );
                        return chunkID;
                    }
                }
            }
        }
        return chunkID;
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of chunkNrows - 1, and a chunk column index of chunkNcols - 1.
     * This is primarily used in constructors that read a data format that
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * presents values in reverse row order.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of chunkNrows - 1, and a chunk column index of chunkNcols - 1.
     * This is primarily used in constructors that read a data format that
     * presents values in reverse row order.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkRowIndex The chunk row index.
     * @param chunkColIndex The chunk column index.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            int chunkRowIndex,
            int chunkColIndex ) {
        ChunkID chunkID = null;
        int cri;
        int cci;
        for ( cri = this.nChunkRows - 1; cri > -1; cri -- ) {
            for ( cci = this.nChunkCols - 1; cci > -1; cci -- ) {
                if ( ! ( ( cri == chunkRowIndex ) && ( cci == chunkColIndex ) ) )  {
                    chunkID = new ChunkID(
                            this.nChunkCols,
                            cri,
                            cci );
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk(
                                cri,
                                cci );
                        return chunkID;
                    }
                }
            }
        }
        return chunkID;
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of chunkNrows - 1, and a chunk column index of chunkNcols - 1. This
     * is primarily used in constructors that read a data format that presents
     * values in reverse row order.
     * @return ChunkID of the Grid2DSquareCellChunkAbstract that is written or null.
     * @param chunkID The ChunkID of the chunk not to be swapToFiled.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return writeToFileGrid2DSquareCellChunkExceptReverseOrder( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File a seriailized version of a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * that with; chunk row index chunkRowIndex, chunk column index chunkColIndex.
     * The first chunk that is attempted to be written is that with a chunk row
     * index of chunkNrows - 1, and a chunk column index of chunkNcols - 1. This
     * is primarily used in constructors that read a data format that presents
     * values in reverse row order.
     * @param chunkID The ChunkID of the chunk not to be swapToFiled.
     */
    protected final ChunkID writeToFileGrid2DSquareCellChunkExceptReverseOrder(
            ChunkID chunkID ) {
        return writeToFileGrid2DSquareCellChunkExceptReverseOrder(
                chunkID.getChunkRowIndex(),
                chunkID.getChunkColIndex() );
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * those that have ChunkID in chunkIDs.
     * @param chunkIDs A hashSet containing ChunkIDs not to be swapToFiled.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void writeToFileGrid2DSquareCellChunksExcept(
            HashSet chunkIDs,
            boolean handleOutOfMemoryError ) {
        try {
            writeToFileGrid2DSquareCellChunksExcept( chunkIDs );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunkExcept( chunkIDs );
                writeToFileGrid2DSquareCellChunksExcept(
                        chunkIDs,
                        handleOutOfMemoryError );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * A method that attempts to write to File seriailized versions of those
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCellChunks except
     * those that have ChunkID in chunkIDs.
     * @param chunkIDs A hashSet containing ChunkIDs not to be swapToFiled.
     */
    protected final void writeToFileGrid2DSquareCellChunksExcept(
            HashSet chunkIDs ) {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkIDToswapToFile;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkIDToswapToFile = new ChunkID(
                        this.nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                if ( ! ( isInCache( chunkIDToswapToFile ) &&
                        chunkIDs.contains( chunkIDToswapToFile ) ) ) {
                    writeToFileGrid2DSquareCellChunk(
                            chunkRowIndex,
                            chunkColIndex );
                }
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The ChunkID of Grid2DSquareCellChunkAbstract swapped or null.
     */
    public final ChunkID swapToFileGrid2DSquareCellChunk(
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunk();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkID = swapToFileGrid2DSquareCellChunk();
                initMemoryReserve();
                return chunkID;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @return The ChunkID of Grid2DSquareCellChunkAbstract swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunk() {
        ChunkID chunkID = writeToFileGrid2DSquareCellChunk();
        clearFromCacheGrid2DSquareCellChunk( chunkID );
        return chunkID;
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The ChunkID of Grid2DSquareCellChunkAbstract swapped or null.
     */
    public final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunkExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkIDSwapped =
                        swapToFileGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return chunkIDSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @return The ChunkID of Grid2DSquareCellChunkAbstract swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            ChunkID chunkID ) {
        int cri;
        int cci;
        ChunkID thisChunkID = null;
        for ( cri = 0; cri < this.nChunkRows; cri ++ ) {
            for ( cci = 0; cci< this.nChunkCols; cci ++ ) {
                thisChunkID = new ChunkID(
                        this.nChunkCols,
                        cri,
                        cci );
                if ( ! thisChunkID.equals( chunkID ) ) {
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( chunkID );
                        clearFromCacheGrid2DSquareCellChunk( chunkID );
                        return thisChunkID;
                    }
                }
            }
        }
        return thisChunkID;
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellChunksExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunksExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped = 0L;
                if ( swapToFileGrid2DSquareCellChunkExcept( chunkID ) != null ) {
                    nGrid2DSquareCellChunksSwapped ++;
                }
                initMemoryReserve();
                nGrid2DSquareCellChunksSwapped +=
                        swapToFileGrid2DSquareCellChunksExcept(
                        chunkID,
                        handleOutOfMemoryError );
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this except that with ChunkID chunkID.
     * @param chunkID ChunkID of the Grid2DSquareCellChunkAbstract not to be swapped.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellChunksExcept(
            ChunkID chunkID ) {
        int cri;
        int cci;
        ChunkID chunkIDToSwap = null;
        long nGrid2DSquareCellChunksSwapped = 0L;
        for ( cri = 0; cri < this.nChunkRows; cri ++ ) {
            for ( cci = 0; cci< this.nChunkCols; cci ++ ) {
                chunkIDToSwap = new ChunkID(
                        this.nChunkCols,
                        cri,
                        cci );
                if ( ! chunkID.equals( chunkIDToSwap ) ) {
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( chunkID );
                        clearFromCacheGrid2DSquareCellChunk( chunkID );
                        nGrid2DSquareCellChunksSwapped ++;
                    }
                }
            }
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The ChunkID of the Grid2DSquareCellChunkAbstract swapped or null.
     */
    public final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            HashSet chunkIDs,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunkExcept( chunkIDs );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkID =
                        swapToFileGrid2DSquareCellChunkExcept( chunkIDs );
                initMemoryReserve();
                return chunkID;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache a Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @return The ChunkID of the Grid2DSquareCellChunkAbstract swapped or null.
     */
    protected final ChunkID swapToFileGrid2DSquareCellChunkExcept(
            HashSet chunkIDs ) {
        int cri;
        int cci;
        ChunkID chunkID = null;
        for ( cri = 0; cri < this.nChunkRows; cri ++ ) {
            for ( cci = 0; cci< this.nChunkCols; cci ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        cri,
                        cci );
                if ( ! chunkIDs.contains( chunkID ) ) {
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( chunkID );
                        clearFromCacheGrid2DSquareCellChunk( chunkID );
                        return chunkID;
                    }
                }
            }
        }
        return chunkID;
    }
    
    
    /**
     * Attempts to write to file and clear from the cache all Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunks(
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellsGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped =
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                nGrid2DSquareCellChunksSwapped +=
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunks(
                        handleOutOfMemoryError );
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache all Grid2DSquareCellChunkAbstract
     * in this.grid2DSquareCells.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunks() {
        Iterator iterator = this.grid2DSquareCells.iterator();
        long nGrid2DSquareCellChunksSwapped = 0;
        Grid2DSquareCellAbstract grid2DSquareCell;
        while ( iterator.hasNext() ) {
            grid2DSquareCell = ( Grid2DSquareCellAbstract ) iterator.next();
            nGrid2DSquareCellChunksSwapped +=
                    grid2DSquareCell.swapToFileGrid2DSquareCellChunks();
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Attempts to swap to file and clear from the cache the Grid2DSquareCellChunkAbstract in
     * this.grid2DSquareCells that do not have a ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract in this.grid2DSquareCells
     *   not to swap and clear from cache.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunksExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellsGrid2DSquareCellChunksExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped =
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                nGrid2DSquareCellChunksSwapped +=
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunksExcept(
                        chunkID,
                        handleOutOfMemoryError );
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to swap to file and clear from the cache the Grid2DSquareCellChunkAbstract in
     * this.grid2DSquareCells that do not have a ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract in this.grid2DSquareCells
     *   not to swap and clear from cache.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunksExcept(
            ChunkID chunkID ) {
        Iterator grid2DSquareCellsIterator = this.grid2DSquareCells.iterator();
        long nGrid2DSquareCellChunksSwapped = 0L;
        Grid2DSquareCellAbstract grid2DSquareCell;
        while ( grid2DSquareCellsIterator.hasNext() ) {
            grid2DSquareCell =
                    ( Grid2DSquareCellAbstract ) grid2DSquareCellsIterator.next();
            nGrid2DSquareCellChunksSwapped +=
                    grid2DSquareCell.swapToFileGrid2DSquareCellChunksExcept( chunkID );
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Attempts to swap to file and clear from the cache a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCells.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunk(
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped =
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                /* if ( nGrid2DSquareCellChunksSwapped == 0L )
                 * initMemoryReserve(); should result in OutOfMemoryError.
                 */
                initMemoryReserve();
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to swap to file and clear from the cache a
     * Grid2DSquareCellChunkAbstract in this.grid2DSquareCells.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() {
        Iterator iterator = this.grid2DSquareCells.iterator();
        Grid2DSquareCellAbstract grid2DSquareCell;
        while ( iterator.hasNext() ) {
            grid2DSquareCell = ( Grid2DSquareCellAbstract ) iterator.next();
            if ( grid2DSquareCell.swapToFileGrid2DSquareCellChunk() != null ) {
                return 1L;
            }
        }
        return 0L;
    }
    
    
    /**
     * Attempts to write to file and clear from the cache (swap) a Grid2DSquareCellChunkAbstract
     * in each Grid2DSquareCellAbstract of this.grid2DSquareCells. All
     * Grid2DSquareCellChunkAbstract written have equal ChunkID that are not equal to chunkID.
     * @param chunkID The ID of Chunks not to swap and clear from cache.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped =
                        swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to write to file and clear from the cache (swap) a Grid2DSquareCellChunkAbstract
     * in each Grid2DSquareCellAbstract of this.grid2DSquareCells. All
     * Grid2DSquareCellChunkAbstract written have equal ChunkID that are not equal to chunkID.
     * @param chunkID The ID of Chunks not to swap and clear from cache.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
            ChunkID chunkID ) {
        Iterator grid2DSquareCellsIterator = this.grid2DSquareCells.iterator();
        Grid2DSquareCellAbstract grid2DSquareCell =
                ( Grid2DSquareCellAbstract ) grid2DSquareCellsIterator.next();
        long nGrid2DSquareCellChunksSwapped = 0L;
        ChunkID chunkSwapped =
                grid2DSquareCell.writeToFileGrid2DSquareCellChunkExcept( chunkID );
        if ( chunkSwapped != null ) {
            grid2DSquareCell.clearFromCacheGrid2DSquareCellChunk( chunkSwapped );
            nGrid2DSquareCellChunksSwapped ++;
            while ( grid2DSquareCellsIterator.hasNext() ) {
                grid2DSquareCell =
                        ( Grid2DSquareCellAbstract ) grid2DSquareCellsIterator.next();
                grid2DSquareCell.writeToFileGrid2DSquareCellChunk( chunkSwapped );
                grid2DSquareCell.clearFromCacheGrid2DSquareCellChunk( chunkSwapped );
                nGrid2DSquareCellChunksSwapped ++;
            }
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Attempts to swap seriailsed version of all grid2DSquareCellChunks. This involves
     * writing them to files and then clearing them from the cache.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellChunks(
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped = 0L;
                if ( swapToFileGrid2DSquareCellChunk() != null ) {
                    nGrid2DSquareCellChunksSwapped ++;
                }
                initMemoryReserve();
                nGrid2DSquareCellChunksSwapped +=
                        swapToFileGrid2DSquareCellChunks( handleOutOfMemoryError );
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to swap seriailsed version of all grid2DSquareCellChunks. This involves
     * writing them to files and then clearing them from the cache.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellChunks() {
        //        Using Iterator as follows risks concurrent modification exception!
        //        Iterator iterator = this.grid2DSquareCellChunks.entrySet().iterator();
        //        ChunkID chunkID;
        //        while ( iterator.hasNext() ) {
        //            chunkID = ( ChunkID ) ( ( Map.Entry ) iterator.next() ).getKey();
        //            writeToFileGrid2DSquareCellChunk( chunkID );
        //            iterator.remove();
        //            System.gc();
        //        }
        int cri;
        int cci;
        ChunkID chunkID;
        long nGrid2DSquareCellChunksSwapped = 0L;
        for ( cri = 0; cri < this.nChunkRows; cri ++ ) {
            for ( cci = 0; cci< this.nChunkCols; cci ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        cri,
                        cci );
                if ( isInCache( chunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( chunkID );
                    clearFromCacheGrid2DSquareCellChunk( chunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Attempts to swap seriailsed version of all Grid2DSquareCellChunk from
     * (cri0, cci0) to (cri1, cci1) in row major order. This involves writing
     * them to files and then clearing them from the cache.
     * @param cri0 The chunk row index of the first Grid2DSquareCellChunk to be
     *   swapped.
     * @param cci0 The chunk column index of the first Grid2DSquareCellChunk to
     *   be swapped.
     * @param cri1 The chunk row index of the last Grid2DSquareCellChunk to be
     *   swapped.
     * @param cci1 The chunk column index of the last Grid2DSquareCellChunk to be
     *   swapped.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    public final long swapToFileGrid2DSquareCellChunks(
            int cri0,
            int cci0,
            int cri1,
            int cci1,
            boolean handleOutOfMemoryError ) {
        try {
            return swapToFileGrid2DSquareCellChunks(
                    cri0,
                    cci0,
                    cri1,
                    cci1 );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                long nGrid2DSquareCellChunksSwapped = 0L;
                if ( swapToFileGrid2DSquareCellChunk() != null ) {
                    nGrid2DSquareCellChunksSwapped ++;
                }
                initMemoryReserve();
                nGrid2DSquareCellChunksSwapped += swapToFileGrid2DSquareCellChunks(
                        cri0,
                        cci0,
                        cri1,
                        cci1,
                        handleOutOfMemoryError );
                return nGrid2DSquareCellChunksSwapped;
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to swap seriailsed version of all Grid2DSquareCellChunk from
     * (cri0, cci0) to (cri1, cci1) in row major order. This involves writing
     * them to files and then clearing them from the cache.
     * @param cri0 The chunk row index of the first Grid2DSquareCellChunks to be
     *   swapped.
     * @param cci0 The chunk column index of the first Grid2DSquareCellChunks to
     *   be swapped.
     * @param cri1 The chunk row index of the last Grid2DSquareCellChunks to be
     *   swapped.
     * @param cci1 The chunk column index of the last Grid2DSquareCellChunks to
     *   be swapped.
     * @return The number of Grid2DSquareCellChunkAbstract swapped.
     */
    protected final long swapToFileGrid2DSquareCellChunks(
            int cri0,
            int cci0,
            int cri1,
            int cci1 ) {
        ChunkID chunkID = null;
        long nGrid2DSquareCellChunksSwapped = 0L;
        if ( cri0 != cri1 ) {
            for ( int cci = cci0; cci < this.nChunkCols; cci ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        cri0,
                        cci );
                if ( isInCache( chunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( chunkID );
                    clearFromCacheGrid2DSquareCellChunk( chunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
            for ( int cri = cri0 + 1; cri < cri1; cri ++ ) {
                for ( int cci = 0; cci < this.nChunkCols; cci ++ ) {
                    chunkID = new ChunkID(
                            this.nChunkCols,
                            cri,
                            cci );
                    if ( isInCache( chunkID ) ) {
                        writeToFileGrid2DSquareCellChunk( chunkID );
                        clearFromCacheGrid2DSquareCellChunk( chunkID );
                        nGrid2DSquareCellChunksSwapped ++;
                    }
                }
            }
            for ( int cci = 0; cci < cci1; cci ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        cri1,
                        cci );
                if ( isInCache( chunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( chunkID );
                    clearFromCacheGrid2DSquareCellChunk( chunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
        } else {
            for ( int cci = cci0; cci < cci1 + 1; cci ++ ) {
                chunkID = new ChunkID(
                        this.nChunkCols,
                        cri0,
                        cci );
                if ( isInCache( chunkID ) ) {
                    writeToFileGrid2DSquareCellChunk( chunkID );
                    clearFromCacheGrid2DSquareCellChunk( chunkID );
                    nGrid2DSquareCellChunksSwapped ++;
                }
            }
        }
        return nGrid2DSquareCellChunksSwapped;
    }
    
    /**
     * Returns true iff grid2DSquareCellChunk given by chunkID is in memory.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk tested to see if
     *   it is swapToFiled.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInCache(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return isInCache( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return isInCache(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff grid2DSquareCellChunk given by chunkID is swapToFiled. This
     * must be an upToDate swapToFile.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk tested to see if
     *   it is swapToFiled.
     */
    protected final boolean isInCache(
            ChunkID chunkID ) {
        return this.grid2DSquareCellChunks.containsKey( chunkID );
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of swapToFileChunk(ChunkID) has been called.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk to be cleared.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunk(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunk( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunk( chunkID );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory. This is usually only
     * done after the equivallent of swapToFileChunk(ChunkID) has been called.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunk(
            ChunkID chunkID ) {
        this.grid2DSquareCellChunks.remove( chunkID );
        System.gc();
    }
    
    /**
     * For releasing all Grid2DSquareCellChunkAbstract in
     * this.grid2DSquareCellChunks.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunks();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunks();
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For releasing all Grid2DSquareCellChunkAbstract in
     * this.grid2DSquareCellChunks.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks() {
        Iterator iterator = this.grid2DSquareCellChunks.entrySet().iterator();
        while ( iterator.hasNext() ) {
            iterator.next();
            iterator.remove();
            System.gc();
        }
    }
    
    /**
     * Clears Grid2DSquareCellChunkAbstracts with ChunkIDs in chunkIDs.
     * @param chunkIDs A HashSet containing the ChunkIDs of the
     *   Grid2DSquareCellChunkAbstracts to be cleared.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            HashSet chunkIDs,
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunks( chunkIDs );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ){
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunks( chunkIDs );
                initMemoryReserve();
            }
        }
    }
    
    /**
     * Clears Grid2DSquareCellChunkAbstracts with ChunkIDs in chunkIDs.
     * @param chunkIDs A HashSet containing the ChunkIDs of the
     *   Grid2DSquareCellChunkAbstracts to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunks(
            HashSet chunkIDs ) {
        Iterator iterator = chunkIDs.iterator();
        ChunkID chunkIDToClear;
        while( iterator.hasNext() ) {
            chunkIDToClear = ( ChunkID ) iterator.next();
            clearFromCacheGrid2DSquareCellChunk( chunkIDToClear );
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunkExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunkExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For releasing a grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunkExcept(
            ChunkID chunkID ) {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkIDToClear;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkIDToClear = new ChunkID(
                        this.nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                if ( ! chunkIDToClear.equals( chunkID ) ) {
                    clearFromCacheGrid2DSquareCellChunk( chunkIDToClear );
                    return;
                }
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunksExcept( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunksExcept( chunkID );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except that with
     * ChunkID given by chunkID.
     * @param chunkID The ChunkID of the grid2DSquareCellChunk not to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            ChunkID chunkID ) {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkIDToswapToFile;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkIDToswapToFile = new ChunkID(
                        nChunkCols,
                        chunkRowIndex,
                        chunkColIndex );
                if ( ! chunkIDToswapToFile.equals( chunkID ) ) {
                    clearFromCacheGrid2DSquareCellChunk( chunkIDToswapToFile );
                }
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except those with
     * ChunkIDs in chunkIDs.
     * @param chunkIDs The HashSet of ChunkIDs of the grid2DSquareCellChunks not
     *   to be cleared.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            HashSet chunkIDs,
            boolean handleOutOfMemoryError ) {
        try {
            clearFromCacheGrid2DSquareCellChunksExcept( chunkIDs );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                clearFromCacheGrid2DSquareCellChunksExcept( chunkIDs );
                initMemoryReserve();
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * For releasing all grid2DSquareCellChunk stored in memory except those with
     * ChunkIDs in chunkIDs.
     * @param chunkIDs The HashSet of ChunkIDs of the grid2DSquareCellChunks not
     *   to be cleared.
     */
    protected final void clearFromCacheGrid2DSquareCellChunksExcept(
            HashSet chunkIDs ) {
        int chunkRowIndex;
        int chunkColIndex;
        ChunkID chunkIDToClear;
        for ( chunkRowIndex = 0; chunkRowIndex < this.nChunkRows; chunkRowIndex ++ ) {
            for ( chunkColIndex = 0; chunkColIndex < this.nChunkCols; chunkColIndex ++ ) {
                chunkIDToClear = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                if ( ! chunkIDs.contains( chunkIDToClear ) ) {
                    clearFromCacheGrid2DSquareCellChunk( chunkIDToClear );
                }
            }
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a swapped cache file.
     * @param chunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be restored.
     * @param chunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be restored.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void loadIntoCacheChunk(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            loadIntoCacheChunk(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                if ( true ) {
                    ChunkID chunkID = new ChunkID(
                            this.nChunkCols,
                            chunkRowIndex,
                            chunkColIndex );
                    swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept(
                            chunkID );
                }
                loadIntoCacheChunk(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Restores a grid2DSquareCellChunk from a swapped cache file.
     * @param chunkRowIndex The chunk row index of the grid2DSquareCellChunk
     *   to be restored.
     * @param chunkColIndex The chunk column index of the grid2DSquareCellChunk
     *   to be restored.
     */
    protected final void loadIntoCacheChunk(
            int chunkRowIndex,
            int chunkColIndex ) {
        boolean handleOutOfMemoryError = true;
        ChunkID chunkID = new ChunkID(
                this.nChunkCols,
                chunkRowIndex,
                chunkColIndex );
        loadIntoCacheChunk(
                chunkID,
                handleOutOfMemoryError );
    }
    
    /**
     * Attempts to load into the memory cache a AbstractGrid2DSquareCellChunk with ChunkID
     * @param chunkID The ChunkID of the grid2DSquareCellChunk to be restored.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final void loadIntoCacheChunk(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            loadIntoCacheChunk( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                loadIntoCacheChunk(
                        chunkID,
                        handleOutOfMemoryErrorFalse );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Attempts to load into the memory cache Grid2DSquareCellChunkAbstract with
     * ChunkID equal to chunkID.
     * @param chunkID The ChunkID of the Grid2DSquareCellChunkAbstract to be restored.
     */
    protected final void loadIntoCacheChunk(
            ChunkID chunkID ) {
        boolean handleOutOfMemoryError = true;
        boolean isInCache = isInCache(
                chunkID,
                handleOutOfMemoryError );
        if ( ! isInCache ) {
            File file = new File(
                    this.directory,
                    "" + chunkID.getChunkRowIndex() +
                    "_" + chunkID.getChunkColIndex() );
            Object object0 = loadObject(
                    file,
                    handleOutOfMemoryError );
            if ( this.getClass() == Grid2DSquareCellInt.class ) {
                if ( object0.getClass() == Grid2DSquareCellIntChunk64CellMap.class ) {
                    Grid2DSquareCellIntChunk64CellMap grid2DSquareCellIntChunk64CellMap =
                            ( Grid2DSquareCellIntChunk64CellMap ) object0;
                    grid2DSquareCellIntChunk64CellMap.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunk64CellMap.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellIntChunk64CellMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkArray.class ) {
                    Grid2DSquareCellIntChunkArray grid2DSquareCellIntChunkArray =
                            ( Grid2DSquareCellIntChunkArray ) object0;
                    grid2DSquareCellIntChunkArray.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkArray.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellIntChunkArray );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkMap.class ) {
                    Grid2DSquareCellIntChunkMap grid2DSquareCellIntChunkMap =
                            ( Grid2DSquareCellIntChunkMap ) object0;
                    grid2DSquareCellIntChunkMap.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkMap.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellIntChunkMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellIntChunkJAI.class ) {
                    Grid2DSquareCellIntChunkJAI grid2DSquareCellIntChunkJAI =
                            ( Grid2DSquareCellIntChunkJAI ) object0;
                    grid2DSquareCellIntChunkJAI.initGrid2DSquareCell( this );
                    grid2DSquareCellIntChunkJAI.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellIntChunkJAI );
                    return;
                }
                System.err.println(
                        "Unrecognised type of Grid2DSquareCellIntChunkAbstract or null " +
                        this.getClass().getName() +
                        ".loadIntoCacheChunk( ChunkID( " + chunkID.toString() + " ) )" );
            } else {
                if ( object0.getClass() == Grid2DSquareCellDoubleChunk64CellMap.class ) {
                    Grid2DSquareCellDoubleChunk64CellMap grid2DSquareCellDoubleChunk64CellMap =
                            ( Grid2DSquareCellDoubleChunk64CellMap ) object0;
                    grid2DSquareCellDoubleChunk64CellMap.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunk64CellMap.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellDoubleChunk64CellMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkArray.class ) {
                    Grid2DSquareCellDoubleChunkArray grid2DSquareCellDoubleChunkArray =
                            ( Grid2DSquareCellDoubleChunkArray ) object0;
                    grid2DSquareCellDoubleChunkArray.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkArray.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellDoubleChunkArray );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkMap.class ) {
                    Grid2DSquareCellDoubleChunkMap grid2DSquareCellDoubleChunkMap =
                            ( Grid2DSquareCellDoubleChunkMap ) object0;
                    grid2DSquareCellDoubleChunkMap.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkMap.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellDoubleChunkMap );
                    return;
                }
                if ( object0.getClass() == Grid2DSquareCellDoubleChunkJAI.class ) {
                    Grid2DSquareCellDoubleChunkJAI grid2DSquareCellDoubleChunkJAI =
                            ( Grid2DSquareCellDoubleChunkJAI ) object0;
                    grid2DSquareCellDoubleChunkJAI.initGrid2DSquareCell( this );
                    grid2DSquareCellDoubleChunkJAI.initChunkID( chunkID );
                    this.grid2DSquareCellChunks.put(
                            chunkID,
                            grid2DSquareCellDoubleChunkJAI );
                    return;
                }
                System.err.println(
                        "Unrecognised type of Grid2DSquareCellDoubleChunkAbstract or null " +
                        this.getClass().getName() +
                        ".loadIntoCacheChunk( ChunkID( " + chunkID.toString() + " ) )" );
            }
        }
    }
    
    /**
     * Loads an object from File file.
     * @param file The file from which the Object is loaded.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            File file,
            boolean handleOutOfMemoryError ) {
        try {
            return loadObject( file );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return loadObject(
                        file,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Loads an object from File file.
     * @param file The file from which the Object is loaded.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            File file ) {
        ObjectInputStream objectInputStream = null;
        try {
            objectInputStream = new ObjectInputStream(
                    new BufferedInputStream(
                    new FileInputStream( file ) ) );
        } catch ( Exception e0 ) {
            if ( objectInputStream != null ) {
                try {
                    objectInputStream.close();
                } catch ( Exception e1 ) {
                    e0.printStackTrace();
                }
            }
            //return null;
        } catch ( Error e0 ) {
            if ( objectInputStream != null ) {
                try {
                    objectInputStream.close();
                } catch ( Exception e1 ) {
                    e0.printStackTrace();
                }
            }
        }
        return loadObject( objectInputStream );
    }
    
    /**
     * Loads an object from the ObjectInputStream objectInputStream.
     * @param file The file from which the Object is loaded.
     * @return The loaded Object or null.
     */
    protected final Object loadObject(
            ObjectInputStream objectInputStream ) {
        Object result = null;
        try {
            result = objectInputStream.readObject();
            objectInputStream.close();
        } catch ( Exception e0 ) {
            e0.printStackTrace();
        }
        return result;
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     * returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     * returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are
     *     initiated, then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs(
            double x,
            double y,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellIDs(
                    x,
                    y,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellIDs(
                        x,
                        y,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids are
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     * returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     * returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     */
    protected final CellID[] getCellIDs(
            double x,
            double y,
            double distance ) {
        return getCellIDs(
                x,
                y,
                getCellRowIndex( y ),
                getCellColIndex( x ),
                distance );
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids would be
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final CellID[] getCellIDs(
            long cellRowIndex,
            long cellColIndex,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellIDs(
                    cellRowIndex,
                    cellColIndex,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellIDs(
                        cellRowIndex,
                        cellColIndex,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns a CellID[] - the cell IDs for cells thats centroids would be
     * intersected by circle with centre at centroid of cell given by cell row
     * index cellRowIndex, cell column index cellColIndex, and radius distance.
     * @param cellRowIndex the row index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param cellColIndex the column index for the cell thats centroid is the
     * circle centre from which cell values are returned.
     * @param distance the radius of the circle for which intersected cell
     * values are returned.
     */
    protected final CellID[] getCellIDs(
            long cellRowIndex,
            long cellColIndex,
            double distance ) {
        return getCellIDs(
                getCellXDouble( cellColIndex ),
                getCellYDouble( cellRowIndex ),
                cellRowIndex,
                cellColIndex,
                distance );
    }
    
    /**
     * Returns double[] cells - the values for cells thats centroids would be
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *        returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *        returned.
     * @param cellRowIndex the row index at y.
     * @param cellColIndex the col index at x.
     * @param distance the radius of the circle for which intersected cell
     *        values are returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID[] getCellIDs(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double distance,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellIDs(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex,
                    distance );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellIDs(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        distance,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] cells - the values for cells thats centroids would be
     * intersected by circle with centre at x-coordinate x, y-coordinate y, and
     * radius distance.
     * @param x the x-coordinate of the circle centre from which cell values are
     *        returned.
     * @param y the y-coordinate of the circle centre from which cell values are
     *        returned.
     * @param cellRowIndex the row index at y.
     * @param cellColIndex the col index at x.
     * @param distance the radius of the circle for which intersected cell
     *        values are returned.
     * TODO:
     * remove need for copy to new array.
     */
    protected CellID[] getCellIDs(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            double distance ) {
        CellID[] cellIDs0;
        int cellDistance = ( int ) Math.ceil( distance / this.dimensions[ 0 ].doubleValue() );
        int limit = ( ( 2 * cellDistance ) + 1 ) * ( ( 2 * cellDistance ) + 1 );
        cellIDs0 = new CellID[ limit ];
        long row;
        long col;
        double thisX;
        double thisY;
        int count = 0;
        //if ( limit > 0 ) {
        for ( row = cellRowIndex - cellDistance; row <= cellRowIndex + cellDistance; row ++ ) {
            thisY = getCellYDouble( row );
            for ( col = cellColIndex - cellDistance; col <= cellColIndex + cellDistance; col ++ ) {
                thisX = getCellXDouble( col );
                if ( Utilities.distance( x, y, thisX, thisY ) <= distance ) {
                    cellIDs0[ count ] = new CellID(
                            this.ncols,
                            row,
                            col );
                    //cellIDs0[ count ] = new CellID( row, col );
                    count ++;
                }
            }
        }
        CellID[] cellIDs = new CellID[ count ];
        System.arraycopy( cellIDs0, 0, cellIDs, 0, count );
        return cellIDs;
        //}
        //return cellIDs0;
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by x-coordinate x, y-coordinate y.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID getNearestCellID(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestCellID(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNearestCellID(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by x-coordinate x, y-coordinate y.
     * @param x the x-coordinate of the point.
     * @param y the y-coordinate of the point.
     */
    protected CellID getNearestCellID(
            double x,
            double y ) {
        return getNearestCellID(
                x,
                y, getCellRowIndex( y ),
                getCellColIndex( x ) );
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by cell row index cellRowIndex, cell column index
     * cellColIndex.
     * @param cellRowIndex the row index from which nearest cell ID is returned.
     * @param cellColIndex the column index from which nearest cell ID is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * return CellID[] as could be more than one nearest CellID
     */
    public CellID getNearestCellID(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestCellID(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNearestCellID(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest cells cellRowIndex and cellColIndex as a long[]
     * from ID to point given by cell row index cellRowIndex, cell column index
     * cellColIndex.
     * @param cellRowIndex the row index from which nearest cell ID is returned.
     * @param cellColIndex the column index from which nearest cell ID is returned.
     * TODO:
     * return CellID[] as could be more than one nearest CellID
     */
    protected CellID getNearestCellID(
            long cellRowIndex,
            long cellColIndex ) {
        return getNearestCellID(
                getCellXDouble( cellColIndex ),
                getCellYDouble( cellRowIndex ),
                cellRowIndex,
                cellColIndex );
    }
    
    /**
     * Returns the the nearest CellID to point given by x-coordinate x,
     * y-coordinate y in position given by cellRowIndex, cellColIndex.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of cell containing point.
     * @param cellColIndex The cell column index of cell containing point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public CellID getNearestCellID(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getNearestCellID(
                    x,
                    y,
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNearestCellID(
                        x,
                        y,
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the the nearest CellID to point given by x-coordinate x,
     * y-coordinate y in position given by cellRowIndex, cellColIndex.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param cellRowIndex The cell row index of cell containing point.
     * @param cellColIndex The cell column index of cell containing point.
     */
    protected CellID getNearestCellID(
            double x,
            double y,
            long cellRowIndex,
            long cellColIndex ) {
        CellID cellID;
        boolean isInGrid = isInGrid(
                x,
                y );
        if ( ! isInGrid ) {
            long row;
            long col;
            if ( x >= this.dimensions[3].doubleValue() ) {
                col = this.ncols - 1;
                if ( y > this.dimensions[4].doubleValue() ) {
                    row = 0;
                } else {
                    if ( y < this.dimensions[2].doubleValue() ) {
                        row = this.nrows - 1;
                    } else {
                        row = getCellRowIndex( y );
                    }
                }
            } else {
                if ( x < this.dimensions[1].doubleValue() ) {
                    col = 0;
                    if ( y >= this.dimensions[4].doubleValue() ) {
                        row = 0;
                    } else {
                        if ( y < this.dimensions[2].doubleValue() ) {
                            row = this.nrows -1;
                        } else {
                            row = getCellRowIndex( y );
                        }
                    }
                } else {
                    col = getCellColIndex( x );
                    if ( y >= dimensions[4].doubleValue() ) {
                        row = 0;
                    } else {
                        row = this.nrows - 1;
                    }
                }
            }
            cellID = new CellID(
                    this.ncols,
                    row,
                    col );
        } else {
            cellID = new CellID(
                    this.ncols,
                    cellRowIndex,
                    cellColIndex );
        }
        return cellID;
    }
    
    /**
     * Returns the height of the grid.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getHeightDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return getHeightDouble();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getHeightDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the height of the grid.
     */
    protected final double getHeightDouble() {
        return getHeightBigDecimal().doubleValue();
    }
    
    /**
     * Returns the height of the grid.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getHeightBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            return getHeightBigDecimal();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getHeightBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the height of the grid.
     */
    protected final BigDecimal getHeightBigDecimal() {
        return ( this.dimensions[ 4 ].subtract( this.dimensions[ 2 ] ) );
    }
    
    /**
     * Returns the width of the grid as a double.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getWidthDouble(
            boolean handleOutOfMemoryError ) {
        try {
            return getWidthDouble();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getWidthDouble( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the width of the grid as a double.
     */
    protected final double getWidthDouble() {
        return getWidthBigDecimal().doubleValue();
    }
    
    /**
     * Returns the width of the grid as a BigDecimal.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getWidthBigDecimal(
            boolean handleOutOfMemoryError ) {
        try {
            return getWidthBigDecimal();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getWidthBigDecimal( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    /**
     * Returns the width of the grid as a BigDecimal.
     */
    protected final BigDecimal getWidthBigDecimal() {
        return ( this.dimensions[ 3 ].subtract( this.dimensions[ 1 ] ) );
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param xBigDecimal The x-coordinate of the point.
     * @param yBigDecimal The y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            BigDecimal xBigDecimal,
            BigDecimal yBigDecimal,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid(
                    xBigDecimal,
                    yBigDecimal );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        xBigDecimal,
                        yBigDecimal,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param xBigDecimal The x-coordinate of the point.
     * @param yBigDecimal The y-coordinate of the point.
     */
    protected final boolean isInGrid(
            BigDecimal xBigDecimal,
            BigDecimal yBigDecimal ) {
        return ( xBigDecimal.compareTo( this.dimensions[ 1 ] ) != -1 &&
                yBigDecimal.compareTo( this.dimensions[ 2 ] ) != -1 &&
                xBigDecimal.compareTo( this.dimensions[ 3 ] ) != 1 &&
                yBigDecimal.compareTo( this.dimensions[ 4 ] ) != 1 );
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff point given by x-coordinate x, y-coordinate y is
     * in the Grid. Anything on the boundary is considered to be in.
     * @param x The x-coordinate of the point.
     * @param y The y-coordinate of the point.
     */
    protected final boolean isInGrid(
            double x,
            double y ) {
        return isInGrid(
                BigDecimal.valueOf( x ),
                BigDecimal.valueOf( y ) );
    }
    
    /**
     * Returns true iff position given by cell row index cellRowIndex, cell
     * column index cellColIndex is in the Grid.
     * @param cellRowIndex The cell row index to test.
     * @param cellColIndex The cell column index to test.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff position given by cell row index cellRowIndex, cell
     * column index cellColIndex is in the Grid.
     * @param cellRowIndex The cell row index to test.
     * @param cellColIndex The cell column index to test.
     */
    protected final boolean isInGrid(
            long cellRowIndex,
            long cellColIndex ) {
        return ( cellRowIndex >= 0 &&
                cellRowIndex < this.nrows &&
                cellColIndex >= 0 &&
                cellColIndex < this.ncols );
    }
    
    /**
     * Returns true iff position given by chunk row index chunkRowIndex, chunk
     * column index chunkColIndex is in the Grid.
     * @param chunkRowIndex The chunk row index to test.
     * @param chunkColIndex The chunk column index to test.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid(
                    chunkRowIndex,
                    chunkColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        chunkRowIndex,
                        chunkColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff position given by chunk row index chunkRowIndex, chunk
     * column index chunkColIndex is in the Grid.
     * @param chunkRowIndex The chunk row index to test.
     * @param chunkColIndex The chunk column index to test.
     */
    protected final boolean isInGrid(
            int chunkRowIndex,
            int chunkColIndex ) {
        return ( chunkRowIndex >= 0 &&
                chunkRowIndex < this.nrows &&
                chunkColIndex >= 0 &&
                chunkColIndex < this.ncols );
    }
    
    /**
     * Returns true iff cell given by cellID is in the Grid.
     * @param cellID The CellID of a cell to test.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff cell given by cellID is in the Grid.
     * @param cellID The CellID of a cell to test.
     */
    protected final boolean isInGrid(
            CellID cellID ) {
        return isInGrid(
                cellID.getCellRowIndex(),
                cellID.getCellColIndex() );
    }
    
    /**
     * Returns true iff ChunkID chunkID is in the Grid.
     * @param chunkID The ChunkID of a cell to test.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            ChunkID chunkID,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid( chunkID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        chunkID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff ChunkID chunkID is in the Grid.
     * @param chunkID The ChunkID of a cell to test.
     */
    protected final boolean isInGrid(
            ChunkID chunkID ) {
        int chunkRowIndex = chunkID.chunkRowIndex;
        int chunkColIndex = chunkID.chunkColIndex;
        return chunkRowIndex > -1 &&
                chunkRowIndex < this.nChunkRows &&
                chunkColIndex > -1 &&
                chunkColIndex < this.nChunkCols;
    }
    
    /**
     * Returns true iff cell given by chunkRowIndex, chunkColIndex,
     * chunkCellRowIndex, chunkCellColIndex is in the Grid.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final boolean isInGrid(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return isInGrid(
                    chunkRowIndex,
                    chunkColIndex,
                    chunkCellRowIndex,
                    chunkCellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return isInGrid(
                        chunkRowIndex,
                        chunkColIndex,
                        chunkCellRowIndex,
                        chunkCellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns true iff cell given by chunkRowIndex, chunkColIndex,
     * chunkCellRowIndex, chunkCellColIndex is in the Grid.
     * @param chunkRowIndex
     * @param chunkColIndex
     * @param chunkCellRowIndex
     * @param chunkCellColIndex
     */
    protected final boolean isInGrid(
            int chunkRowIndex,
            int chunkColIndex,
            int chunkCellRowIndex,
            int chunkCellColIndex ) {
        return isInGrid(
                ( ( long ) chunkRowIndex * ( long ) this.chunkNcols ) + ( long ) chunkCellRowIndex,
                ( ( long ) chunkColIndex * ( long ) this.chunkNrows ) + ( long ) chunkCellColIndex );
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a BigDecimal.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellXBigDecimal( cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellXBigDecimal(
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a BigDecimal.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            long cellColIndex,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellXBigDecimal( cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellXBigDecimal(
                        cellColIndex,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellXBigDecimal(
                        cellColIndex,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a BigDecimal.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     */
    protected final BigDecimal getCellXBigDecimal(
            long cellColIndex ) {
        BigDecimal bigDecimal0 = BigDecimal.valueOf( cellColIndex );
        BigDecimal bigDecimal1 = dimensions[ 0 ].multiply( bigDecimal0 );
        BigDecimal bigDecimal2 = dimensions[ 1 ].add( bigDecimal1 );
        bigDecimal1 =  BigDecimal.valueOf( 2L );
        bigDecimal0 = dimensions[ 0 ].divide(
                bigDecimal1,
                this.dimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return bigDecimal2.add( bigDecimal0 );
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellXDouble(
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellXDouble(
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param chunkCellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param chunkRowIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            int chunkCellColIndex,
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
        return getCellXDouble( chunkCellColIndex, chunkColIndex );
    } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkID = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return getCellXDouble(
                        chunkCellColIndex, chunkRowIndex, chunkColIndex, 
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param chunkCellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    protected final double getCellXDouble(
            int chunkCellColIndex,
            int chunkColIndex ) {
        return getCellXDouble( getCellColIndex( chunkCellColIndex, chunkColIndex ) );
    }
    
    /**
     * Returns the x-coordinate of the centroid for cells with column index
     * cellColIndex as a double.
     * @param cellColIndex The cell column index thats centroid x-coordinate is
     * returned.
     */
    protected final double getCellXDouble(
            long cellColIndex ) {
        return getCellXBigDecimal( cellColIndex ).doubleValue();
    }
    
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a BigDecimal.
     * @param cellID
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellXBigDecimal( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellXBigDecimal(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid of cell with CellID cellID as a
     * getCellXBigDecimal.
     * @param cellID The CellID of the cell thats centroid is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final BigDecimal getCellXBigDecimal(
            CellID cellID,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellXBigDecimal( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellXBigDecimal(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellXBigDecimal(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a BigDecimal.
     * @param cellID
     */
    protected final BigDecimal getCellXBigDecimal(
            CellID cellID ) {
        BigDecimal bigDecimal0 = BigDecimal.valueOf( cellID.cellColIndex );
        BigDecimal bigDecimal1 = dimensions[ 0 ].multiply( bigDecimal0 );
        BigDecimal bigDecimal2 = dimensions[ 1 ].add( bigDecimal1 );
        bigDecimal0 = BigDecimal.valueOf( 2L );
        bigDecimal1 = dimensions[ 0 ].divide(
                bigDecimal0,
                this.dimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return bigDecimal2.add( bigDecimal1 );
    }
    
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a double
     * @param cellID
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellXDouble(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellXDouble( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellXDouble(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the x-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID The CellID of the cell thats centroid is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final double getCellXDouble(
            CellID cellID,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellXDouble( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellXDouble(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellXDouble(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
   
    /**
     * Returns the x-coordinate of the centroid for cell with cell ID cellID as
     * a double
     * @param cellID
     */
    protected final double getCellXDouble(
            CellID cellID ) {
        return getCellXBigDecimal( cellID ).doubleValue();
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a BigDecimal.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            long cellRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellYBigDecimal( cellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellYBigDecimal(
                        cellRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a BigDecimal.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            long cellRowIndex,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellYBigDecimal( cellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellYBigDecimal(
                        cellRowIndex,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellYBigDecimal(
                        cellRowIndex,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a BigDecimal.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     *   returned.
     */
    protected final BigDecimal getCellYBigDecimal(
            long cellRowIndex ) {
        BigDecimal bigDecimal0 = BigDecimal.valueOf( cellRowIndex );
        BigDecimal bigDecimal1 = dimensions[ 0 ].multiply( bigDecimal0 );
        BigDecimal bigDecimal2 = dimensions[ 2 ].add( bigDecimal1 );
        bigDecimal0 = BigDecimal.valueOf( 2L );
        bigDecimal1 = dimensions[ 0 ].divide(
                bigDecimal0,
                this.dimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return bigDecimal2.add( bigDecimal1 );
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            long cellRowIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellYDouble( cellRowIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellYDouble(
                        cellRowIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param chunkCellRowIndex the chunk cell column index thats centroid 
     *   y-coordinate is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            int chunkCellRowIndex,
            int chunkRowIndex,
            int chunkColIndex,
            boolean handleOutOfMemoryError ) {
        try {
        return getCellYDouble( chunkCellRowIndex, chunkRowIndex );
    } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                ChunkID chunkID = new ChunkID( nChunkCols, chunkRowIndex, chunkColIndex );
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID );
                initMemoryReserve();
                return getCellYDouble(
                        chunkCellRowIndex, chunkRowIndex, chunkColIndex, 
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param chunkCellRowIndex the chunk cell column index thats centroid 
     *   y-coordinate is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    protected final double getCellYDouble(
            int chunkCellRowIndex,
            int chunkRowIndex ) {
       return getCellYDouble( getCellRowIndex( chunkCellRowIndex, chunkRowIndex ) );
    }
    
    /**
     * Returns the y-coordinate of the centroid for cells with row index
     * cellRowIndex as a double.
     * @param cellRowIndex the cell column index thats centroid y-coordinate is
     * returned.
     */
    protected final double getCellYDouble(
            long cellRowIndex ) {
        return getCellYBigDecimal( cellRowIndex ).doubleValue();
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * BigDecimal.
     * @param cellID The CellID of the cell thats centroid is returned.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellYBigDecimal( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellYBigDecimal(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
      * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * BigDecimal.
     * @param cellID The CellID of the cell thats centroid is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final BigDecimal getCellYBigDecimal(
            CellID cellID,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellYBigDecimal( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellYBigDecimal(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellYBigDecimal(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * BigDecimal.
     * @param cellID The CellID of the cell thats centroid is returned.
     */
    protected final BigDecimal getCellYBigDecimal(
            CellID cellID ) {
        BigDecimal bigDecimal0 = BigDecimal.valueOf( cellID.cellRowIndex );
        BigDecimal bigDecimal1 = dimensions[ 0 ].multiply( bigDecimal0 );
        BigDecimal bigDecimal2 = dimensions[ 2 ].add( bigDecimal1 );
        bigDecimal0 = BigDecimal.valueOf( 2L );
        bigDecimal1 = dimensions[ 0 ].divide(
                bigDecimal0,
                this.dimensionsScale,
                BigDecimal.ROUND_HALF_EVEN );
        return bigDecimal2.add( bigDecimal1 );
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double getCellYDouble(
            CellID cellID,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellYDouble( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellYDouble(
                        cellID,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
      * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID The CellID of the cell thats centroid is returned.
     * @param chunkRowIndex The chunk row index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     * @param chunkColIndex The chunk column index of the 
     *   Grid2DSquareCellChunkAbstract not to be swapped if an OutOfMemoryError 
     *   is thrown.
     */
    public final double getCellYDouble(
            CellID cellID,
            int chunkRowIndex,
            int chunkColIndex ) {
        try {
            return getCellYDouble( cellID );
        } catch ( OutOfMemoryError oome0 ) {
            clearMemoryReserve();
            ChunkID chunkID = new ChunkID(
                    this.nChunkCols, 
                    chunkRowIndex, 
                    chunkColIndex );
            if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunkExcept( chunkID ) > 0 ) {
                initMemoryReserve();
                return getCellYDouble(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
            } else {
                if ( swapToFileGrid2DSquareCellsGrid2DSquareCellChunk() > 0 ) {
                    initMemoryReserve();
                    return getCellYDouble(
                        cellID,
                        chunkRowIndex,
                        chunkColIndex );
                } else {
                    throw oome0;
                }
            }
        }
    }
    
    /**
     * Returns the y-coordinate of the centroid of cell with CellID cellID as a
     * double.
     * @param cellID
     */
    protected final double getCellYDouble(
            CellID cellID ) {
        return getCellYBigDecimal( cellID ).doubleValue();
    }
    
    /**
     * Returns gridBounds (the bounding box of the grid) as a double[] where;
     * gridBounds[0] xmin, left most x-coordinate of this
     * gridBounds[1] ymin, lowest y-coordinate of this
     * gridBounds[2] xmax, right most x-coordinate of this
     * gridBounds[3] ymax, highest y-coordinate of this
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Are bounds in double range?
     * Is there more than cellsize difference with precision?
     *   Throw appropriate exceptions.
     */
    public final double[] getGridBounds(
            boolean handleOutOfMemoryError ) {
        try {
            return getGridBounds();
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getGridBounds( handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns gridBounds (the bounding box of the grid) as a double[] where;
     * gridBounds[0] xmin, left most x-coordinate of this
     * gridBounds[1] ymin, lowest y-coordinate of this
     * gridBounds[2] xmax, right most x-coordinate of this
     * gridBounds[3] ymax, highest y-coordinate of this
     * TODO:
     * Are bounds in double range?
     * Is there more than cellsize difference with precision?
     *   Throw appropriate exceptions.
     */
    protected final double[] getGridBounds() {
        double[] result = new double[ 4 ];
        result[ 0 ] = dimensions[ 1 ].doubleValue();
        result[ 1 ] = dimensions[ 2 ].doubleValue();
        result[ 2 ] = dimensions[ 3 ].doubleValue();
        result[ 3 ] = dimensions[ 4 ].doubleValue();
        return result;
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell at (rowIndex,colIndex)
     * double[1] ymin, lowest y-coordinate of cell at (rowIndex,colIndex)
     * double[2] xmax, right most x-coordinate of cell at (rowIndex,colIndex)
     * double[3] ymax, highest y-coordinate of cell at (rowIndex,colIndex)
     * @param cellRowIndex the row index of the cell for which the bounds are
     *   returned
     * @param cellColIndex the column index of the cell for which the bounds are
     *   returned
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public final double[] getCellBoundsDoubleArray(
            long cellRowIndex,
            long cellColIndex,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellBoundsDoubleArray(
                    cellRowIndex,
                    cellColIndex );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellBoundsDoubleArray(
                        cellRowIndex,
                        cellColIndex,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell at (rowIndex,colIndex)
     * double[1] ymin, lowest y-coordinate of cell at (rowIndex,colIndex)
     * double[2] xmax, right most x-coordinate of cell at (rowIndex,colIndex)
     * double[3] ymax, highest y-coordinate of cell at (rowIndex,colIndex)
     * @param cellRowIndex the row index of the cell for which the bounds are
     *   returned
     * @param cellColIndex the column index of the cell for which the bounds are
     *   returned
     */
    protected final double[] getCellBoundsDoubleArray(
            long cellRowIndex,
            long cellColIndex ) {
        return getCellBoundsDoubleArray(
                getCellXDouble( cellColIndex ),
                getCellYDouble( cellRowIndex ) );
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell that intersects point at (x,y)
     * double[1] ymin, lowest y-coordinate of cell that intersects point at (x,y)
     * double[2] xmax, right most x-coordinate of cell that intersects point at (x,y)
     * double[3] ymax, highest y-coordinate of cell that intersects point at (x,y)
     * @param x the x-coordinate in the cell for which the bounds are returned
     * @param y the y-coordinate in the cell for which the bounds are returned
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Precision may compromise result. More precision is available via
     * BigDecimal arithmetic.
     */
    public final double[] getCellBoundsDoubleArray(
            double x,
            double y,
            boolean handleOutOfMemoryError ) {
        try {
            return getCellBoundsDoubleArray(
                    x,
                    y );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getCellBoundsDoubleArray(
                        x,
                        y,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns double[] where;
     * double[0] xmin, left most x-coordinate of cell that intersects point at (x,y)
     * double[1] ymin, lowest y-coordinate of cell that intersects point at (x,y)
     * double[2] xmax, right most x-coordinate of cell that intersects point at (x,y)
     * double[3] ymax, highest y-coordinate of cell that intersects point at (x,y)
     * @param x the x-coordinate in the cell for which the bounds are returned
     * @param y the y-coordinate in the cell for which the bounds are returned
     * TODO:
     * Precision may compromise result. More precision is available via
     * BigDecimal arithmetic.
     */
    protected final double[] getCellBoundsDoubleArray(
            double x,
            double y ) {
        double[] cellBounds = new double[ 4 ];
        double halfCellsize = dimensions[ 0 ].doubleValue() / 2.0d;
        cellBounds[ 0 ] = x - halfCellsize;
        cellBounds[ 1 ] = y - halfCellsize;
        cellBounds[ 2 ] = x + halfCellsize;
        cellBounds[ 3 ] = y + halfCellsize;
        return cellBounds;
    }
    
    /**
     * Returns the noDataValue of this as a BigDecimal.
     */
    public abstract BigDecimal getNoDataValueBigDecimal(
            boolean handleOutOfMemoryError );
    
    /**
     * Returns the next ChunkID in row major order from chunkID, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getNextChunk(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols,
            boolean handleOutOfMemoryError ) {
        try {
            return getNextChunk(
                    chunkID,
                    nChunkRows,
                    nChunkCols );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return  getNextChunk(
                        chunkID,
                        nChunkRows,
                        nChunkCols,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row major order from chunkID, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    protected ChunkID getNextChunk(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols ) {
        int chunkRowIndex = chunkID.getChunkRowIndex();
        int chunkColIndex = chunkID.getChunkColIndex();
        if ( chunkColIndex < nChunkCols - 1 ) {
            return new ChunkID(
                    nChunkCols,
                    chunkRowIndex,
                    chunkColIndex + 1 );
        } else {
            if ( chunkRowIndex < nChunkRows - 1  ) {
                return new ChunkID(
                        nChunkCols,
                        chunkRowIndex + 1,
                        0 );
            }
        }
        return null;
    }
    
    /**
     * Returns the next ChunkID in row minor order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getNextChunkRowMinor(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols,
            boolean handleOutOfMemoryError ) {
        try {
            return getNextChunkRowMinor(
                    chunkID,
                    nChunkRows,
                    nChunkCols );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getNextChunkRowMinor(
                        chunkID,
                        nChunkRows,
                        nChunkCols,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row minor order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    protected ChunkID getNextChunkRowMinor(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols ) {
        int chunkRowIndex = chunkID.getChunkRowIndex();
        int chunkColIndex = chunkID.getChunkColIndex();
        if ( chunkColIndex < nChunkCols - 1 ) {
            return new ChunkID(
                    nChunkCols,
                    chunkRowIndex,
                    chunkColIndex + 1 );
        } else {
            if ( chunkRowIndex > 0 ) {
                return new ChunkID(
                        nChunkCols,
                        chunkRowIndex - 1,
                        0 );
            }
        }
        return null;
    }
    
    /**
     * Returns the next ChunkID in row major order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    public ChunkID getPreviousChunk(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols,
            boolean handleOutOfMemoryError ) {
        try {
            return getPreviousChunk(
                    chunkID,
                    nChunkRows,
                    nChunkCols );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                clearMemoryReserve();
                swapToFileGrid2DSquareCellsGrid2DSquareCellChunk();
                initMemoryReserve();
                return getPreviousChunk(
                        chunkID,
                        nChunkRows,
                        nChunkCols,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Returns the next ChunkID in row major order from this, or null.
     * @param chunkID
     * @param nChunkRows
     * @param nChunkCols
     * TODO:
     * Is it inefficient to pass in nChunkRows and nChunkCols since this has them already?
     */
    protected ChunkID getPreviousChunk(
            ChunkID chunkID,
            int nChunkRows,
            int nChunkCols ) {
        int chunkRowIndex = chunkID.getChunkRowIndex();
        int chunkColIndex = chunkID.getChunkColIndex();
        if ( chunkColIndex > 0 ) {
            return new ChunkID(
                    nChunkCols,
                    chunkRowIndex,
                    chunkColIndex - 1 );
        } else {
            if ( chunkRowIndex > 0  ) {
                return new ChunkID(
                        nChunkCols,
                        chunkRowIndex - 1,
                        0 );
            }
        }
        return null;
    }
    
    
    /**
     * A simple CellID class for distinguishing cells.
     */
    public static class CellID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the row in which the cell is positioned
         */
        protected long cellRowIndex;
        
        /**
         * For storing the column in which the cell is positioned
         */
        protected long cellColIndex;
        
        /**
         * For storing a hashCode for the cell.
         * NB. When the number of cells in a grid is more than the maximum int
         * value care should be taken that hashCodes are distributed in a
         * relatively even way. Operations will slow down greatly if a large
         * number of cells have a CellID with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public CellID() {}
        
        /**
         * Constructs a new CellID that is a clone of cellID
         * @param cellID The CellID to clone.
         */
        public CellID( CellID cellID ) {
            this.cellColIndex = cellID.cellColIndex;
            this.cellRowIndex = cellID.cellRowIndex;
            this.hashCode = cellID.hashCode;
        }
        
        /**
         * Constructs a new CellID
         * @param cellRowIndex The row of the cell.
         * @param cellColIndex The column of the cell.
         * @param ncols The number of columns in a grid2DSquareCellAbstract.
         */
        public CellID( long ncols, long cellRowIndex, long cellColIndex ) {
            //this( nrows, ncols, cellRowIndex, cellColIndex );
            this.cellRowIndex = cellRowIndex;
            this.cellColIndex = cellColIndex;
            this.hashCode = ( int ) ( ( ncols * cellRowIndex ) + cellColIndex );
        }
        
        //public CellID( long cellRowIndex, long cellColIndex ) {
        //    this( ncols, cellRowIndex, cellColIndex );
        //    this.cellRowIndex = cellRowIndex;
        //    this.cellColIndex = cellColIndex;
        //    this.hashCode = ( int ) ( ( this.ncols * cellRowIndex ) + cellColIndex );
        //}
        
        /**
         * Returns hashCode of this CellID
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * Returns cellRowIndex of this CellID
         */
        public long getCellRowIndex() {
            return this.cellRowIndex;
        }
        
        /**
         * Returns cellColIndex of this CellID
         */
        public long getCellColIndex() {
            return this.cellColIndex;
        }
        
        /**
         * Returns a description of this CellID
         */
        public String toString() {
            return "CellID( hashCode( " + hashCode() + "), cellRowIndex( " + getCellRowIndex() + " ), cellColIndex( " + getCellColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            CellID anotherCellID = ( CellID ) object;
            return ( ( this.cellColIndex == anotherCellID.cellColIndex ) && ( this.cellRowIndex == anotherCellID.cellRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should result in NullPointerException or ClassCastException
            CellID cellID = ( CellID ) object;
            if ( cellID.cellRowIndex > this.cellRowIndex ) {
                return 1;
            }
            if ( cellID.cellRowIndex < this.cellRowIndex ) {
                return -1;
            }
            if ( cellID.cellColIndex > this.cellColIndex ) {
                return 1;
            }
            if ( cellID.cellColIndex < this.cellColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
    }
    
    /**
     * A simple ChunkID class for distinguishing chunks.
     */
    public static class ChunkID extends Object implements Serializable, Comparable {
        
        /**
         * For storing the chunk row index
         */
        protected int chunkRowIndex;
        
        /**
         * For storing the chunk column index
         */
        protected int chunkColIndex;
        
        /**
         * For storing a hashCode for this ChunkID.
         * NB. When the number of cells in a chunk is more than the maximum int
         * value care should be taken to distributed hashcodes in a relatively
         * even way. Operations may slow down greatly if a large number of
         * chunks have ChunkIDs with the same hashCode.
         */
        protected int hashCode;
        
        /**
         * Default constructor.
         */
        public ChunkID() {
            this.chunkColIndex = Integer.MIN_VALUE;
            this.chunkRowIndex = Integer.MIN_VALUE;
            this.hashCode = Integer.MIN_VALUE;
        }
        
        /**
         * Constructs a new ChunkID that is a clone of chunkID
         * @param chunkID The ChunkID to clone.
         */
        public ChunkID( ChunkID chunkID ) {
            this.chunkColIndex = chunkID.chunkColIndex;
            this.chunkRowIndex = chunkID.chunkRowIndex;
            this.hashCode = chunkID.hashCode;
        }
        
        /**
         * Constructs a new ChunkID
         * @param chunkRowIndex the row of the
         * @param chunkColIndex the column of the
         * @param nChunkCols The number of chunk columns (of the grid2DSquareCellAbstract).
         */
        public ChunkID(
                int nChunkCols,
                int chunkRowIndex,
                int chunkColIndex ) {
            this.chunkRowIndex = chunkRowIndex;
            this.chunkColIndex = chunkColIndex;
            this.hashCode = ( int ) ( ( nChunkCols * chunkRowIndex ) + chunkColIndex );
        }
        
        /**
         * Returns this.hashCode
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * Returns this.chunkRowIndex
         */
        public int getChunkRowIndex() {
            return this.chunkRowIndex;
        }
        
        /**
         * Returns this.chunkColIndex
         */
        public int getChunkColIndex() {
            return this.chunkColIndex;
        }
        
        /**
         * Returns a description of this
         */
        public String toString() {
            return  "ChunkID( " +
                    "hashCode( " + hashCode() + "), " +
                    "chunkRowIndex( " + getChunkRowIndex() + " ), " +
                    "chunkColIndex( " + getChunkColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            ChunkID anotherChunkID = ( ChunkID ) object;
            return ( ( this.chunkColIndex == anotherChunkID.chunkColIndex ) && ( this.chunkRowIndex == anotherChunkID.chunkRowIndex ) );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            // Cast failure should throw NullPointerException or ClassCastException
            ChunkID chunkID = ( ChunkID ) object;
            if ( chunkID.chunkRowIndex > this.chunkRowIndex ) {
                return 1;
            }
            if ( chunkID.chunkRowIndex < this.chunkRowIndex ) {
                return -1;
            }
            if ( chunkID.chunkColIndex > this.chunkColIndex ) {
                return 1;
            }
            if ( chunkID.chunkColIndex < this.chunkColIndex ) {
                return -1;
            }
            return 0;
            //} catch ( Exception e0 ) {
            //    return -1;
            //}
        }
        
    }
    
}